/*
threads.cpp - C Thread support library

This file is a part of Fiku.
Copyright (C) 2021 Akib Azmain Turja

Fiku is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Fiku is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

Authors: Akib Azmain Turja
*/

#include <threads.h>
#include <fiku/proc.hpp>
#include <string.h>
#include <errno.h>
#include <fiku/cpu.h>

struct _thrd_info
{
  proc::thrd_t handle;
  static constexpr size_t stack_size = 0x1000;
  char stack[stack_size];
  volatile bool exited;
  volatile int retval;
};

namespace
{
  __attribute__ ((noreturn)) void
  start_thrd (thrd_start_t func, void *arg)
  {
    thrd_exit (func (arg));
  }

  void
  delete_thrd (void */* thrd */)
  {
    // delete reinterpret_cast < thrd_t > (thrd); // <-- Triple fault

    // We can't delete the structure, as it contains the stack, and it is
    // possible that this function is executed while we are using that
    // stack. If we unmap our stack, accessing the stack will cause a page
    // fault, then the processor will try to push data and will cause
    // another page fault, resulting a double fault, followed by a page
    // fault again, causing triple fault (and reset).
  }
}

int
thrd_create (thrd_t *thrdp, thrd_start_t func, void *arg)
{
  thrd_t thrd = new _thrd_info;

  if (errno == EAGAIN)
    {
      return thrd_nomem;
    }

  *thrdp = thrd;

  memset (thrd, 0, sizeof (_thrd_info));
  thrd->handle.after_detach = delete_thrd;
  thrd->handle.after_detach_arg = thrd;

  asm volatile ("movl %%cr3, %0"
                : "=a" (thrd->handle.vas));

  thrd->handle.reg_state = new idt::registers_t;
  thrd->handle.reg_state->eflags = 0x202;
  thrd->handle.reg_state->cs = 0x8;
  thrd->handle.reg_state->ds = 0x10;
  thrd->handle.reg_state->eip = reinterpret_cast < size_t > (start_thrd);
  thrd->handle.reg_state->esp =
    (reinterpret_cast < size_t > (thrd->stack + thrd->stack_size)
     - sizeof (size_t) * 3); // 2 for arguments, another for returning
                             // (unused)

  // Now write the arguments into stack
  (reinterpret_cast < size_t *> (thrd->stack + thrd->stack_size) - 3)[1]
    = (reinterpret_cast < size_t > (func));
  (reinterpret_cast < size_t *> (thrd->stack + thrd->stack_size) - 3)[2]
    = (reinterpret_cast < size_t > (arg));

  proc::thrd_prep (&thrd->handle);
  delete thrd->handle.reg_state;

  thrd->handle.thrd_state = thrd->handle.thrd_runnable;
  proc::thrd_add (&thrd->handle);

  return thrd_success;
}

void
thrd_yield ()
{

  // Don't yield if cli count is non zero
  if (cpu_current ()->cli_count > 0)
    {
      return;
    }

  // Call thread yield system call, number 0x0 (we're using system calls
  // because we need to save register state and pass the amount of time we
  // have used in this thread to the thread switcher)
  asm volatile ("int $0xff"
                :
                : "a" (0x0));
}
