/*
cpuid.h - Functions for executing CPUID functions

This file is a part of Fiku.
Copyright (C) 2021 Akib Azmain Turja

Fiku is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Fiku is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

Authors: Akib Azmain Turja
*/

#include <fiku/cpuid.h>
#include <fiku/idt.hpp>
#include <string.h>

extern "C" bool cpuid_check_eflags_id_bit_ ();

namespace
{
  bool cpuid_supported;

  void invalid_opcode_handler_tmp (uint8_t, uint8_t, idt::registers_t &)
  {
    cpuid_supported = false;
  }
}

bool
cpuid_check ()
{
  static bool executed_previously;
  static bool supported;

  // If this function is executed before, return the cached value
  if (!executed_previously)
    {
      executed_previously = true;
      if (cpuid_check_eflags_id_bit_ ())
        {
          return supported = true;
        }

      // Fliping the ID bit isn't supported, execute cpuid directly and
      // wait for invalid opcode exception

      // This will be set to false if cpuid is unsupported
      cpuid_supported = true;

      // Backup previous handler for invalid opcode exception
      idt::interrupt_handler previous_handler = idt::get_handler (0x6);

      // Set custom handler
      idt::set_handler (0x6, invalid_opcode_handler_tmp);

      // Temporary space for putting unused values
      cpuid_data_t null;

      // Execute cpuid
      cpuid_exec_p ({ 0x0, 0x0, 0x0, 0x0 }, &null);

      // Restore previous handler
      idt::set_handler (0x6, previous_handler);

      // If any invaild opcode exception has occurred, the custom handler
      // will set "cpuid_supported" to false
      supported = cpuid_supported;
    }

  return supported;
}

bool
cpuid_check_extended ()
{
  static bool previously_executed;
  static bool ret;
  if (!previously_executed)
    {
      ret = cpuid_exec ({ 0x80000000, 0x0, 0x0, 0x0 }).eax > 0x80000000;
      previously_executed = true;
    }
  return ret;
}

cpuid_data_t
cpuid_exec (cpuid_data_t function)
{
  cpuid_data_t ret;
  cpuid_exec_p (function, &ret);
  return ret;
}

const char*
cpuid_vendor ()
{
  static bool executed_previously;
  static char vendor[13]; // One extra for trailing null

  // If this function is executed before, do nothing and return the cached
  // value
  if (!executed_previously)
    {
      cpuid_data_t ret = cpuid_exec ({ 0x0, 0x0, 0x0, 0x0 });
      memcpy(vendor, &ret.ebx, 4);
      memcpy(vendor + 4, &ret.edx, 4);
      memcpy(vendor + 8, &ret.ecx, 4);
      vendor[12] = 0;
      executed_previously = true;
    }

  return vendor;
}

cpuid_data_t
cpuid_getfeatures ()
{
  static bool executed_previously;
  static cpuid_data_t ret;

  // If this function is executed before, do nothing and return the cached
  // value
  if (!executed_previously)
    {
      cpuid_exec_p ({ 0x0, 0x0, 0x0, 0x0 }, &ret);
      executed_previously = true;
    }

  return ret;
}

bool
cpuid_feature_pse ()
{
  return cpuid_getfeatures ().edx >> 3 & 1;
}

bool
cpuid_feature_pae ()
{
  return cpuid_getfeatures ().edx >> 6 & 1;
}

uint8_t
cpuid_pae_maxphyaddr ()
{
  static bool executed_previously;
  static uint8_t maxphyaddr;

  // If this function is executed before, do nothing and return the cached
  // value
  if (!executed_previously)
    {
      if (cpuid_check_extended ())
        {
          maxphyaddr = (cpuid_exec ({ 0x80000008, 0x0, 0x0, 0x0 }).eax
                        & 0xff);
        }
      else
        {
          maxphyaddr = 36;
        }

      executed_previously = true;
    }

  return maxphyaddr;
}

bool
cpuid_feature_nx ()
{
  static bool executed_previously;
  static bool ret;

  // If this function is executed before, do nothing and return the cached
  // value
  if (!executed_previously)
    {
      if (cpuid_check_extended ())
        {
          ret = (cpuid_exec ({ 0x80000001, 0x0, 0x0, 0x0 }).edx >> 20);
        }

      executed_previously = true;
    }

  return ret;
}

bool
cpuid_feature_msr ()
{
  return cpuid_getfeatures ().edx >> 5 & 1;
}
