/*
idt_routine_common.cpp - Functions for dispatching interrupts for i386
architecture

This file is a part of Fiku.
Copyright (C) 2021 Akib Azmain Turja

Fiku is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Fiku is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

Authors: Akib Azmain Turja
*/

#include <fiku/idt.hpp>
#include <fiku/io.h>
#include <fiku/panic.h>
#include "idt_routines.hpp"

namespace
{
  idt::interrupt_handler handlers[256];
}

namespace idt
{
  void
  set_handler (uint8_t interrupt, interrupt_handler handler)
  {
    handlers[interrupt] = handler;
  }

  interrupt_handler
  get_handler (uint8_t interrupt)
  {
    return handlers[interrupt];
  }
}

extern "C" void
isr_common_ (idt_routine::interrupt_frame *pushed_registers)
{

  // Whether it is a hardware interrupt generated by the PIC
  bool is_irq = (pushed_registers->interrupt >= 0x20
                 && pushed_registers->interrupt < 0x30);

  // Check if the interrupt is spurious
  if (is_irq)
    {

      // Check if it is a spurious interrupt by the master PIC
      if (pushed_registers->interrupt == 0x7 + 0x20)
        {

          // Get In-Service Register value
          io_outb (0xb, 0x20);
          if (!(io_inb (0x20) >> 7 & 1))
            {
              return;
            }
        }

      // Check if it is a spurious interrupt by the slave PIC
      else if (pushed_registers->interrupt == 0x7 + 0x28)
        {

          // Get In-Service Register value
          io_outb (0xb, 0xa0);
          if (!(io_inb (0xa0) >> 7 & 1))
            {

              // Send EOI (end of interrupt) to the master PIC
              io_outb (0x20, 0x20);
              return;
            }
        }
    }

  // Check if a handler is registered for the interrupt
  if (handlers[pushed_registers->interrupt] != nullptr)
    {

      // Setup the structure to pass to the handler
      idt::registers_t registers = {
        static_cast < uint16_t > (pushed_registers->cs),
        static_cast < uint16_t > (pushed_registers->ds),
        pushed_registers->eip,
        pushed_registers->eflags,
        pushed_registers->eax,
        pushed_registers->ebx,
        pushed_registers->ecx,
        pushed_registers->edx,
        0,
        pushed_registers->ebp,
        pushed_registers->esi,
        pushed_registers->edi,
        false
      };

      // Set esp according to cs
      if ((pushed_registers->cs & 0x3) == 0x0)
        {
          registers.esp = (reinterpret_cast < uint32_t >
                           (&pushed_registers->esp));
        }
      else
        {
          registers.esp = pushed_registers->esp;
        }

      // Call the handler
      handlers[pushed_registers->interrupt] (pushed_registers->interrupt,
                                             pushed_registers->code,
                                             registers);

      // Update registers
#define UPDATE(x) pushed_registers->x = registers.x
      UPDATE (cs);
      UPDATE (ds);
      UPDATE (eip);
      UPDATE (eflags);
      UPDATE (eax);
      UPDATE (ebx);
      UPDATE (ecx);
      UPDATE (edx);
      UPDATE (ebp);
      UPDATE (esi);
      UPDATE (edi);

      // Update stack segment and %esp if we're going to return to lower
      // ring level
      if (pushed_registers->cs & 0x3)
       {
         UPDATE (esp);
         pushed_registers->ss = pushed_registers->ds;
       }
#undef UPDATE

      // Send EOI signals to PICs if required
      if (is_irq && !registers.eoi_sent)
        {

          // If the IRQ is from the slave PIC, send EOI to it
          if (pushed_registers->interrupt >= 0x28)
            {
              io_outb (0x20, 0xa0);
            }

          // In any case, send EOI to the master PIC
          io_outb (0x20, 0x20);
        }
    }

  // Otherwise, show "unhandled interrupt" and halt
  else
    {

      // Show unhandled interrupt error and panic
      panic ("unhandled interrupt: %#lx", pushed_registers->interrupt);
    }
}
