/*
start.s - Instructions for starting the kernel for i386 architecture

This file is a part of Fiku.
Copyright (C) 2021 Akib Azmain Turja

Fiku is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Fiku is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

Authors: Akib Azmain Turja
*/

  .section .data
  .align 0x1000
  .type init_page_s_, @object
init_page_s_:

  // Four PDEs, hopefully the kernel will fit in 16 MiB (or atleast the
  // required parts to handle if any problem occurs)
  .long (0x0000 + 0x1000) | 0x3
  .long (0x1000 + 0x1000) | 0x3
  .long (0x2000 + 0x1000) | 0x3
  .long (0x3000 + 0x1000) | 0x3
  // NOTE: Add more PDEs here

  // Empty page directory entries
  .zero (960 - 4) * 4
  //           |    |- each PDE is 4 bytes
  //           |
  // 4 of PDEs are defined above, increase if added more

  // Four more PDEs to map the kernel at 0xf0000000
  .long (0x0000 + 0x1000) | 0x3
  .long (0x1000 + 0x1000) | 0x3
  .long (0x2000 + 0x1000) | 0x3
  .long (0x3000 + 0x1000) | 0x3

  // Empty page directory entries
  .zero (1024 - (960 + 4)) * 4

  .macro fpte64 base, i=0

  // Fill 64 PTEs (we can't fill more than 100 by recursion, it is the limit)
  .ifne 64 - \i
  .long \base | 0x3
  fpte64 "(\base + 0x1000)", "(\i + 1)"
  .endif
  .endm

  .macro fpte base

  // Fill 64 PTEs 16 times, filling 1024 (64 * 16) PTEs
  fpte64 "(\base + 0x40000 * 0)"
  fpte64 "(\base + 0x40000 * 1)"
  fpte64 "(\base + 0x40000 * 2)"
  fpte64 "(\base + 0x40000 * 3)"
  fpte64 "(\base + 0x40000 * 4)"
  fpte64 "(\base + 0x40000 * 5)"
  fpte64 "(\base + 0x40000 * 6)"
  fpte64 "(\base + 0x40000 * 7)"
  fpte64 "(\base + 0x40000 * 8)"
  fpte64 "(\base + 0x40000 * 9)"
  fpte64 "(\base + 0x40000 * 10)"
  fpte64 "(\base + 0x40000 * 11)"
  fpte64 "(\base + 0x40000 * 12)"
  fpte64 "(\base + 0x40000 * 13)"
  fpte64 "(\base + 0x40000 * 14)"
  fpte64 "(\base + 0x40000 * 15)"
  //       |- base   |       |
  //                 |       |- page table index
  // Every 64 PTEs map 0x40000 (4096 * 64 or 0x1000 * 0x40) bytes
  .endm

  // Fill four page tables
  fpte (0x400000 * 0)
  fpte (0x400000 * 1)
  fpte (0x400000 * 2)
  fpte (0x400000 * 3)
  //      |        |      NOTE: Add more page tables above this line
  //      |        |- page table index
  // Each page table maps 0x400000 (4096 * 1024 or 0x1000 * 0x400) bytes
  .size init_page_s_, . - init_page_s_

  .section .text
  .global _start
  .type _start, @function

  // Execution starts from here
_start:

  // First stop interrupts, or they may annoy us
  cli

  // The PDEs contains the offset of PTEs from "init_page_s__", make them
  // physical
  addl $(init_page_s_ - 0xf0000000), init_page_s_ - 0xf0000000 + (0 * 4)
  addl $(init_page_s_ - 0xf0000000), init_page_s_ - 0xf0000000 + (1 * 4)
  addl $(init_page_s_ - 0xf0000000), init_page_s_ - 0xf0000000 + (2 * 4)
  addl $(init_page_s_ - 0xf0000000), init_page_s_ - 0xf0000000 + (3 * 4)
  addl $(init_page_s_ - 0xf0000000), init_page_s_ - 0xf0000000 + (960 * 4)
  addl $(init_page_s_ - 0xf0000000), init_page_s_ - 0xf0000000 + (961 * 4)
  addl $(init_page_s_ - 0xf0000000), init_page_s_ - 0xf0000000 + (962 * 4)
  addl $(init_page_s_ - 0xf0000000), init_page_s_ - 0xf0000000 + (963 * 4)

  // Load page directory
  movl $(init_page_s_ - 0xf0000000), %ecx
  movl %ecx, %cr3

  // Set the PG bit in %cr0 and enable paging
  movl %cr0, %ecx
  orl $0x80000000, %ecx
  movl %ecx, %cr0

  // Jump to higher half!
  leal 1f, %ecx
  jmp *%ecx
1:

  // Setup the stack
  movl $stack__, %esp
  movl %esp, %ebp

  // Set eflags to something that we know
  pushl $0x2
  popf

  // Clear BSS
  movl $bss_begin_, %edi
  movl $bss_end_, %ecx
  subl %edi, %ecx
  movl %eax, %edx // Backup Multiboot2 boot magic
  xorl %eax, %eax
  rep stosb
  movl %edx, %eax // Restore Multiboot2 boot magic

  // Check if the kernel is booted with Multiboot2 compliant bootloader
  cmpl $0x36d76289, %eax
  jne 2f

  // Convert to virtual and save %ebx that contains Multiboot2 boot info
  addl $0xf0000000, %ebx
  movl %ebx, multiboot_boot_info_ptr_

  // Call global constructors
  call _init

  // Call high level C++ code
  call main

  // Call global destructors
  call _fini

  // Execution should never come to this point, and if it ever comes,
  // freeze the CPU
  jmp cpu_freeze

2:
  call invalid_eax_error_
