/*
idt.cpp - IDT related functions for i386 architecture

This file is a part of Fiku.
Copyright (C) 2021 Akib Azmain Turja

Fiku is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Fiku is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

Authors: Akib Azmain Turja
*/

#include <fiku/idt.hpp>
#include <fiku/io.h>
#include <fiku/vga.h>
#include <stdio.h>

namespace
{
  struct pointer
  {

    // Size of IDT
    uint16_t limit;

    // Base address of IDT
    uint32_t base;
  }
  __attribute__((packed));

  pointer idt_ptr;

  void
  debug_break (uint8_t, uint8_t, idt::registers_t &registers)
  {
    printf (("Interrupt 3 received from %#lx. Breakpoint. Waiting for"
             " debugger...\n"),
            registers.eip);
    vga_flush ();

    volatile bool continue_execution = false;
    while (!continue_execution);

    printf ("Continuing...\n");
    vga_flush ();
  }
}

extern "C" void
idt_flush__ (void *idt_ptr);

#define DEF_ISR(n) extern "C"                   \
    void isr_##n##__()

DEF_ISR (0x00);
DEF_ISR (0x01);
DEF_ISR (0x02);
DEF_ISR (0x03);
DEF_ISR (0x04);
DEF_ISR (0x05);
DEF_ISR (0x06);
DEF_ISR (0x07);
DEF_ISR (0x08);
DEF_ISR (0x09);
DEF_ISR (0x0a);
DEF_ISR (0x0b);
DEF_ISR (0x0c);
DEF_ISR (0x0d);
DEF_ISR (0x0e);
DEF_ISR (0x0f);
DEF_ISR (0x10);
DEF_ISR (0x11);
DEF_ISR (0x12);
DEF_ISR (0x13);
DEF_ISR (0x14);
DEF_ISR (0x15);
DEF_ISR (0x16);
DEF_ISR (0x17);
DEF_ISR (0x18);
DEF_ISR (0x19);
DEF_ISR (0x1a);
DEF_ISR (0x1b);
DEF_ISR (0x1c);
DEF_ISR (0x1d);
DEF_ISR (0x1e);
DEF_ISR (0x1f);
DEF_ISR (0x20);
DEF_ISR (0x21);
DEF_ISR (0x22);
DEF_ISR (0x23);
DEF_ISR (0x24);
DEF_ISR (0x25);
DEF_ISR (0x26);
DEF_ISR (0x27);
DEF_ISR (0x28);
DEF_ISR (0x29);
DEF_ISR (0x2a);
DEF_ISR (0x2b);
DEF_ISR (0x2c);
DEF_ISR (0x2d);
DEF_ISR (0x2e);
DEF_ISR (0x2f);
DEF_ISR (0x30);
DEF_ISR (0x31);
DEF_ISR (0x32);
DEF_ISR (0x33);
DEF_ISR (0x34);
DEF_ISR (0x35);
DEF_ISR (0x36);
DEF_ISR (0x37);
DEF_ISR (0x38);
DEF_ISR (0x39);
DEF_ISR (0x3a);
DEF_ISR (0x3b);
DEF_ISR (0x3c);
DEF_ISR (0x3d);
DEF_ISR (0x3e);
DEF_ISR (0x3f);
DEF_ISR (0x40);
DEF_ISR (0x41);
DEF_ISR (0x42);
DEF_ISR (0x43);
DEF_ISR (0x44);
DEF_ISR (0x45);
DEF_ISR (0x46);
DEF_ISR (0x47);
DEF_ISR (0x48);
DEF_ISR (0x49);
DEF_ISR (0x4a);
DEF_ISR (0x4b);
DEF_ISR (0x4c);
DEF_ISR (0x4d);
DEF_ISR (0x4e);
DEF_ISR (0x4f);
DEF_ISR (0x50);
DEF_ISR (0x51);
DEF_ISR (0x52);
DEF_ISR (0x53);
DEF_ISR (0x54);
DEF_ISR (0x55);
DEF_ISR (0x56);
DEF_ISR (0x57);
DEF_ISR (0x58);
DEF_ISR (0x59);
DEF_ISR (0x5a);
DEF_ISR (0x5b);
DEF_ISR (0x5c);
DEF_ISR (0x5d);
DEF_ISR (0x5e);
DEF_ISR (0x5f);
DEF_ISR (0x60);
DEF_ISR (0x61);
DEF_ISR (0x62);
DEF_ISR (0x63);
DEF_ISR (0x64);
DEF_ISR (0x65);
DEF_ISR (0x66);
DEF_ISR (0x67);
DEF_ISR (0x68);
DEF_ISR (0x69);
DEF_ISR (0x6a);
DEF_ISR (0x6b);
DEF_ISR (0x6c);
DEF_ISR (0x6d);
DEF_ISR (0x6e);
DEF_ISR (0x6f);
DEF_ISR (0x70);
DEF_ISR (0x71);
DEF_ISR (0x72);
DEF_ISR (0x73);
DEF_ISR (0x74);
DEF_ISR (0x75);
DEF_ISR (0x76);
DEF_ISR (0x77);
DEF_ISR (0x78);
DEF_ISR (0x79);
DEF_ISR (0x7a);
DEF_ISR (0x7b);
DEF_ISR (0x7c);
DEF_ISR (0x7d);
DEF_ISR (0x7e);
DEF_ISR (0x7f);
DEF_ISR (0x80);
DEF_ISR (0x81);
DEF_ISR (0x82);
DEF_ISR (0x83);
DEF_ISR (0x84);
DEF_ISR (0x85);
DEF_ISR (0x86);
DEF_ISR (0x87);
DEF_ISR (0x88);
DEF_ISR (0x89);
DEF_ISR (0x8a);
DEF_ISR (0x8b);
DEF_ISR (0x8c);
DEF_ISR (0x8d);
DEF_ISR (0x8e);
DEF_ISR (0x8f);
DEF_ISR (0x90);
DEF_ISR (0x91);
DEF_ISR (0x92);
DEF_ISR (0x93);
DEF_ISR (0x94);
DEF_ISR (0x95);
DEF_ISR (0x96);
DEF_ISR (0x97);
DEF_ISR (0x98);
DEF_ISR (0x99);
DEF_ISR (0x9a);
DEF_ISR (0x9b);
DEF_ISR (0x9c);
DEF_ISR (0x9d);
DEF_ISR (0x9e);
DEF_ISR (0x9f);
DEF_ISR (0xa0);
DEF_ISR (0xa1);
DEF_ISR (0xa2);
DEF_ISR (0xa3);
DEF_ISR (0xa4);
DEF_ISR (0xa5);
DEF_ISR (0xa6);
DEF_ISR (0xa7);
DEF_ISR (0xa8);
DEF_ISR (0xa9);
DEF_ISR (0xaa);
DEF_ISR (0xab);
DEF_ISR (0xac);
DEF_ISR (0xad);
DEF_ISR (0xae);
DEF_ISR (0xaf);
DEF_ISR (0xb0);
DEF_ISR (0xb1);
DEF_ISR (0xb2);
DEF_ISR (0xb3);
DEF_ISR (0xb4);
DEF_ISR (0xb5);
DEF_ISR (0xb6);
DEF_ISR (0xb7);
DEF_ISR (0xb8);
DEF_ISR (0xb9);
DEF_ISR (0xba);
DEF_ISR (0xbb);
DEF_ISR (0xbc);
DEF_ISR (0xbd);
DEF_ISR (0xbe);
DEF_ISR (0xbf);
DEF_ISR (0xc0);
DEF_ISR (0xc1);
DEF_ISR (0xc2);
DEF_ISR (0xc3);
DEF_ISR (0xc4);
DEF_ISR (0xc5);
DEF_ISR (0xc6);
DEF_ISR (0xc7);
DEF_ISR (0xc8);
DEF_ISR (0xc9);
DEF_ISR (0xca);
DEF_ISR (0xcb);
DEF_ISR (0xcc);
DEF_ISR (0xcd);
DEF_ISR (0xce);
DEF_ISR (0xcf);
DEF_ISR (0xd0);
DEF_ISR (0xd1);
DEF_ISR (0xd2);
DEF_ISR (0xd3);
DEF_ISR (0xd4);
DEF_ISR (0xd5);
DEF_ISR (0xd6);
DEF_ISR (0xd7);
DEF_ISR (0xd8);
DEF_ISR (0xd9);
DEF_ISR (0xda);
DEF_ISR (0xdb);
DEF_ISR (0xdc);
DEF_ISR (0xdd);
DEF_ISR (0xde);
DEF_ISR (0xdf);
DEF_ISR (0xe0);
DEF_ISR (0xe1);
DEF_ISR (0xe2);
DEF_ISR (0xe3);
DEF_ISR (0xe4);
DEF_ISR (0xe5);
DEF_ISR (0xe6);
DEF_ISR (0xe7);
DEF_ISR (0xe8);
DEF_ISR (0xe9);
DEF_ISR (0xea);
DEF_ISR (0xeb);
DEF_ISR (0xec);
DEF_ISR (0xed);
DEF_ISR (0xee);
DEF_ISR (0xef);
DEF_ISR (0xf0);
DEF_ISR (0xf1);
DEF_ISR (0xf2);
DEF_ISR (0xf3);
DEF_ISR (0xf4);
DEF_ISR (0xf5);
DEF_ISR (0xf6);
DEF_ISR (0xf7);
DEF_ISR (0xf8);
DEF_ISR (0xf9);
DEF_ISR (0xfa);
DEF_ISR (0xfb);
DEF_ISR (0xfc);
DEF_ISR (0xfd);
DEF_ISR (0xfe);
DEF_ISR (0xff);

#undef DEF_ISR

namespace idt
{
  void
  flush (const gate * table, size_t count)
  {
    idt_ptr = {
      static_cast < uint16_t > (sizeof (gate) * count - 1),
      reinterpret_cast < uint32_t > (table)
    };

    idt_flush__ (&idt_ptr);
  }

  gate *
  init (bool flush, bool pic)
  {

    // Remap the PIC
    if (pic)
      {
        io_outb (0x11, 0x20);
        io_outb (0x11, 0xa0);
        io_outb (0x20, 0x21);
        io_outb (0x28, 0xa1);
        io_outb (0x4, 0x21);
        io_outb (0x2, 0xa1);
        io_outb (0x1, 0x21);
        io_outb (0x1, 0xa1);

        // Disable the PIC
        io_outb (0xff, 0x21);
        io_outb (0xff, 0xa1);
      }

    static gate idt[256];

#define SET_GATE(n, pl) idt[n] =                                          \
      gate (reinterpret_cast < uint32_t > (isr_##n##__), 0x8, 0b1110, pl)

    SET_GATE (0x00, 0x0);
    SET_GATE (0x01, 0x0);
    SET_GATE (0x02, 0x0);
    SET_GATE (0x03, 0x3);
    SET_GATE (0x04, 0x0);
    SET_GATE (0x05, 0x0);
    SET_GATE (0x06, 0x0);
    SET_GATE (0x07, 0x0);
    SET_GATE (0x08, 0x0);
    SET_GATE (0x09, 0x0);
    SET_GATE (0x0a, 0x0);
    SET_GATE (0x0b, 0x0);
    SET_GATE (0x0c, 0x0);
    SET_GATE (0x0d, 0x0);
    SET_GATE (0x0e, 0x0);
    SET_GATE (0x0f, 0x0);
    SET_GATE (0x10, 0x0);
    SET_GATE (0x11, 0x0);
    SET_GATE (0x12, 0x0);
    SET_GATE (0x13, 0x0);
    SET_GATE (0x14, 0x0);
    SET_GATE (0x15, 0x0);
    SET_GATE (0x16, 0x0);
    SET_GATE (0x17, 0x0);
    SET_GATE (0x18, 0x0);
    SET_GATE (0x19, 0x0);
    SET_GATE (0x1a, 0x0);
    SET_GATE (0x1b, 0x0);
    SET_GATE (0x1c, 0x0);
    SET_GATE (0x1d, 0x0);
    SET_GATE (0x1e, 0x0);
    SET_GATE (0x1f, 0x0);
    SET_GATE (0x20, 0x0);
    SET_GATE (0x21, 0x0);
    SET_GATE (0x22, 0x0);
    SET_GATE (0x23, 0x0);
    SET_GATE (0x24, 0x0);
    SET_GATE (0x25, 0x0);
    SET_GATE (0x26, 0x0);
    SET_GATE (0x27, 0x0);
    SET_GATE (0x28, 0x0);
    SET_GATE (0x29, 0x0);
    SET_GATE (0x2a, 0x0);
    SET_GATE (0x2b, 0x0);
    SET_GATE (0x2c, 0x0);
    SET_GATE (0x2d, 0x0);
    SET_GATE (0x2e, 0x0);
    SET_GATE (0x2f, 0x0);
    SET_GATE (0x30, 0x0);
    SET_GATE (0x31, 0x0);
    SET_GATE (0x32, 0x0);
    SET_GATE (0x33, 0x0);
    SET_GATE (0x34, 0x0);
    SET_GATE (0x35, 0x0);
    SET_GATE (0x36, 0x0);
    SET_GATE (0x37, 0x0);
    SET_GATE (0x38, 0x0);
    SET_GATE (0x39, 0x0);
    SET_GATE (0x3a, 0x0);
    SET_GATE (0x3b, 0x0);
    SET_GATE (0x3c, 0x0);
    SET_GATE (0x3d, 0x0);
    SET_GATE (0x3e, 0x0);
    SET_GATE (0x3f, 0x0);
    SET_GATE (0x40, 0x0);
    SET_GATE (0x41, 0x0);
    SET_GATE (0x42, 0x0);
    SET_GATE (0x43, 0x0);
    SET_GATE (0x44, 0x0);
    SET_GATE (0x45, 0x0);
    SET_GATE (0x46, 0x0);
    SET_GATE (0x47, 0x0);
    SET_GATE (0x48, 0x0);
    SET_GATE (0x49, 0x0);
    SET_GATE (0x4a, 0x0);
    SET_GATE (0x4b, 0x0);
    SET_GATE (0x4c, 0x0);
    SET_GATE (0x4d, 0x0);
    SET_GATE (0x4e, 0x0);
    SET_GATE (0x4f, 0x0);
    SET_GATE (0x50, 0x0);
    SET_GATE (0x51, 0x0);
    SET_GATE (0x52, 0x0);
    SET_GATE (0x53, 0x0);
    SET_GATE (0x54, 0x0);
    SET_GATE (0x55, 0x0);
    SET_GATE (0x56, 0x0);
    SET_GATE (0x57, 0x0);
    SET_GATE (0x58, 0x0);
    SET_GATE (0x59, 0x0);
    SET_GATE (0x5a, 0x0);
    SET_GATE (0x5b, 0x0);
    SET_GATE (0x5c, 0x0);
    SET_GATE (0x5d, 0x0);
    SET_GATE (0x5e, 0x0);
    SET_GATE (0x5f, 0x0);
    SET_GATE (0x60, 0x0);
    SET_GATE (0x61, 0x0);
    SET_GATE (0x62, 0x0);
    SET_GATE (0x63, 0x0);
    SET_GATE (0x64, 0x0);
    SET_GATE (0x65, 0x0);
    SET_GATE (0x66, 0x0);
    SET_GATE (0x67, 0x0);
    SET_GATE (0x68, 0x0);
    SET_GATE (0x69, 0x0);
    SET_GATE (0x6a, 0x0);
    SET_GATE (0x6b, 0x0);
    SET_GATE (0x6c, 0x0);
    SET_GATE (0x6d, 0x0);
    SET_GATE (0x6e, 0x0);
    SET_GATE (0x6f, 0x0);
    SET_GATE (0x70, 0x0);
    SET_GATE (0x71, 0x0);
    SET_GATE (0x72, 0x0);
    SET_GATE (0x73, 0x0);
    SET_GATE (0x74, 0x0);
    SET_GATE (0x75, 0x0);
    SET_GATE (0x76, 0x0);
    SET_GATE (0x77, 0x0);
    SET_GATE (0x78, 0x0);
    SET_GATE (0x79, 0x0);
    SET_GATE (0x7a, 0x0);
    SET_GATE (0x7b, 0x0);
    SET_GATE (0x7c, 0x0);
    SET_GATE (0x7d, 0x0);
    SET_GATE (0x7e, 0x0);
    SET_GATE (0x7f, 0x0);
    SET_GATE (0x80, 0x0);
    SET_GATE (0x81, 0x0);
    SET_GATE (0x82, 0x0);
    SET_GATE (0x83, 0x0);
    SET_GATE (0x84, 0x0);
    SET_GATE (0x85, 0x0);
    SET_GATE (0x86, 0x0);
    SET_GATE (0x87, 0x0);
    SET_GATE (0x88, 0x0);
    SET_GATE (0x89, 0x0);
    SET_GATE (0x8a, 0x0);
    SET_GATE (0x8b, 0x0);
    SET_GATE (0x8c, 0x0);
    SET_GATE (0x8d, 0x0);
    SET_GATE (0x8e, 0x0);
    SET_GATE (0x8f, 0x0);
    SET_GATE (0x90, 0x0);
    SET_GATE (0x91, 0x0);
    SET_GATE (0x92, 0x0);
    SET_GATE (0x93, 0x0);
    SET_GATE (0x94, 0x0);
    SET_GATE (0x95, 0x0);
    SET_GATE (0x96, 0x0);
    SET_GATE (0x97, 0x0);
    SET_GATE (0x98, 0x0);
    SET_GATE (0x99, 0x0);
    SET_GATE (0x9a, 0x0);
    SET_GATE (0x9b, 0x0);
    SET_GATE (0x9c, 0x0);
    SET_GATE (0x9d, 0x0);
    SET_GATE (0x9e, 0x0);
    SET_GATE (0x9f, 0x0);
    SET_GATE (0xa0, 0x0);
    SET_GATE (0xa1, 0x0);
    SET_GATE (0xa2, 0x0);
    SET_GATE (0xa3, 0x0);
    SET_GATE (0xa4, 0x0);
    SET_GATE (0xa5, 0x0);
    SET_GATE (0xa6, 0x0);
    SET_GATE (0xa7, 0x0);
    SET_GATE (0xa8, 0x0);
    SET_GATE (0xa9, 0x0);
    SET_GATE (0xaa, 0x0);
    SET_GATE (0xab, 0x0);
    SET_GATE (0xac, 0x0);
    SET_GATE (0xad, 0x0);
    SET_GATE (0xae, 0x0);
    SET_GATE (0xaf, 0x0);
    SET_GATE (0xb0, 0x0);
    SET_GATE (0xb1, 0x0);
    SET_GATE (0xb2, 0x0);
    SET_GATE (0xb3, 0x0);
    SET_GATE (0xb4, 0x0);
    SET_GATE (0xb5, 0x0);
    SET_GATE (0xb6, 0x0);
    SET_GATE (0xb7, 0x0);
    SET_GATE (0xb8, 0x0);
    SET_GATE (0xb9, 0x0);
    SET_GATE (0xba, 0x0);
    SET_GATE (0xbb, 0x0);
    SET_GATE (0xbc, 0x0);
    SET_GATE (0xbd, 0x0);
    SET_GATE (0xbe, 0x0);
    SET_GATE (0xbf, 0x0);
    SET_GATE (0xc0, 0x0);
    SET_GATE (0xc1, 0x0);
    SET_GATE (0xc2, 0x0);
    SET_GATE (0xc3, 0x0);
    SET_GATE (0xc4, 0x0);
    SET_GATE (0xc5, 0x0);
    SET_GATE (0xc6, 0x0);
    SET_GATE (0xc7, 0x0);
    SET_GATE (0xc8, 0x0);
    SET_GATE (0xc9, 0x0);
    SET_GATE (0xca, 0x0);
    SET_GATE (0xcb, 0x0);
    SET_GATE (0xcc, 0x0);
    SET_GATE (0xcd, 0x0);
    SET_GATE (0xce, 0x0);
    SET_GATE (0xcf, 0x0);
    SET_GATE (0xd0, 0x0);
    SET_GATE (0xd1, 0x0);
    SET_GATE (0xd2, 0x0);
    SET_GATE (0xd3, 0x0);
    SET_GATE (0xd4, 0x0);
    SET_GATE (0xd5, 0x0);
    SET_GATE (0xd6, 0x0);
    SET_GATE (0xd7, 0x0);
    SET_GATE (0xd8, 0x0);
    SET_GATE (0xd9, 0x0);
    SET_GATE (0xda, 0x0);
    SET_GATE (0xdb, 0x0);
    SET_GATE (0xdc, 0x0);
    SET_GATE (0xdd, 0x0);
    SET_GATE (0xde, 0x0);
    SET_GATE (0xdf, 0x0);
    SET_GATE (0xe0, 0x0);
    SET_GATE (0xe1, 0x0);
    SET_GATE (0xe2, 0x0);
    SET_GATE (0xe3, 0x0);
    SET_GATE (0xe4, 0x0);
    SET_GATE (0xe5, 0x0);
    SET_GATE (0xe6, 0x0);
    SET_GATE (0xe7, 0x0);
    SET_GATE (0xe8, 0x0);
    SET_GATE (0xe9, 0x0);
    SET_GATE (0xea, 0x0);
    SET_GATE (0xeb, 0x0);
    SET_GATE (0xec, 0x0);
    SET_GATE (0xed, 0x0);
    SET_GATE (0xee, 0x0);
    SET_GATE (0xef, 0x0);
    SET_GATE (0xf0, 0x0);
    SET_GATE (0xf1, 0x0);
    SET_GATE (0xf2, 0x0);
    SET_GATE (0xf3, 0x0);
    SET_GATE (0xf4, 0x0);
    SET_GATE (0xf5, 0x0);
    SET_GATE (0xf6, 0x0);
    SET_GATE (0xf7, 0x0);
    SET_GATE (0xf8, 0x0);
    SET_GATE (0xf9, 0x0);
    SET_GATE (0xfa, 0x0);
    SET_GATE (0xfb, 0x0);
    SET_GATE (0xfc, 0x0);
    SET_GATE (0xfd, 0x0);
    SET_GATE (0xfe, 0x0);
    SET_GATE (0xff, 0x0);

#undef SET_GATE

    // Flush the IDT
    if (flush)
      {
        idt::flush (idt, 256);
      }

    set_handler (0x3, debug_break);
    return idt;
  }
}
