/*
paging.cpp - Paging related function implementation for i386 architecture

This file is a part of Fiku.
Copyright (C) 2021 Akib Azmain Turja

Fiku is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Fiku is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

Authors: Akib Azmain Turja
*/

#include <fiku/paging.hpp>
#include <config.h>
#include <stddef.h>
#include <stdio.h>
#include <fiku/panic.h>
#include <fiku/vga.h>
#include <fiku/idt.hpp>
#include <fiku/multiboot2.hpp>
#include <fiku/cpuid.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <fiku/spinlock.h>
#include <fiku/cpu.h>

extern "C"
{
  char text_begin_[0];
  char text_end_[0];
  char init_begin_[0];
  char init_end_[0];
  char fini_begin_[0];
  char fini_end_[0];
  char rodata_begin_[0];
  char rodata_end_[0];
  char ctors_begin_[0];
  char ctors_end_[0];
  char dtors_begin_[0];
  char dtors_end_[0];
  char data_begin_[0];
  char data_end_[0];
  char bss_begin_[0];
  char bss_end_[0];

  // Writes to %cr3
  void
  paging_flush_ (size_t directory);

  // Evaluates "%cr0 | mask"
  void
  paging_enable_bits_cr0_ (uint32_t mask);

  // Evaluates "%cr4 | mask"
  void
  paging_enable_bits_cr4_ (uint32_t mask);

  // Enables PAE paging
  void
  paging_enable_pae_ (uint32_t cr3);
}

#ifdef PAE_SUPPORT
#define PT_TOTAL_SIZE 0x80000
#define PDPT_TOTAL_SIZE 0x20
#else
#define PT_TOTAL_SIZE 0x40000
#define PDPT_TOTAL_SIZE 0x0
#endif // PAE_SUPPORT
#define PD_TOTAL_SIZE 0x1000

// Page tables (65536 required for the kernel in both PAE and non-PAE)
char pt_[PT_TOTAL_SIZE] __attribute__ ((aligned (0x1000)));

// Page directory (only 1 required for the kernel in both PAE and non-PAE)
char pd_[PD_TOTAL_SIZE] __attribute__ ((aligned (0x1000)));

// Page directory pointer table
char pdpt_[PDPT_TOTAL_SIZE] __attribute__ ((aligned (0x20)));

namespace
{

  // Holds allocation data for superblocks
  struct superblock_table_t
  {

    // Entry count
    size_t count;

    // Entries (each holds the count of blocks with free blocks)
    uint16_t *entries;

    // Entry coverage (the amount of memory covered by a entry)
    size_t entry_coverage;
  };

  // Page allocation data
  struct phys_alloc_data_t
  {

    // Page count
    size_t count;

    // Page allocation and reference data
    uint16_t *data;

    // Stack for caching free pages
    size_t *stack;

    // Current stack pointer
    size_t *stack_ptr;

    // Stack size
    size_t stack_size;

    // Table count
    size_t table_count;

    // Allocation data
    superblock_table_t *tables;
  };

  // Data about physical block allocation
  phys_alloc_data_t phys_alloc_data;

  // Returns reference to the reference count of the page
  auto
  get_page (size_t page) -> decltype (phys_alloc_data.data[page]) &
  {
    return phys_alloc_data.data[page];
  }

  // Checks whether the page is used
  size_t
  check_page (size_t page)
  {

    // Make sure the page is in range
    if (page >= phys_alloc_data.count)
      {
        errno = ERANGE;
        return 0;
      }

    return get_page (page);
  }

  // Sets a page as used
  void
  set_page (size_t page)
  {

    // Make sure the page is in range
    if (page >= phys_alloc_data.count)
      {
        errno = ERANGE;
        return;
      }

    // Make sure page is available
    if (check_page (page) == 0)
      {

        // Set
        phys_alloc_data.data[page] = 1;

        // Write to all entries referenced this page
        for (size_t i = phys_alloc_data.table_count; i != 0; --i)
          {
            size_t index = page / (phys_alloc_data.tables
                                   [i - 1].entry_coverage / 0x1000);

            // A page is set to used, decrement free block count
            phys_alloc_data.tables[i - 1].entries[index]--;

            // If this block still references atleast a free block, break
            if (phys_alloc_data.tables[i - 1].entries[index] != 0)
              {
                break;
              }
          }
      }
  }

  // Sets page as free
  void
  unset_page (size_t page)
  {

    // Make sure the page is in range
    if (page >= phys_alloc_data.count)
      {
        errno = ERANGE;
        return;
      }

    // Make sure page is used
    if (check_page (page) == 1)
      {

        // Unset
        phys_alloc_data.data[page] = 0;

        // Write to all entries referenced this page
        for (size_t i = phys_alloc_data.table_count; i != 0; --i)
          {
            size_t index = page / (phys_alloc_data.tables
                                   [i - 1].entry_coverage / 0x1000);

            // A page is set to free, increment free block count
            phys_alloc_data.tables[i - 1].entries[index]++;

            // If this block references any other block as free, break
            if (phys_alloc_data.tables[i - 1].entries[index] > 1)
              {
                break;
              }
          }

        // Push the page address into stack
        if (sizeof (size_t) * (phys_alloc_data.stack
                               - phys_alloc_data.stack_ptr)
            < phys_alloc_data.stack_size)
          {
            *(--phys_alloc_data.stack_ptr) = page;
          }
      }
  }

  size_t
  find_page ()
  {

    // If there are free pages in the stack, pop and return
    if (phys_alloc_data.stack != phys_alloc_data.stack_ptr)
      {
        return *(phys_alloc_data.stack_ptr++);
      }

    size_t offset = 0;

    for (size_t i = 0; i < phys_alloc_data.table_count; ++i)
      {
        for (size_t j = offset / phys_alloc_data.tables[i].entry_coverage;
             j < phys_alloc_data.tables[i].count; ++j)
          {
            if (phys_alloc_data.tables[i].entries[j] == 0)
              {
                offset += phys_alloc_data.tables[i].entry_coverage;
              }
            else
              {
                break;
              }
          }
      }

    bool page_found = false;
    size_t page_addr = 0;

    for (size_t i = offset / 0x1000; i < phys_alloc_data.count; ++i)
      {
        if (check_page (i) == 0)
          {
            page_addr = i;
            page_found = true;
            break;
          }
      }

    if (!page_found)
      {
        errno = EAGAIN;
      }

    return page_addr;
  }

  // Initializes page_alloc_data
  void
  page_alloc_data_init (multiboot2::boot_info &multiboot_info,
                        uint16_t *data, size_t count)
  {

    // Initialize variables
    phys_alloc_data.count = count;
    phys_alloc_data.data = data;
    phys_alloc_data.table_count = 0;
    phys_alloc_data.tables = nullptr;
    phys_alloc_data.stack = nullptr;
    phys_alloc_data.stack_ptr = nullptr;
    phys_alloc_data.stack_size = 0;
    memset (phys_alloc_data.data, 0,
            2048 * sizeof (*phys_alloc_data.data));

    // Set unavailable memory as used
    for (auto &tag : multiboot_info)
      {

        // If the tag holds memory map, process it, otherwise continue
        if (tag.type == 6)
          {
            auto &ctag = tag.cast < multiboot2::boot_info::memmap_t > ();

            uint64_t prev_base = 0;
            uint64_t prev_length = 0;

            // Process all entries
            for (size_t i = 0; i < ctag.entry_count (); ++i)
              {

                // If the indicated memory isn't usable...
                if (ctag[i].type != 1)
                  {

                    // Then set it as used
                    for (size_t offset = 0;
                         offset < ((ctag[i].length + 0xfff) & ~0xfff);
                         ++offset)
                      {
                        set_page ((ctag[i].base_addr + offset) >> 12);

                        // Ignore range error
                        if (errno == ERANGE)
                          {
                            errno = 0;
                            break;
                          }
                      }
                  }

                // Set the unspecified skipped area as used too
                for (size_t offset = ((prev_base + prev_length + 0xfff)
                                      & ~0xfff);
                     offset < ((ctag[i].base_addr + 0xfff) & ~0xfff);
                     ++offset)
                  {
                    set_page (offset >> 12);

                    // Ignore range error
                    if (errno == ERANGE)
                      {
                        errno = 0;
                        break;
                      }
                  }

                prev_base = ctag[i].base_addr;
                prev_length = ctag[i].length;
              }
            break;
          }
      }
  }

  void
  page_alloc_data_reinit (multiboot2::boot_info &multiboot_info)
  {

    // Calculate the page count in all available memory (including holes)
    uint64_t page_count = 0;
    for (auto &tag : multiboot_info)
      {

        // If the tag holds memory map, process it, otherwise continue
        if (tag.type == 6)
          {
            auto &ctag = tag.cast < multiboot2::boot_info::memmap_t > ();

            // Process all entries
            for (size_t i = 0; i < ctag.entry_count (); ++i)
              {

                // If the indicated memory is usable, then increase
                // required bits if required
                if (ctag[i].type == 1)
                  {
                    uint64_t page_count_c = (ctag[i].base_addr +
                                                  ctag[i].length) / 0x1000;

                    if (page_count < page_count_c)
                      {
                        page_count = page_count_c;
                      }
                  }
              }
            break;
          }
      }

    /*
     * NOTE: Don't fill structures (except with zero or other patterns)
     * data before the last dynamic allocation. Doing so might cause
     * corruption.
     */

    phys_alloc_data_t prev_alloc_data = phys_alloc_data;

    // Allocate for new page allocation and reference count data
    void *block_alloc_data =
      aligned_alloc (alignof (decltype (phys_alloc_data.data)),
                     page_count * sizeof (phys_alloc_data.data));

    // Make sure we have the memory
    if (errno == ERANGE)
      {
        panic ("not enough memory on this system");
      }

    memset (block_alloc_data, 0,
            page_count * sizeof (phys_alloc_data.data));

    // Calculate the number of superblock tables required
    uint64_t table_count = 0;
    constexpr uint16_t table_max_entries = 1 << 10;
    for (uint64_t i = page_count; i > table_max_entries;
         (i = (i + (table_max_entries - 1)) / table_max_entries,
          ++table_count));

    // Allocate superblock tables
    auto superblock_tables = static_cast < superblock_table_t *>
      (aligned_alloc (alignof (superblock_table_t),
                      table_count * sizeof (superblock_table_t)));

    // Make sure we have the memory
    if (errno == ERANGE)
      {
        panic ("not enough memory on this system");
      }

    memset (superblock_tables, 0,
            table_count * sizeof (superblock_table_t));

    // Allocate superblock table entries
    for (size_t i = 0, coverage = 0x1000 * table_max_entries;
         i < table_count;
         ++i, coverage *= table_max_entries)
      {
        size_t entry_count = ((page_count + (coverage / 0x1000 - 1))
                              / (coverage/ 0x1000));
        void *entries = aligned_alloc (alignof (uint16_t),
                                       entry_count * sizeof (uint16_t));

        // Make sure we have the memory
        if (errno == ERANGE)
          {
            panic ("not enough memory on this system");
          }

        superblock_tables[i].entry_coverage = coverage;
        superblock_tables[i].entries = static_cast < uint16_t *> (entries);
        superblock_tables[i].count = entry_count;
      }

    // Allocate a stack for finding free page faster
    size_t stack_size = sizeof (size_t) * (page_count / 64);
    size_t *stack = (static_cast < size_t *>
                     (aligned_alloc (alignof (size_t), stack_size)));

    // Setup physical block allocation data
    phys_alloc_data.count = page_count;
    phys_alloc_data.data = (static_cast < decltype (phys_alloc_data.data) >
                            (block_alloc_data));
    phys_alloc_data.table_count = table_count;
    phys_alloc_data.tables = superblock_tables;
    phys_alloc_data.stack = stack + stack_size / sizeof (size_t);
    phys_alloc_data.stack_ptr = stack + stack_size / sizeof (size_t);
    phys_alloc_data.stack_size = stack_size;

    // Set free block / superblock counts in superblock tables
    for (size_t i = 0; i < table_count; ++i)
      {
        for (size_t j = 0; j < phys_alloc_data.tables[i].count; ++j)
          {
            if (i + 1 < table_count)
              {

                // Index in next table of the superblock with the same
                // offset as current one
                size_t next_table_i =
                  ((j * phys_alloc_data.tables[i].entry_coverage)
                   / (phys_alloc_data.tables[i + 1].entry_coverage));

                // Set the count of free blocks / superblocks
                if (phys_alloc_data.tables[i].count - next_table_i < 1024)
                  {
                    phys_alloc_data.tables[i].entries[j] =
                      phys_alloc_data.tables[i].count - next_table_i;
                  }
                else
                  {
                    phys_alloc_data.tables[i].entries[j] =
                      table_max_entries;
                  }
              }
            else
              {

                // Index in the block with the same offset as current one
                size_t page_i =
                  (j * phys_alloc_data.tables[i].entry_coverage) / 0x1000;

                // Set the count of free blocks / superblocks
                if (page_count - page_i < 1024)
                  {
                    phys_alloc_data.tables[i].entries[j] = (page_count
                                                            - page_i);
                  }
                else
                  {
                    phys_alloc_data.tables[i].entries[j] =
                      table_max_entries;
                  }
              }
          }
      }

    // Fill data and superblock tables
    for (uint64_t i = 0; i < prev_alloc_data.count; ++i)
      {
        if (prev_alloc_data.data[i])
          {
            set_page (i);

            if (errno == ERANGE)
              {
                errno = 0;
                break;
              }
          }
      }
  }

  // Indices for allocating tables
  size_t pt_alloc_index = 0;
  size_t pd_alloc_index = 0;
  size_t pdpt_alloc_index = 0;

  // Returns whether to use PAE
  bool
  use_pae ()
  {
    return ((PAE_SUPPORT && !PAE_SUPPORT_RUNTIME)
            || (PAE_SUPPORT_RUNTIME && cpuid_check ()
                && cpuid_feature_pae ()));
  }

  // Returns whether to use PSE
  bool
  use_pse ()
  {
    return cpuid_check () && cpuid_feature_pse ();
  }

  // Checks whether the given page is in the given range
  bool
  in_range (uint32_t begin, uint32_t end, uint32_t addr)
  {

    // If the range is 0, return false immediately ("end" maybe zero to
    // represent 2^32, make sure "end" isn't zero)
    if (end && begin >= end)
      return false;

    // Check
    return addr >= begin && (!end || addr < end);
  }

  // Handles a page fault
  void
  page_fault_handler (uint8_t, uint8_t code, idt::registers_t &registers)
  {
    uint32_t error_address;
    asm ("mov %%cr2, %0"
         : "=r" (error_address));

    const char *const error_msgs[] =
      {
      // If set...
      // I: fault caused by instruction fetch
      // R: fault caused by corrupted paging structure
      // U: fault caused by user mode
      // W: fault caused by a write
      // P: fault not caused by non-present page
      /* IRUWP */
      /* 00000 */ ("invalid read at unmapped %#lx from %#lx by"
                   " supervisor"),
      /* 00001 */ ("invalid read at %#lx from %#lx by supervisor"),
      /* 00010 */ ("invalid write at unmapped %#lx from %#lx by"
                   " supervisor"),
      /* 00011 */ ("invalid write at %#lx from %#lx by supervisor"),
      /* 00100 */ ("invalid read at unmapped %#lx from %#lx by user"),
      /* 00101 */ ("invalid read at %#lx from %#lx by user"),
      /* 00110 */ ("invalid write at unmapped %#lx from %#lx by user"),
      /* 00111 */ ("invalid write at %#lx from %#lx by user"),
      /* 01000 */ ("corrupted paging structure, invalid read at unmapped"
                   " %#lx from %#lx by supervisor"),
      /* 01001 */ ("corrupted paging structure, invalid read at %#lx from"
                   " %#lx by supervisor"),
      /* 01110 */ ("corrupted paging structure, invalid write at unmapped"
                   " %#lx from %#lx by supervisor"),
      /* 01011 */ ("corrupted paging structure, invalid write at %#lx from"
                   " %#lx by supervisor"),
      /* 01100 */ ("corrupted paging structure, invalid read at unmapped"
                   " %#lx from %#lx by user"),
      /* 01101 */ ("corrupted paging structure, invalid read at %#lx from"
                   " %#lx by user"),
      /* 01110 */ ("corrupted paging structure, invalid write at unmapped"
                   " %#lx from %#lx by user"),
      /* 01111 */ ("corrupted paging structure, invalid write at %#lx from"
                   " %#lx by user"),
      /* 10000 */ ("invalid execution (read) at unmapped %#lx from %#lx by"
                   " supervisor"),
      /* 10001 */ ("invalid execution (read) at %#lx from %#lx by"
                   " supervisor"),
      /* 10010 */ ("invalid execution (write) at unmapped %#lx from %#lx"
                   " by supervisor"),
      /* 10011 */ ("invalid execution (write) at %#lx from %#lx by"
                   " supervisor"),
      /* 10100 */ ("invalid execution (read) at unmapped %#lx from %#lx"
                   " by user"),
      /* 10101 */ ("invalid execution (read) at %#lx from %#lx by user"),
      /* 10110 */ ("invalid execution (write) at unmapped %#lx from %#lx"
                   " by user"),
      /* 10111 */ ("invalid execution (write) at %#lx from %#lx by user"),
      /* 11000 */ ("corrupted paging structure, invalid execution (read)"
                   " at unmapped %#lx from %#lx by supervisor"),
      /* 11001 */ ("corrupted paging structure, invalid execution (read)"
                   " at %#lx from %#lx by supervisor"),
      /* 11110 */ ("corrupted paging structure, invalid execution (write)"
                   " at unmapped %#lx from %#lx by supervisor"),
      /* 11011 */ ("corrupted paging structure, invalid execution (write)"
                   " at %#lx from %#lx by supervisor"),
      /* 11100 */ ("corrupted paging structure, invalid execution (read)"
                   " at unmapped %#lx from %#lx by user"),
      /* 11101 */ ("corrupted paging structure, invalid execution (read)"
                   " at %#lx from %#lx by user"),
      /* 11110 */ ("corrupted paging structure, invalid execution (write)"
                   " at unmapped %#lx from %#lx by user"),
      /* 11111 */ ("corrupted paging structure, invalid execution (write)"
                   " at %#lx from %#lx by user"),
    };

    printf ("page fault: ");
    printf (error_msgs[code & 0b11111], error_address, registers.eip);
    printf (" (code %d)\n", code);
    vga_flush ();
    cpu_freeze ();
  }

  // Structure holding an special address range
  struct addr_range_t
  {
    uint32_t begin;
    uint32_t end;
  };

  // Returns reference to the lower 32 bit of a paging structure
  uint32_t &
  pg_struct_low (void* page_struct)
  {
    return static_cast < uint32_t *> (page_struct)[0];
  }

  // Returns reference to the higher 32 bit of a paging structure
  uint32_t &
  pg_struct_high (void* page_struct)
  {
    return static_cast < uint32_t *> (page_struct)[1];
  }

  void fill_pt (size_t base, addr_range_t *addrs, size_t size)
  {
    size_t pte_count = use_pae () ? 512 : 1024;
    size_t pte_size = 0x1000 / pte_count;

    for (size_t i = 0; i < pte_count; i++)
      {

        // Do nothing if address is outside of kernel space
        if (in_range (0x0, 0xf0000000, base + i * 0x1000))
          {
            continue;
          }

        auto check = [&] (void *begin, void *end) -> bool
        {
          size_t start = reinterpret_cast < uint32_t > (begin);
          size_t stop = reinterpret_cast < uint32_t > (end) + 0xfff;

          start &= ~0xfff;
          stop &= ~0xfff;

          return in_range (start, stop, base + i * 0x1000);
        };

        auto check_special = [&] () -> bool
        {
          for (size_t j = 0; j < size; j++)
            {
              size_t start = addrs[j].begin;
              size_t stop = addrs[j].end + 0xfff;

              start &= ~0xfff;
              stop &= ~0xfff;

              if (in_range (start, stop, base + i * 0x1000))
                {
                  return true;
                }
            }

          return false;
        };

        if (check (text_begin_, text_end_)
            || check (init_begin_, init_end_)
            || check (fini_begin_, fini_end_))
          {
            paging::pte (pt_ + pt_alloc_index)
              .init (base + i * 0x1000 - 0xf0000000, false, true);

            set_page (((base - 0xf0000000) + i * 0x1000) / 0x1000);
          }
        else if (check (rodata_begin_, rodata_end_)
                 || check (ctors_begin_, ctors_end_)
                 || check (dtors_begin_, dtors_end_))
          {
            paging::pte (pt_ + pt_alloc_index)
              .init (base + i * 0x1000 - 0xf0000000, false, false);

            set_page (((base - 0xf0000000) + i * 0x1000) / 0x1000);
          }
        else if (check (data_begin_, data_end_)
                 || check (bss_begin_, bss_end_)
                 || check_special ())
          {
            paging::pte (pt_ + pt_alloc_index)
              .init (base + i * 0x1000 - 0xf0000000, true, false);

            set_page (((base - 0xf0000000) + i * 0x1000) / 0x1000);
          }
        else
          {
            paging::pte (pt_ + pt_alloc_index).init ();
          }

        pt_alloc_index += pte_size;
      }
  }

  void fill_pd (size_t base, addr_range_t *addrs, size_t size)
  {
    bool pae = use_pae ();
    size_t pde_count = pae ? 512 : 1024;
    size_t pde_size = 0x1000 / pde_count;
    size_t coverage = pae ? 0x200000 : 0x400000;

    for (size_t i = 0; i < pde_count; i++)
      {
        auto check = [&] (void *begin, void *end) -> bool
        {
          size_t start = reinterpret_cast < uint32_t > (begin);
          size_t stop = (reinterpret_cast < uint32_t > (end)
                         + (coverage - 1));

          start &= ~(coverage - 1);
          stop &= ~(coverage - 1);

          return in_range (start, stop, base + i * coverage);
        };

        // Only map if required by kernel
        if (check (reinterpret_cast < void *> (0xf0000000), nullptr))
          {
            paging::pde (pd_ + pd_alloc_index)
              .init (reinterpret_cast < uint32_t > (pt_)
                     + pt_alloc_index - 0xf0000000);

            fill_pt(base + i * coverage, addrs, size);
          }
        else
          {
            paging::pde (pd_ + pd_alloc_index).init ();
          }

        pd_alloc_index += pde_size;
      }
  }

  void fill_pdpt (size_t base, addr_range_t *addrs, size_t size)
  {
    size_t pdpte_size = 8;
    size_t coverage = 0x40000000;

    for (size_t i = 0; i < 4; ++i)
      {
        auto check = [&] (void *begin, void *end) -> bool
        {
          size_t start = reinterpret_cast < uint32_t > (begin);
          size_t stop = (reinterpret_cast < uint32_t > (end)
                         + (coverage - 1));

          start &= ~(coverage - 1);
          stop &= ~(coverage - 1);

          return in_range (start, stop, base + i * coverage);
        };

        // Only map if required by kernel
        if (check (reinterpret_cast < void *> (0xf0000000), nullptr))
          {
            paging::pdpte (pdpt_ + pdpt_alloc_index)
              .init (reinterpret_cast < uint32_t > (pd_)
                     + pd_alloc_index - 0xf0000000);

            fill_pd(base + i * coverage, addrs, size);
          }
        else
          {
            paging::pdpte (pdpt_ + pdpt_alloc_index).init ();
          }

        pdpt_alloc_index += pdpte_size;
      }
  }

  void
  enable (bool paging, bool supervisor_writable, bool pse, bool pae)
  {
    uint32_t bit_mask_cr0 = 0;
    uint32_t bit_mask_cr4 = 0;

    // Set the paging (PG) bit
    bit_mask_cr0 |= paging << 31;

    // Set the WP (Write Protect) bit
    bit_mask_cr0 |= (!supervisor_writable) << 16;

    // Set the PSE bit
    bit_mask_cr4 |= (use_pse () && pse) << 4;

    // Set the PAE bit
    bit_mask_cr4 |= (use_pae () && pae) << 5;

    // Write to CR4
    paging_enable_bits_cr4_ (bit_mask_cr4);

    // Write to CR0
    paging_enable_bits_cr0_ (bit_mask_cr0);
  }

  spnlk_t phymem_lock;
  spnlk_t virmem_lock;

  int
  accuire (spnlk_t *lock)
  {
    cpu_cli ();
    return spnlk_lock (lock);
  }

  int
  release (spnlk_t *lock)
  {
    int ret = spnlk_unlock (lock);
    cpu_sti ();
    return ret;
  }
}

namespace paging
{
  pde_pte_base::pde_pte_base (void *page_struct)
    : page_struct (page_struct)
  {}

  void
  pde_pte_base::init ()
  {
    pg_struct_low (page_struct) = 0;
    if (use_pae ())
      {
        pg_struct_high (page_struct) = 0;
      }
  }

  void
  pde_pte_base::init (uint32_t physaddr, bool writable, bool executable,
                      bool privileged, bool present)
  {
    pg_struct_low (page_struct) = 0;
    pde_pte_base::present (present);
    pde_pte_base::writable (writable);
    pde_pte_base::privileged (privileged);
    if (use_pae ())
      {
        pg_struct_high (page_struct) = 0;
        pde_pte_base::executable (executable);
      }
    address (physaddr);
  }

  uint64_t
  pde_pte_base::address () const
  {
    uint64_t addr;

    // If PAE is enabled, address is 63 bit (page aligned)
    if (use_pae ())
      {
        addr = (*static_cast < uint64_t *> (page_struct) &
                ~(static_cast < uint64_t > (1) << 63));
      }

    // Otherwise address is 32 bit
    else
      {
        addr = pg_struct_low (page_struct);
      }

    // Clear lower 12 bits
    return addr & ~0xfff;
  }

  void
  pde_pte_base::address (uint64_t value)
  {
    value &= ~0xfff;

    if (use_pae ())
      {

        // Clear reseved bits
        for (uint_fast8_t i = 62; i >= cpuid_pae_maxphyaddr (); --i)
          {
            value &= ~(static_cast < uint64_t > (1) << i);
          }

        // Clear previous value
        *static_cast < uint64_t *> (page_struct) &= ~0x7ffff000;

        // Write new value
        *static_cast < uint64_t *> (page_struct) |= value;
      }
    else
      {

        // Clear previous value
        pg_struct_low (page_struct) &= 0xfff;

        // Write new value
        pg_struct_low (page_struct) |= value;
      }
  }

  bool
  pde_pte_base::writable () const
  {
    return pg_struct_low (page_struct) >> 1 & 1;
  }

  void
  pde_pte_base::writable (bool value)
  {

    // Clear previous value
    pg_struct_low (page_struct) &= ~(1 << 1);

    // Write new value
    pg_struct_low (page_struct) |= value << 1;
  }

  bool
  pde_pte_base::executable () const
  {
    return cpuid_feature_nx () && (pg_struct_high (page_struct) >> 31) & 1;
  }

  void
  pde_pte_base::executable (bool value)
  {
    if (cpuid_feature_nx ())
      {

        // Clear previous value
        pg_struct_high (page_struct) &= ~(1 << 31);

        // Write new value
        pg_struct_high (page_struct) |= !value << 31;
      }
  }

  bool
  pde_pte_base::privileged () const
  {
    return !(pg_struct_low (page_struct) >> 2 & 1);
  }

  void
  pde_pte_base::privileged (bool value)
  {

    // Clear previous value
    pg_struct_low (page_struct) &= ~(1 << 2);

    // Write new value
    pg_struct_low (page_struct) |= (!value) << 2;
  }

  bool
  pde_pte_base::accessed () const
  {
    return pg_struct_low (page_struct) >> 5 & 1;
  }

  void
  pde_pte_base::accessed (bool value)
  {

    // Clear previous value
    pg_struct_low (page_struct) &= ~(1 << 5);

    // Write new value
    pg_struct_low (page_struct) |= value << 5;
  }

  bool
  pde_pte_base::written () const
  {
    return pg_struct_low (page_struct) >> 6 & 1;
  }

  void
  pde_pte_base::written (bool value)
  {

    // Clear previous value
    pg_struct_low (page_struct) &= ~(1 << 6);

    // Write new value
    pg_struct_low (page_struct) |= value << 6;
  }

  bool
  pde_pte_base::global () const
  {
    return pg_struct_low (page_struct) >> 8 & 1;
  }

  void
  pde_pte_base::global (bool value)
  {

    // Clear previous value
    pg_struct_low (page_struct) &= ~(1 << 8);

    // Write new value
    pg_struct_low (page_struct) |= value << 8;
  }

  bool
  pde_pte_base::present () const
  {
    return pg_struct_low (page_struct) & 1;
  }

  void
  pde_pte_base::present (bool value)
  {

    // Clear previous value
    pg_struct_low (page_struct) &= ~0x1;

    // Write new value
    pg_struct_low (page_struct) |= value;
  }

  bool
  pde_pte_base::cache () const
  {
    return pg_struct_low (page_struct) >> 4 & 1;
  }

  void
  pde_pte_base::cache (bool value)
  {

    // Clear previous value
    pg_struct_low (page_struct) &= ~(1 << 4);

    // Write new value
    pg_struct_low (page_struct) |= value << 4;
  }

  bool
  pde_pte_base::writet () const
  {
    return pg_struct_low (page_struct) >> 3 & 1;
  }

  void
  pde_pte_base::writet (bool value)
  {

    // Clear previous value
    pg_struct_low (page_struct) &= ~(1 << 3);

    // Write new value
    pg_struct_low (page_struct) |= value << 3;
  }

  pdpte::pdpte (void *page_struct)
    : pde_pte_base (page_struct)
  {}

  void
  pdpte::init ()
  {
    pde_pte_base::init ();
  }

  void
  pdpte::init (uint32_t physaddr, bool present)
  {
    init ();
    pde_pte_base::present (present);
    address (physaddr);
  }

  pde::pde (void *page_struct)
    : pde_pte_base (page_struct)
  {}

  void
  pde::init ()
  {
    pde_pte_base::init ();
  }

  void
  pde::init (uint32_t physaddr, bool writable, bool executable,
             bool privileged, bool size, bool present, bool swapped)
  {
    pde_pte_base::init (0, writable, executable, privileged, present);
    pde::swapped (swapped);
    pde::size (size);
    address (physaddr);
  }

  uint64_t
  pde::address () const
  {
    return pde_pte_base::address ();
  }

  void
  pde::address (uint64_t value)
  {
    if (size ())
      {
        if (use_pae ())
          {
            value &= ~0xfffff;
          }
        else
          {
            value &= ~0x1fffff;
          }
      }

    pde_pte_base::address (value);
  }

  bool
  pde::written () const
  {
    return use_pse () && present () && pde_pte_base::written ();
  }

  void
  pde::written (bool value)
  {
    if (use_pse ())
      {
        pde_pte_base::written (present () && value);
      }
  }

  bool
  pde::swapped () const
  {
    return pg_struct_low (page_struct) >> 11 & 1;
  }

  void
  pde::swapped (bool value)
  {

    // Clear previous value
    pg_struct_low (page_struct) &= ~(1 << 11);

    // Write new value
    pg_struct_low (page_struct) |= value << 11;
  }

  bool
  pde::size () const
  {
    return use_pse () && pg_struct_low (page_struct) >> 7 & 1;
  }

  void
  pde::size (bool value)
  {
    if (use_pse ())
      {

        // Clear previous value
        pg_struct_low (page_struct) &= ~(1 << 7);

        // Write new value
        pg_struct_low (page_struct) |= value << 7;
      }
  }

  pte::pte (void *page_struct)
    : pde_pte_base (page_struct)
  {}

  void
  pte::init ()
  {
    pde_pte_base::init ();
  }

  bool
  pte::swapped () const
  {
    return pg_struct_low (page_struct) >> 11 & 1;
  }

  void
  pte::swapped (bool value)
  {

    // Clear previous value
    pg_struct_low (page_struct) &= ~(1 << 11);

    // Write new value
    pg_struct_low (page_struct) |= value << 11;
  }

  void
  pte::init (uint32_t physaddr, bool writable, bool executable,
             bool privileged, bool present, bool swapped)
  {
    pde_pte_base::init (0, writable, executable, privileged, present);
    pte::swapped (swapped);
    address (physaddr);
  }

#if TARGET_CPU != i386
  void
  invlpg (void *address, bool)
  {
    asm volatile ("invlpg (%0)"
                  :
                  : "r" (address)
                  : "memory");
  }
#else
  void
  invlpg (void *, bool no_reload)
  {

    // Reload only if ``no_reload'' is false
    if (!no_reload)
      {
        asm volatile ("movl %%cr3, %%eax;"
                      "movl %%eax, %%eax"
                      :
                      :
                      : "eax", "memory");
      }
  }
#endif

  void
  flush (size_t structure)
  {
    paging_flush_ (structure);
  }

  void *
  init (const void *boot_info, bool flush)
  {

    // Check is PAE support but PAE support is unavailable
    if (PAE_SUPPORT && !PAE_SUPPORT_RUNTIME && !(cpuid_check () &&
                                                 cpuid_feature_pae ()))
      {
        panic ("error: compiled with PAE support but PAE support is "
               "unavailable on this system, please compile with "
               "\"--enable-pae=auto\"");
      }

    // Initialize spinlocks
    spnlk_init (&phymem_lock);
    spnlk_init (&virmem_lock);

    multiboot2::boot_info multiboot_info (boot_info);

    addr_range_t special_addrs_fixed[] =
      {

        // VGA buffer
        {0xf00b8000, 0xf00b8000 + (25 * 80 * 2)},

        // Multiboot information structure
        {reinterpret_cast < uint32_t > (boot_info),
         (reinterpret_cast < uint32_t > (boot_info)
          + multiboot_info.size ())}
      };

    // How many elements?
    size_t special_addrs_count = sizeof(special_addrs_fixed) /
      sizeof(addr_range_t);

    for (const auto &tag : multiboot_info)
      {
        if (tag.type == 3)
          {
            special_addrs_count++;
          }
      }

    // Special address ranges to be mapped
    addr_range_t special_addrs[special_addrs_count];

    // Fillup the array holding the special address ranges to the mapped
    {
      size_t i = 0;

      for (i = 0; i < sizeof (special_addrs_fixed) / sizeof (addr_range_t);
           i++)
        {
          special_addrs[i] = special_addrs_fixed[i];
        }

      // Write ranges for Multiboot2 modules
      for (const auto& tag : multiboot_info)
        {
          if (tag.type == 3)
            {
              const auto &ctag =
                tag.cast<multiboot2::boot_info::module_t> ();

              special_addrs[i++] =
                {ctag.start + 0xf0000000, ctag.end + 0xf0000000};
            }
        }
    }

    // Set up physical page allocation data
    constexpr size_t block_alloc_page_count = 0x4000;
    uint16_t block_alloc_data[block_alloc_page_count];
    page_alloc_data_init (multiboot_info, block_alloc_data,
                          block_alloc_page_count);

    // Fill the structures recursively
    if (use_pae ())
      {
        fill_pdpt (0, special_addrs, special_addrs_count);
      }
    else
      {
        fill_pd (0, special_addrs, special_addrs_count);
      }

    // Set the fault handler
    idt::set_handler (14, page_fault_handler);

    // Flush if instructed
    if (flush)
      {

        // Flush the structure generated
        if (use_pae ())
          {

            // Enable NX (No eXecute) bit if supported
            if (cpuid_feature_nx ())
              {
                cpu_wrmsr (cpu_rdmsr (0xc0000080) | (1 << 11),
                              0xc0000080);
              }

            auto tmp_fault_handler = [] (uint8_t, uint8_t,
                                         idt::registers_t& registers)
            {

              // Increase %eip
              registers.eip += 0xf0000000;
            };

            // Set page fault handler to change %eip
            idt::set_handler(0xe, tmp_fault_handler);

            // Switch to PAE paging
            paging_enable_pae_ ((reinterpret_cast < uint32_t >
                                 (use_pae () ? pdpt_ : pd_))
                                - 0xf0000000);
          }
        else
          {

            // Enable bits
            enable (true, false, use_pse (), false);

            // Flush
            paging::flush (reinterpret_cast < size_t > (pd_) - 0xf0000000);
          }

        // Set the fault handler
        idt::set_handler (0xe, page_fault_handler);
      }

    // Reinitializes page allocation data with dynamically allocated memory
    page_alloc_data_reinit (multiboot_info);

    // Return the paging structure
    return use_pae () ? pdpt_ : pd_;
  }

  uint64_t
  palloc ()
  {
    accuire (&phymem_lock);
    uint64_t addr = find_page ();

    if (errno == EAGAIN)
      {
        release (&phymem_lock);
        return 0;
      }

    set_page (addr);
    release (&phymem_lock);
    return addr * 0x1000;
  }

  void
  palloc (uint64_t addr)
  {
    accuire (&phymem_lock);

    if (check_page (addr / 0x1000) > 0)
      {
        ++get_page (addr / 0x1000);
      }
    else
      {
        set_page (addr / 0x1000);
      }

    release (&phymem_lock);
  }

  void
  pfree (uint64_t addr)
  {
    accuire (&phymem_lock);

    if (check_page (addr / 0x1000) > 1)
      {
        --get_page (addr / 0x1000);
      }
    else
      {
        unset_page (addr / 0x1000);
      }

    release (&phymem_lock);
  }

  bool
  pused (uint64_t addr)
  {
    return puser (addr);
  }

  size_t
  puser (uint64_t addr)
  {
    accuire (&phymem_lock);
    size_t ret = check_page (addr / 0x1000);
    release (&phymem_lock);
    return ret;
  }

  void *
  kpalloc (size_t count, size_t alignment, bool writable, bool executable)
  {
    if (alignment > 0x1000) // i.e 0x2000, 0x4000...
      {
        alignment >>= 12;
      }
    else
      {
        alignment = 0;
      }

    // Physical addresses to allocated physical blocks
    size_t page_addr[count];

    for (size_t i = 0; i < count; ++i)
      {
        page_addr[i] = palloc ();

        if (errno == EAGAIN)
          {

            // Revert previously set physical blocks
            for (size_t j = 0; j < i; ++j)
              {
                pfree (page_addr[j]);
              }

            break;
          }
      }

    // Make sure no error occurred
    if (errno == 0)
      {
        size_t pte_size = use_pae () ? 8 : 4;
        size_t last_free_pages = 0;
        size_t i = 0;
        accuire (&virmem_lock);

        // Find contiguous `count' free pages
        while (i < sizeof (pt_) / pte_size && last_free_pages < count)
          {
            if (pte (pt_ + i * pte_size).present ())
              {
                last_free_pages = 0;

                // Align if required
                if (alignment)
                  {
                    i = (i + alignment) & ~(alignment - 1);
                    continue;
                  }
              }
            else
              {
                ++last_free_pages;
              }

            ++i;
          }

        // If free pages are found, continue
        if (last_free_pages == count)
          {
            i -= count;

            // Set pages
            for (size_t j = 0; j < count; ++j)
              {
                pte pt_entry = pte (pt_ + (i + j) * pte_size);
                pt_entry.init (page_addr[j], writable, executable);
                pt_entry.global (true);
              }

            void *addr = (reinterpret_cast < void *>
                          (i * 0x1000 + 0xf0000000));

            // Update TLB; if ``invlpg'' isn't supported, reload %cr3
#if TARGET_CPU == i386
            invlpg (nullptr); // No matter what we pass, it is ignored
#else
            for (size_t i = 0; i < count; ++i)
              {
                invlpg (static_cast < char *> (addr) + i * 0x1000);
              }
#endif

            if (errno == EAGAIN)
              {
                release (&virmem_lock);
                return nullptr;
              }

            release (&virmem_lock);
            return addr;
          }
      }

    release (&virmem_lock);
    return nullptr;
  }

  void
  kpmap (void *page, uint64_t addr, bool writable, bool executable)
  {
    pte pt_entry = kpte (page);
    accuire (&virmem_lock);
    pt_entry.init (addr, writable, executable);
    pt_entry.global (true);
    release (&virmem_lock);

    // Update TLB
    invlpg (page);
  }

  void *
  kpmap (uint64_t addr, bool writable, bool executable)
  {
    size_t pte_size = use_pae () ? 8 : 4;
    for (size_t i = 0; i < sizeof (pt_) / pte_size; ++i)
      {
        void *page = reinterpret_cast < void *> (0xf0000000 + i * 0x1000);
        pte pt_entry = kpte (page);
        accuire (&virmem_lock);
        if (!pt_entry.present ())
          {
            pt_entry.present (true);
            release (&virmem_lock);
            kpmap (page, addr, writable, executable);
            return page;
          }
        else
          {
            release (&virmem_lock);
          }
      }

    errno = EAGAIN;
    return nullptr;
  }

  uint64_t
  kpmapping (void *page)
  {
    return kpte (page).address ();
  }

  void
  kpunmap (void *page)
  {
    accuire (&virmem_lock);
    kpte (page).present (false);
    release (&virmem_lock);
  }

  pte
  kpte (void *page)
  {
    size_t pte_size = use_pae () ? 8 : 4;
    return pte (pt_ + ((reinterpret_cast < size_t > (page)
                        - 0xf0000000) / 0x1000 * pte_size));
  }

  void
  kpfree (void *page, size_t count)
  {
    for (size_t i = 0; i < count; ++i)
      {
        accuire (&virmem_lock);
        pte pt_entry = kpte (static_cast < char *> (page) + i * 0x1000);

        if (!pt_entry.present ())
          {
            continue;
          }

        pt_entry.present (false);
        pfree (pt_entry.address () / 0x1000);
        release (&virmem_lock);
      }

    // Update TLB; if ``invlpg'' isn't supported, reload %cr3
#if TARGET_CPU == i386
    invlpg (nullptr); // No matter what we pass, it is ignored
#else
    for (size_t i = 0; i < count; ++i)
      {
        invlpg (static_cast < char *> (page) + i * 0x1000);
      }
#endif
  }
}
