/*
proc.cpp - Process management and thread scheduling functions

This file is a part of Fiku.
Copyright (C) 2021 Akib Azmain Turja

Fiku is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Fiku is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

Authors: Akib Azmain Turja
*/

#include <fiku/proc.hpp>
#include <fiku/panic.h>
#include "context.h"
#include "idt_routines.hpp"
#include <fiku/cpu.h>
#include <stdio.h>
#include <fiku/vga.h>
#include <string.h>
#include <fiku/gdt.hpp>
#include <fiku/paging.hpp>

extern "C" void
proc_init_thrd_ ();

namespace
{

  // Current thread
  proc::thrd_t *current_thrd;

  // First thread in queue
  proc::thrd_t *first_thrd;

  // Last thread in queue
  proc::thrd_t *last_thrd;

  // Task State Segment
  proc::tss_t tss;

  // The timer being used
  timer::basic_timer *timer_;

  __attribute__ ((constructor (10))) void
  init ()
  {

    // Nullify globals variables
    current_thrd = nullptr;
    first_thrd = nullptr;
    last_thrd = nullptr;
    timer_ = nullptr;

    // Setup TSS
    memset (&tss, 0, sizeof (tss));
    tss.cs = 0x1b;
    tss.ds = 0x23;
    tss.ss = 0x23;
    tss.es = 0x23;
    tss.fs = 0x23;
    tss.gs = 0x23;
    tss.ss0 = 0x10;

    // Setup TSS entry
    gdt::get ()[5] = gdt::tsse (reinterpret_cast < uint32_t > (&tss),
                                sizeof (tss));

    // Flush TSS
    asm volatile ("movw $0x28, %ax;"
                  "ltr %ax;");
  }

  proc::thrd_t *
  pop_thrd ()
  {

    // If there is no thread, return
    if (!first_thrd)
      {
        return nullptr;
      }

    auto thrd = first_thrd;
    first_thrd = thrd->next_thrd;
    thrd->next_thrd = nullptr;
    return thrd;
  }

  void
  push_thrd (proc::thrd_t *thrd)
  {

    // If there are threads, push the thread back, otherwise put it in the
    // front
    if (first_thrd)
      {
        last_thrd->next_thrd = thrd;
      }
    else
      {
        first_thrd = thrd;
      }

    last_thrd = thrd;
  }

  proc::thrd_t *
  get_thrd ()
  {

    // Get a thread
    proc::thrd_t *next = pop_thrd ();

    // The first thread we have tested, used to prevent deadlocks
    proc::thrd_t *first_tested_thrd = next;

    for (size_t i = 0; true; ++i)
      {

        // If the thread in nullptr, break
        if (!next)
          {
            break;
          }

        // Make sure we don't stuck in an infinite loop
        else if (next == first_tested_thrd && i > 0)
          {
            next = nullptr;
            break;
          }

        // If the thread is detached, remove it from the queue
        else if (!next->attached)
          {

            // Call detach handler
            if (next->after_detach)
              {
                next->after_detach (next->after_detach_arg);
              }

            next = pop_thrd ();
          }

        // If the thread isn't runnable, skip it
        else if (next->thrd_state != next->thrd_runnable)
          {
            push_thrd (next);
            next = pop_thrd ();
          }

        // Otherwise, return it
        else
          {
            push_thrd (next);
            break;
          }
      }

    return next;
  }

  proc::thrd_t *&
  get_current_thrd ()
  {
    return current_thrd;
  }
}

namespace proc
{
  void
  thrd_prep (thrd_t *thrd)
  {
    struct init_context_t
    {
      proc::thrd_t **current_thrd_store_ptr;
      proc::thrd_t *thrd;
      uint32_t cr3;
    };

    struct contexts_t
    {
      context_t load_context;
      init_context_t thrd_init_context;
      idt_routine::interrupt_frame iret_context;
    };

    size_t esp;

    // Set the esp for the building the contexts according RPL in %cs
    if ((thrd->reg_state->cs & 0x3) == 0x0)
      {
        esp = thrd->reg_state->esp;
        thrd->usr_thrd = false;
      }
    else
      {
        esp = reinterpret_cast < uint32_t > (thrd->kernel_stack);
        thrd->usr_thrd = true;
      }

    // Reserve memory for contexts
    esp -= sizeof (contexts_t);

    // If the thread is a kernel thread, remove ss and esp from contexts
    if (!thrd->usr_thrd)
      {
        esp += (sizeof (idt_routine::interrupt_frame::ss)
                + sizeof (idt_routine::interrupt_frame::esp));
      }

    contexts_t *contexts = reinterpret_cast < contexts_t *> (esp);

    // Setup contexts
    contexts->load_context.b = 0;
    contexts->load_context.si = 0;
    contexts->load_context.di = 0;
    contexts->load_context.bp = 0;
    contexts->load_context.ip = (reinterpret_cast < uint32_t >
                                 (proc_init_thrd_));

    contexts->thrd_init_context.current_thrd_store_ptr =
      &get_current_thrd ();
    contexts->thrd_init_context.thrd = thrd;
    contexts->thrd_init_context.cr3 = thrd->vas;

    contexts->iret_context.eax = thrd->reg_state->eax;
    contexts->iret_context.edx = thrd->reg_state->ebx;
    contexts->iret_context.ecx = thrd->reg_state->ecx;
    contexts->iret_context.edx = thrd->reg_state->edx;
    contexts->iret_context.esi = thrd->reg_state->esi;
    contexts->iret_context.edi = thrd->reg_state->edi;
    contexts->iret_context.ebp = thrd->reg_state->ebp;
    contexts->iret_context.ds = thrd->reg_state->ds;
    contexts->iret_context.eip = thrd->reg_state->eip;
    contexts->iret_context.cs = thrd->reg_state->cs;
    contexts->iret_context.eflags = thrd->reg_state->eflags;

    if ((thrd->reg_state->cs & 0x3) != 0)
      {
        contexts->iret_context.ss = thrd->reg_state->ds;
        contexts->iret_context.esp = thrd->reg_state->esp;
      }

    thrd->kernel_context = contexts;

    // Run thread for 50 milliseconds before switching
    thrd->ttl = 50;
  }

  void
  thrd_add (thrd_t *thrd)
  {
    thrd->attached = true;
    push_thrd (thrd);
  }

  void
  thrd_rm (thrd_t *thrd)
  {
    thrd->attached = false;
  }

  proc::thrd_t *
  thrd_current ()
  {
    return get_current_thrd ();
  }

  void
  thrd_yield (size_t time_lived, idt::registers_t *regs)
  {

    // Get a thread
    thrd_t *thrd = get_thrd ();

    // If there is not thread to execute, panic
    if (!thrd)
      {
        panic ("panic: no threads to execute");
      }

    // Get current thread and update it
    thrd_t *current = thrd_current ();
    current->reg_state = regs;
    current->cpu_time += time_lived;

    // If this is the next thread to execute, just return
    if (current == thrd)
      {
        return;
      }

    // Set timer
    get_timer ().set (thrd->ttl);

    // Update TSS
    if (thrd->usr_thrd)
      {
        tss.esp0 = reinterpret_cast < uint32_t > (thrd->kernel_stack);
      }

    // Switch contexts
    context_switch ((reinterpret_cast < context_t **>
                     (&current->kernel_context)),
                    (static_cast < context_t *>
                     (thrd->kernel_context)));

    // Update current thread and VAS
    get_current_thrd () = current;
    paging::flush (current->vas);
  }

  void
  thrd_start ()
  {

    // Get a thread
    thrd_t *thrd = get_thrd ();

    // If there is no thread, panic
    if (!thrd)
      {
        panic ("error: no threads to execute");
      }

    // Set and start the timer
    get_timer ().set (thrd->ttl);
    get_timer ().start ();

    // Update TSS
    if (thrd->usr_thrd)
      {
        tss.esp0 = reinterpret_cast < uint32_t > (thrd->kernel_stack);
      }

    // Load the context
    asm volatile ("call context_load"
                  :
                  : "d" (thrd->kernel_context));

    // Never executed, as context_load never returns
    cpu_freeze ();
  }

  void
  set_timer (timer::basic_timer &timer)
  {
    timer_ = &timer;
  }

  timer::basic_timer &
  get_timer ()
  {
    return *timer_;
  }

  port_t *
  port_alloc (bool protection)
  {
    port_t *port = new port_t;
    port->protection = protection;
    port->receivers = nullptr;
    port->ref_count = 0;
    return port;
  }

  void
  port_free (port_t *port)
  {
    delete port;
  }

  void
  port_priv (port_t *port, thrd_t *thrd, size_t rights);

  void
  port_send (port_t *port, size_t time_lived, idt::registers_t *regs);

  idt::registers_t *
  port_recv (port_t *port, size_t time_lived, idt::registers_t *regs);

  tss_t *
  get_tss ()
  {
    return &tss;
  }
}
