/*
pit.cpp - PIT management code

This file is a part of Fiku.
Copyright (C) 2021 Akib Azmain Turja

Fiku is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Fiku is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

Authors: Akib Azmain Turja
*/

#include <fiku/timers/pit.hpp>
#include <fiku/io.h>

#define PIT_FREQUENCY 1193182

namespace
{

  // How many ticks wanted?
  size_t tick_count;

  // How namy ticks elapsed?
  size_t ticks_elapsed;

  // The interrupt handler
  timer::pit_t::handler_t interrupt_handler;

  // Whether to call the handler
  bool call_handler;

  // Returns the minimum value
  constexpr size_t
  min (size_t a, size_t b)
  {
    return a < b ? a : b;
  }

  // Sets PIT tick count
  void
  pit_set_tick_count (uint16_t tick_count)
  {
    io_outb (0x36, 0x43);
    io_outb (tick_count & 0xff, 0x40);
    io_outb (tick_count >> 8, 0x40);
  }

  // Calls the handler if it's right time, or waits
  void
  dispatcher (uint8_t, uint8_t, idt::registers_t &regs)
  {

    // If it's not the right time, wait
    if (ticks_elapsed < tick_count)
      {

        // If pending tick count is less than 0xffff, reset counter
        if (tick_count - ticks_elapsed < 0xffff)
          {
            ticks_elapsed += tick_count - ticks_elapsed;
            pit_set_tick_count (tick_count - ticks_elapsed);
          }

        // Otherwise, the counter is already correctly setup
        else
          {
            ticks_elapsed += 0xffff;
            return;
          }
      }

    // Otherwise, call the handler
    else
      {
        size_t time = ticks_elapsed * 1000 / PIT_FREQUENCY;
        ticks_elapsed = min (tick_count, 0xffff);
        interrupt_handler (0x20, time, regs);
      }
  }
}

namespace timer
{
  pit_t::pit_t ()
  {

    // Disable PIT
    io_outb (io_inb (0x21) | (1 << 0), 0x21);

    // Initialize
    tick_count = 0;
    ticks_elapsed = 0;
    interrupt_handler = nullptr;
    call_handler = false;
    idt::set_handler (0x20, dispatcher);
  }

  void
  pit_t::set (size_t interval, pit_t::handler_t handler)
  {
    if (interval == 0)
      {

        // Set tick count to maximum, on x86 it'll take about an hour to
        // complete this amount to ticks
        tick_count = SIZE_MAX;
        call_handler = false;
      }

    else
      {
        tick_count = (interval * PIT_FREQUENCY) / 1000;
        ticks_elapsed = 0;

        // If handler is specified, update handler
        if (handler)
          {
            interrupt_handler = handler;
          }

        if (!interrupt_handler)
          {
            call_handler = false;
          }
      }

    // The PIT doesn't support waiting for more than 0xffff ticks, so
    // we have to *emulate* it
    uint16_t pit_tick_count = min (tick_count, 0xffff);
    ticks_elapsed += pit_tick_count;

    pit_set_tick_count (pit_tick_count);
  }

  void
  pit_t::start ()
  {
    io_outb (io_inb (0x21) & ~(1 << 0), 0x21);
  }

  size_t
  pit_t::read () const
  {
    io_outb (0x00, 0x43);
    uint32_t pit_tick_left = io_inb (0x40);
    pit_tick_left |= io_inb (0x40) << 8;

    // Convert to milliseconds
    return (ticks_elapsed - pit_tick_left) * 1000 / PIT_FREQUENCY;
  }

  void
  pit_t::stop ()
  {
    io_outb (io_inb (0x21) | (1 << 0), 0x21);
  }

  pit_t pit;
}
