/*
locks.cpp - Lock implementations

This file is a part of Fiku.
Copyright (C) 2021 Akib Azmain Turja

Fiku is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Fiku is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

Authors: Akib Azmain Turja
*/

#include <threads.h>
#include <fiku/spinlock.h>

namespace
{
  bool
  trylock (volatile bool *lock)
  {
    bool locked_prev;

    // Try to lock
    asm volatile ("movb $1, %0;"
                  "lock xchgb %0, (%1);"
                  : "=a" (locked_prev)
                  : "b" (lock));

    return locked_prev;
  }
}

int
mtx_lock (mtx_t *mtx)
{

  // Loop until success
  while (true)
    {
      int status = mtx_trylock (mtx);

      // Failed, try again
      if (status == thrd_busy)
        {
          thrd_yield ();
        }

      // Either succeeded, or an unexpected error occurred, report it
      else
        {
          return status;
        }
    }
}

int
mtx_trylock (mtx_t *mtx)
{

  // Lock the mutex itself
  if (trylock (&mtx->self_locked))
    {
      return thrd_busy;
    }

  // If it's already locked, return error
  if (mtx->locker &&
      (!mtx->recursive || !thrd_equal (thrd_current (), mtx->locker)))
    {
      mtx->self_locked = false;
      return thrd_busy;
    }

  // Otherwise, if the mutex is not locked, lock it
  else if (!mtx->locker)
    {
      mtx->locker = thrd_current ();
    }

  if (mtx->recursive)
    {

      // Increase recursion count
      ++mtx->recursion_count;
    }

  mtx->self_locked = false;
  return thrd_success;
}

int
mtx_unlock (mtx_t *mtx)
{

  // Lock the mutex itself
  if (trylock (&mtx->self_locked))
    {
      return thrd_error;
    }

  // If the locker is not the current thread, return failure
  if (!thrd_equal (thrd_current (), mtx->locker))
    {
      mtx->self_locked = false;
      return thrd_busy;
    }

  // If the mutex is recursive, decrement recursion count
  if (mtx->recursive)
    {
      --mtx->recursion_count;
    }

  // If the mutex is plain or the recursion count in zero, unlock it
  if (!mtx->recursive || mtx->recursion_count == 0)
    {
      mtx->locker = nullptr;
    }

  mtx->self_locked = false;
  return thrd_success;
}

int
spnlk_init (spnlk_t *lock)
{
  lock->locked = false;
  return spnlk_success;
}

int
spnlk_trylock (spnlk_t *lock)
{
  if (trylock (&lock->locked) == true)
    {
      return spnlk_busy;
    }
  else
    {
      return spnlk_success;
    }
}

int
spnlk_lock (spnlk_t *lock)
{
  int status;

  do
    {
      status = spnlk_trylock (lock);
    }
  while (status == spnlk_busy);

  return status;
}

int
spnlk_unlock (spnlk_t *lock)
{
  lock->locked = false;
  return spnlk_success;
}

int
spnlk_destroy (spnlk_t *)
{
  // Nothing to do...
  return spnlk_success;
}
