/*
vas.cpp - Virtual Address Space management functions for i386 architecture

This file is a part of Fiku.
Copyright (C) 2021 Akib Azmain Turja

Fiku is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Fiku is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

Authors: Akib Azmain Turja
*/

/*
 * WARNING: This file contains contains critical functions which accesses
 * and modifies the physical memory directly (actually, it maps that part
 * of the physical memory to somewhere in the virtual address space). Be
 * very careful when modifing this file, as incorrect modification may (and
 * will) introduce severe bugs which will be very difficult to find out and
 * even might go unnoticed (it took about 40 hours for the original author,
 * Akib, to find a bug in the clone function). YOU HAVE BEEN WARNED!
 */

#include <fiku/vas.hpp>
#include <fiku/cpuid.h>
#include <errno.h>
#include <string.h>
#include <config.h>
#include <fiku/panic.h>

namespace
{

  // Returns whether to use PAE
  bool
  use_pae ()
  {
    return ((PAE_SUPPORT && !PAE_SUPPORT_RUNTIME)
            || (PAE_SUPPORT_RUNTIME && cpuid_check ()
                && cpuid_feature_pae ()));
  }

  struct mapping_t
  {
    uint64_t key; // Physical address
    void *data; // Virtual address
    size_t ref_count; // How many references to the mapping?

    // Fields used by AVL (Adelson-Velsky and Landis) tree algorithm
    size_t height;
    mapping_t *left;
    mapping_t *right;
  };

  struct pager_vas_info_t
  {
    uint64_t root_paging_table;
    char *pdpt;
    char *pd;
    char *pt;
    mapping_t *cached_mappings;
    mapping_t *used_mappings;
    void *data_offset;
    mapping_t *mappings;
    size_t mapping_create_index;
    size_t max_mapping_count;
  };

  pager_vas_info_t pager_vas_info;

  template < class arg_t > constexpr arg_t
  max (arg_t a, arg_t b)
  {
    return a < b ? b : a;
  }

  template < class node_t > node_t *
  avl_rotate_left (node_t *root)
  {

    // Rotate
    node_t *pivot = root->right;
    root->right = pivot->left;
    pivot->left = root;

    // Update heights
    auto update_height = [] (node_t *node) -> void
    {
      node->height = 1 + max ((node->left ? node->left->height : 0),
                              (node->right ? node->right->height : 0));
    };

    update_height (root);
    update_height (pivot);

    // Return new root
    return pivot;
  }

  template < class node_t > node_t *
  avl_rotate_right (node_t *root)
  {

    // Rotate
    node_t *pivot = root->left;
    root->left = pivot->right;
    pivot->right = root;

    // Update heights
    auto update_height = [] (node_t *node) -> void
    {
      node->height = 1 + max ((node->left ? node->left->height : 0),
                              (node->right ? node->right->height : 0));
    };

    update_height (root);
    update_height (pivot);

    // Return new root
    return pivot;
  }

  template < class node_t > node_t *
  avl_insert (node_t *root, node_t *node)
  {
    if (!root)
      {
        node->height = 1;
        return node;
      }

    // Insert
    if (node->key < root->key)
      {
        root->left = avl_insert (root->left, node);
      }
    else
      {
        root->right = avl_insert (root->right, node);
      }

    // Update height
    root->height = 1 + max ((root->left ? root->left->height : 0),
                            (root->right ? root->right->height : 0));

    // Balance
    auto bf = [] (node_t *node) -> ptrdiff_t
    {
      return ((node->left ? node->left->height : 0)
              - (node->right ? node->right->height : 0));
    };

    if (bf (root) > 1)
      {
        if (node->key < root->left->key)
          {
            root = avl_rotate_right (root);
          }
        else
          {
            root->left = avl_rotate_left (root->left);
            root = avl_rotate_right (root);
          }
      }
    else if (bf (root) < -1)
      {
        if (node->key < root->right->key)
          {
            root->right = avl_rotate_right (root->right);
            root = avl_rotate_left (root);
          }
        else
          {
            root = avl_rotate_left (root);
          }
      }

    return root;
  }

  template < class node_t > node_t *
  avl_delete (node_t *root, node_t *node)
  {

    // If root is null, return null
    if (!root)
      {
        return nullptr;
      }

    // If the current node is the node to delete, delete it
    else if (root == node)
      {

        // If the node has no child, simple delete it
        if (!node->left && !node->right)
          {
            root = nullptr;
          }

        // Otherwise, if the node has one child, replace it with that
        else if (!node->left)
          {
            root = node->right;
          }
        else if (!node->right)
          {
            root = node->left;
          }

        // Otherwise, the node has two children, replace the node with the
        // node with the maximum key value
        else
          {

            // Get the node with minimum key in right sub-tree
            node_t *right_min = node->right;
            while (right_min->left)
              {
                right_min = right_min->left;
              }

            right_min->right = avl_delete (root->right, right_min);
            right_min->left = root->left;
            root = right_min;
          }

        // Remove external references from node to prevent infinite loops
        node->left = nullptr;
        node->right = nullptr;
      }

    // Otherwise, find and delete the node in sub-tree
    else if (node->key < root->key)
      {
        root->left = avl_delete (root->left, node);
      }
    else
      {
        root->right = avl_delete (root->right, node);
      }

    // If root in nullptr, return
    if (!root)
      {
        return nullptr;
      }

    // Update height
    root->height = 1 + max ((root->left ? root->left->height : 0),
                            ((root->right) ? root->right->height : 0));

    // Balance
    auto bf = [] (node_t *node) -> ptrdiff_t
    {
      return ((node->left ? node->left->height : 0)
              - (node->right ? node->right->height : 0));
    };

    if (bf (root) > 1)
      {
        if (bf (root->left) >= 0)
          {
            root = avl_rotate_right (root);
          }
        else
          {
            root->left = avl_rotate_left (root->left);
            root = avl_rotate_right (root);
          }
      }
    else if (bf (root) < -1)
      {
        if (bf (root->right) <= 0)
          {
            root = avl_rotate_left (root);
          }
        else
          {
            root->right = avl_rotate_right (root->right);
            root = avl_rotate_left (root);
          }
      }

    // Return new node
    return root;
  }

  template < class node_t > node_t *
  avl_search (node_t *root, decltype (node_t::key) key)
  {

    // If root is a valid node, test it and it's children
    if (root)
      {
        if (key == root->key)
          {
            return root;
          }
        else if (key < root->key)
          {
            return avl_search (root->left, key);
          }
        else
          {
            return avl_search (root->right, key);
          }
      }

    // Otherwise, not found, and return nullptr
    else
      {
        return nullptr;
      }
  }

  mapping_t *
  create_mapping ()
  {

    // Make sure we can allocate
    if (pager_vas_info.mapping_create_index
        >= pager_vas_info.max_mapping_count)
      {
        return nullptr;
      }

    size_t i = pager_vas_info.mapping_create_index++;

    // Fill mapping object
    pager_vas_info.mappings[i].key = i << 12 | 0xfff; // Invalid key, just
                                                      // for making node
                                                      // keys unique
    pager_vas_info.mappings[i].data =
      reinterpret_cast < char *> (pager_vas_info.data_offset) + i * 0x1000;

    pager_vas_info.mappings[i].ref_count = 0;
    pager_vas_info.mappings[i].height = 0;
    pager_vas_info.mappings[i].left = nullptr;
    pager_vas_info.mappings[i].right = nullptr;

    return &pager_vas_info.mappings[i];
  }

  // Mapping a physical address somewhere into VAS (Virtual Address Space)
  mapping_t *
  map (uint64_t addr)
  {

    // Check if the page is already in use some other area
    auto node = avl_search (pager_vas_info.used_mappings, addr);
    if (node)
      {
        ++node->ref_count;
        return node;
      }

    // Otherwise, check if the page is cached
    else
      {
        node = avl_search (pager_vas_info.cached_mappings, addr);
        if (node)
          {

            // Remove it from the list of cached pages
            pager_vas_info.cached_mappings =
              avl_delete (pager_vas_info.cached_mappings, node);

            ++node->ref_count;

            // Insert it to the list of used pages
            pager_vas_info.used_mappings =
              avl_insert (pager_vas_info.used_mappings, node);

            // Return
            return node;
          }
      }

    // The page is neither cached nor in use in other thread, expand
    // mappings
    node = create_mapping ();

    // Still no mapping found, remove a cached page and map it
    if (!node && pager_vas_info.cached_mappings)
      {
        node = pager_vas_info.cached_mappings;
        pager_vas_info.cached_mappings =
          avl_delete (pager_vas_info.cached_mappings, node);
      }

    // Make sure we've a node to use
    if (node)
      {

        // Set address
        node->key = addr;

        // Increase reference count
        ++node->ref_count;

        // Map it
        paging::pte (pager_vas_info.pt
                     + (vas::vas_t::to_page (node->data)
                        * (use_pae () ? 8 : 4))).init (addr, true, false);

        // Update TLB
        paging::invlpg (node->data);

        // Insert it to the list of used pages
        pager_vas_info.used_mappings =
          avl_insert (pager_vas_info.used_mappings, node);

        // Return
        return node;
      }

    // No node, no memory, resource busy.  Stop.
    else
      {
        errno = EBUSY;
        return nullptr;
      }
  }

  // Unmaps (or more accurately, marks as unused) a temporary mapping
  void
  unmap (mapping_t *mapping)
  {

    // Decrease reference count
    --mapping->ref_count;

    // If it is totally free...
    if (mapping->ref_count == 0)
      {

        // Remove it from the list of used pages
        pager_vas_info.used_mappings =
          avl_delete (pager_vas_info.used_mappings, mapping);

        // Insert the page into the tree of caches
        pager_vas_info.cached_mappings =
          avl_insert (pager_vas_info.cached_mappings, mapping);
      }
  }

  // RAII (Resource Acquisition Is Initialization) class which maps an
  // address on construction (or declaration) and unmaps on destruction
  // (when it goes out of scope) (one of the reasons why I chose C++ over
  // C)
  class smart_mapping
  {
  private:
    mapping_t *mapping_;

  public:

    // Maps to the given address
    smart_mapping (uint64_t addr)
      : mapping_ (map (addr))
    {}

    // Initializes and attaches given mapping
    smart_mapping (mapping_t *mapping)
      : mapping_ (mapping)
    {}

    // We don't want to cpoy it implicitly
    smart_mapping (const smart_mapping &) = delete;

    // Destructor, unmaps mapping
    ~smart_mapping ()
    {

      // Only unmap if mapping isn't null (may be due to error during
      // construction, or the call function explicitly detached it)
      if (mapping_)
        {
          unmap (mapping_);
        }
    }

    // Returns the mapping structure
    const mapping_t *
    mapping () const
    {
      return mapping_;
    }

    // Unmaps previous mapping or attaches new one
    void
    mapping (mapping_t *mapping)
    {
      if (mapping_)
        {
          unmap (mapping_);
        }

      mapping_ = mapping;
    }

    // Attaches a mapping
    void
    attach (mapping_t *mapping)
    {
      mapping_ = mapping;
    }

    // Detach the mapping
    mapping_t *
    detach ()
    {
      auto prev_mapping = mapping_;
      mapping_ = nullptr;
      return prev_mapping;
    }

    // Returns pointer to the mapping
    template < class type = void *> type
    ptr ()
    {
      return mapping_ ? static_cast < type > (mapping_->data) : nullptr;
    }

    // We don't want to copy it implicitly
    smart_mapping &
    operator = (const smart_mapping &) = delete;
  };

  // Return mapping to page table
  smart_mapping
  get_pt (uint64_t addr, size_t index, bool create = false,
          uint64_t *pdpte_p = nullptr, uint64_t *pde_p = nullptr)
  {

    // PTE size
    size_t pte_size = use_pae () ? 8 : 4;

    // PDE size
    size_t pde_size = pte_size;

    // PD coverage (i.e amount of pages each PD controls)
    size_t pde_cov = 0x1000 / pte_size;

    // PDPT coverage (i.e amount of pages each PDPT controls)
    size_t pdpte_cov = 0x40000000 / 0x1000;

    // Nullifies a page
    auto nullify = [] (uint64_t addr)
    {
      smart_mapping page = addr;

      if (errno == EBUSY)
        {
          return;
        }

      memset (page.ptr (), 0, 0x1000);
    };

    // Returns mapping to PT
    auto get_pt_pd = [&] (uint64_t addr, size_t index) -> mapping_t *
    {

      // Map the memory
      smart_mapping page = addr;

      // If failed, return
      if (errno == EBUSY)
        {
          return nullptr;
        }

      // Get the PDE
      auto pde = paging::pde (page.ptr < char *> () +
                              index / pde_cov * pde_size);
      uint64_t new_phys_pg = UINT64_MAX;

      // If the PDE is non-present, create it or return
      if (!pde.present ())
        {
          if (create)
            {

              // Allocate a page
              uint64_t phys_pg = paging::palloc ();

              // If page frame allocation failed, return
              if (errno == EAGAIN)
                {
                  return nullptr;
                }

              // Nothing to explain, fill the memory with zeros
              nullify (phys_pg);

              // If nullification failed, cleanup and return
              if (errno == EBUSY)
                {
                  paging::pfree (phys_pg);
                  return nullptr;
                }

              // Setup the PDE
              pde.init (phys_pg, true, true, false);
              new_phys_pg = phys_pg;
            }

          // PT not found, return
          else
            {
              return nullptr;
            }
        }

      // If this is a modifing operation and the PT is COW (Copy On Write,
      // not to be with the animal), copy it if required
      else if (create && pde.present () && pde.swapped ()
               && !pde.writable ())
        {

          // If the PT is refered more than once (i.e shared between VASs),
          // copy it
          if (paging::puser (pde.address ()) > 1)
            {

              // Map source page
              smart_mapping src = pde.address ();

              if (errno == EBUSY)
                {
                  return nullptr;
                }

              // Allocate a page
              uint64_t phys_pg = paging::palloc ();

              if (errno == EAGAIN)
                {
                  return nullptr;
                }

              // Map destination page
              smart_mapping dest = phys_pg;

              if (errno == EBUSY)
                {
                  paging::pfree (phys_pg);
                  return nullptr;
                }

              // Copy PT
              memcpy (dest.ptr (), src.ptr (), 0x1000);

              // Mark all PTEs as swapped
              for (size_t i = 0; i < 0x1000 / pte_size; ++i)
                {
                  auto pte = paging::pte (dest.ptr < char *> ()
                                          + i * pte_size);

                  if (pte.writable ())
                    {
                      pte.writable (false);
                      pte.swapped (true);

                      // Increment reference count
                      paging::palloc (pte.address ());
                    }
                }

              // Decrement reference count
              paging::pfree (pde.address ());

              // Set new address
              pde.address (phys_pg);
            }

          // Mark as normal
          pde.writable (true);
          pde.swapped (false);
        }

      // Now map it
      mapping_t *ptr = map (pde.address ());

      // Write to pointer if given
      if (pde_p)
        {
          *pde_p = addr + index / pde_cov * pde_size;
        }

      // If errors occured, cleanup and return
      if (new_phys_pg != UINT64_MAX && (errno == EBUSY || errno == EAGAIN))
        {
          pde.present (false);
          paging::pfree (new_phys_pg);
          return nullptr;
        }

      // Return the result
      return ptr;
    };

    auto get_pt_pdpt = [&] (uint64_t addr, size_t index) -> mapping_t *
    {

      // Map the memory
      smart_mapping page = addr;

      // If failed, return
      if (errno == EBUSY)
        {
          return nullptr;
        }

      // Get the PDPTE
      auto pdpte = paging::pdpte (page.ptr < char *> () +
                                  index / pdpte_cov * 8);
      uint64_t new_phys_pg = UINT64_MAX;

      // If the PDPTE is non-present, create it or return
      if (!pdpte.present ())
        {
          if (create)
            {
              uint64_t phys_pg = paging::palloc ();

              // If page frame allocation failed, return
              if (errno == EAGAIN)
                {
                  return nullptr;
                }

              // Nothing to explain, fill the memory with zeros
              nullify (phys_pg);

              // If nullification failed, cleanup and return
              if (errno == EBUSY)
                {
                  paging::pfree (phys_pg);
                  return nullptr;
                }

              // Setup the PDPTE
              pdpte.init (phys_pg);
              new_phys_pg = phys_pg;
            }

          // PT not found, return
          else
            {
              return nullptr;
            }
        }

      // Now find
      mapping_t *ptr = get_pt_pd (pdpte.address (), index % pdpte_cov);

      // Write to pointer if given
      if (pdpte_p)
        {
          *pdpte_p = addr + index / pdpte_cov * 8;
        }

      // If errors occured, cleanup and return
      if (new_phys_pg != UINT64_MAX && (errno == EBUSY || errno == EAGAIN))
        {
          pdpte.present (false);
          paging::pfree (new_phys_pg);
          return nullptr;
        }

      // Return the result
      return ptr;
    };

    // RETURN THE PTE
    if (use_pae ())
      {
        return get_pt_pdpt (addr, index);
      }
    else
      {
        return get_pt_pd (addr, index);
      }
  }
}

// Page tables
extern char pt_[];

// Page directory
extern char pd_[];

// Page directory pointer table
extern char pdpt_[];

namespace vas
{
  vas_t::vas_t ()
    : root_paging_table (0), initialized (false)
  {

    // Allocate a block from physical memory
    root_paging_table = paging::palloc ();

    // Make sure allocation
    if (errno == EAGAIN)
      {
        return;
      }

    // Map it
    smart_mapping page = root_paging_table;

    // Make sure it is mapped
    if (errno == EBUSY)
      {
        paging::pfree (root_paging_table);
        return;
      }

    if (use_pae ())
      {

        // Nullify first three PDPTE
        for (size_t i = 0; i < 3; ++i)
          {
            paging::pdpte (page.ptr < char *> () + i * 8).init ();
          }

        // Allocate physical block for page directory
        uint64_t pd = paging::palloc ();

        // Make sure allocation
        if (errno == EAGAIN)
          {
            paging::pfree (root_paging_table);
            return;
          }

        // Setup the last PDPTE
        paging::pdpte (page.ptr < char *> () + 3 * 8).init (pd);

        // Map kernel page directory
        smart_mapping pd_pg = pd;

        // Make sure it is mapped
        if (errno == EBUSY)
          {

            // Release memory and return
            paging::pfree (pd);
            paging::pfree (root_paging_table);
            return;
          }

        // Copy kernel page directory
        memcpy (pd_pg.ptr (), pd_, 0x1000);
      }
    else
      {

        // Copy kernel page directory
        memcpy (page.ptr (), pd_, 0x1000);
      }

    initialized = true;
  }

  vas_t::vas_t (vas_t &&other)
    : root_paging_table (other.root_paging_table),
      initialized (other.initialized)
  {
    other.initialized = false;
  }

  vas_t::~vas_t ()
  {
    _free ();
  }

  void
  vas_t::map (size_t page, bool writable, bool executable)
  {

    // Allocate a page
    uint64_t phymem = paging::palloc ();

    if (errno != EAGAIN)
      {

        // Map it
        map (page, phymem, writable, executable);

        // If failed, cleanup
        if (errno == ERANGE)
          {
            paging::pfree (phymem);
          }
      }
  }

  void
  vas_t::map (size_t page, uint64_t physaddr, bool writable,
              bool executable)
  {

    // Make sure in range
    if (page * 0x1000 >= 0xf0000000)
      {
        errno = ERANGE;
        return;
      }

    // PTE size
    size_t pte_size = use_pae () ? 8 : 4;

    // Index in PT
    size_t pte_i = (use_pae ()
                    ? ((page % 0x40000) % 0x200)
                    : (page % 0x400));

    // Get the PT
    smart_mapping &&pt_pg = get_pt (root_paging_table, page, true);

    if (errno == EBUSY || errno == EAGAIN)
      {
        return;
      }

    // Write the PT
    paging::pte (pt_pg.ptr < char *> ()
                 + pte_i * pte_size).init (physaddr, writable,
                                           executable, false);
  }

  uint64_t
  vas_t::mapping (size_t page) const
  {

    // Make sure in range
    if (page * 0x1000 >= 0xf0000000)
      {
        errno = ERANGE;
        return 0;
      }

    // PTE size
    size_t pte_size = use_pae () ? 8 : 4;

    // Index in PT
    size_t pte_i = (use_pae ()
                    ? ((page % 0x40000) % 0x200)
                    : (page % 0x400));

    // Get the PT
    smart_mapping &&pt_pg = get_pt (root_paging_table, page);

    if (errno == EBUSY || errno == EAGAIN || !pt_pg.mapping ())
      {
        return UINT64_MAX;
      }

    // Get the physical address
    return paging::pte (pt_pg.ptr < char *> ()
                        + pte_i * pte_size).address ();
  }

  bool
  vas_t::writable (size_t page) const
  {

    // Make sure in range
    if (page * 0x1000 >= 0xf0000000)
      {
        errno = ERANGE;
        return 0;
      }

    // PTE size
    size_t pte_size = use_pae () ? 8 : 4;

    // Index in PT
    size_t pte_i = (use_pae ()
                    ? ((page % 0x40000) % 0x200)
                    : (page % 0x400));

    // Get the PT
    smart_mapping &&pt_pg = get_pt (root_paging_table, page);

    if (errno == EBUSY || errno == EAGAIN)
      {
        return false;
      }

    // Get the PTE
    auto pte =  paging::pte (pt_pg.ptr < char *> ()
                             + pte_i * pte_size);

    // If the page is readonly and swapped, it means it is COW
    return pte.writable () || pte.swapped ();
  }

  bool
  vas_t::executable (size_t page) const
  {

    // Make sure in range
    if (page * 0x1000 >= 0xf0000000)
      {
        errno = ERANGE;
        return 0;
      }

    // PTE size
    size_t pte_size = use_pae () ? 8 : 4;

    // Index in PT
    size_t pte_i = (use_pae ()
                    ? ((page % 0x40000) % 0x200)
                    : (page % 0x400));

    // Get the PT
    smart_mapping &&pt_pg = get_pt (root_paging_table, page);

    if (errno == EBUSY || errno == EAGAIN)
      {
        return false;
      }

    // Get whether the page is executable
    return paging::pte (pt_pg.ptr < char *> ()
                        + pte_i * pte_size).executable ();
  }

  void
  vas_t::touch (size_t page)
  {

    // Make sure in range
    if (page * 0x1000 >= 0xf0000000)
      {
        errno = ERANGE;
        return;
      }

    // PTE size
    size_t pte_size = use_pae () ? 8 : 4;

    // Index of PTE
    size_t pte_i = (use_pae ()
                    ? ((page % 0x40000) % 0x200)
                    : (page % 0x400));

    // Get the PT
    smart_mapping &&pt_pg = get_pt (root_paging_table, page, true);

    if (errno == EBUSY || errno == EAGAIN)
      {
        return;
      }

    auto pte = paging::pte (pt_pg.ptr < char *> () + pte_i * pte_size);

    // If the page is swapped, bring it in an immdiately usable state
    if (pte.swapped ())
      {

        // If the page is present and readable, it is COW
        if (pte.present () && !pte.writable ())
          {

            // If the page is shared (not to be confused with shared
            // memory, I means other VASs also refer to it), copy it
            if (paging::puser (pte.address ()) > 1)
              {

                // Map source page
                smart_mapping src = pte.address ();

                if (errno == EBUSY)
                  {
                    return;
                  }

                // Allocate a page
                uint64_t dest_pg = paging::palloc ();

                if (errno == EAGAIN)
                  {
                    return;
                  }

                // Map destnation page
                smart_mapping dest = dest_pg;

                if (errno == EBUSY)
                  {
                    paging::pfree (dest_pg);
                    return;
                  }

                // Copy
                memcpy (dest.ptr (), src.ptr (), 0x1000);

                // Decrement reference count
                paging::pfree (pte.address ());

                // Set new address
                pte.address (dest_pg);
              }

            // Mark as normal and usable
            pte.writable (true);
            pte.swapped (false);
          }

        // TODO: implement swapping
      }
  }

  void
  vas_t::unmap (size_t page, bool free)
  {

    // Make sure in range
    if (page * 0x1000 >= 0xf0000000)
      {
        errno = ERANGE;
        return;
      }

    // PTE size
    size_t pte_size = use_pae () ? 8 : 4;

    // PTE index
    size_t pte_i = (use_pae ()
                    ? ((page % 0x40000) % 0x200)
                    : (page % 0x400));

    // Get the PTE (note: we set create to true so that get_pt treats it as
    // a modifing operation and bring the associate structure back into
    // memory in an modifiable state, it should not alter the structures in
    // any other ways as the page is already there unless the call to this
    // function is invalid)
    uint64_t pdpte_addr, pde_addr;
    smart_mapping &&pt_pg = get_pt (root_paging_table, page, true,
                                    &pdpte_addr, &pde_addr);

    if (errno == EBUSY || errno == EAGAIN)
      {
        return;
      }

    // Get the PTE
    auto pte = paging::pte (pt_pg.ptr < char *> () + pte_i * pte_size);

    // Free if instructed
    if (free)
      {
        paging::pfree (pte.address ());
      }

    // Clean PTE (i.e clear all bits)
    pte.init ();

    // Now free unnecessary memory (e.g unused PT, PD)...

    // Check if the page table is empty
    bool empty = true;
    for (size_t i = 0; i < 0x1000 / pte_size; ++i)
      {
        if (paging::pte (pt_pg.ptr < char *> () + i * pte_size).present ())
          {

            // Atleast one is used, don't free
            empty = false;
            break;
          }
      }

    // If the page table is empty, free it too
    if (empty)
      {

        // Map PD
        smart_mapping pd_pg = pde_addr & ~0xfff;

        if (errno == EBUSY)
          {
            return;
          }

        // Mark the PDE as non-present
        auto pde = paging::pde (pd_pg.ptr < char *> ()
                                + (pde_addr & 0xfff));
        pde.present (false);

        // Free the PT
        paging::pfree (pde.address ());

        // If using PAE, check if the page directory is empty
        if (use_pae ())
          {
            for (size_t i = 0; i < 0x1000 / pte_size; ++i)
              {
                if (paging::pde (pd_pg.ptr < char *> ()
                                 + i * pte_size).present ())
                  {

                    // Atleast one PDE is used, don't free
                    empty = false;
                    break;
                  }
              }

            // If empty, deallocate it too
            if (empty)
              {

                // Map the PDPT
                smart_mapping pdpt_pg = pdpte_addr & ~0xfff;

                if (errno == EBUSY)
                  {
                    return;
                  }

                // Mark the PDPTE ans non-present
                auto pdpte = paging::pdpte (pdpt_pg.ptr < char *> ()
                                            + (pdpte_addr & 0xfff));
                pdpte.present (false);

                // Free PD
                paging::pfree (pdpte.address ());
              }
          }
      }
  }

  bool
  vas_t::is_mapped (size_t page) const
  {

    // Make sure in range
    if (page * 0x1000 >= 0xf0000000)
      {
        errno = ERANGE;
        return false;
      }

    // PTE size and index
    size_t pte_size = use_pae () ? 8 : 4;
    size_t pte_i = (use_pae ()
                    ? ((page % 0x40000) % 0x200)
                    : (page % 0x400));

    // Get the PT
    smart_mapping &&pt_pg = get_pt (root_paging_table, page);

    if (!pt_pg.mapping () || errno == EBUSY || errno == EAGAIN)
      {
        return false;
      }

    // Get whether the page is executable
    return paging::pte (pt_pg.ptr < char *> ()
                        + pte_i * pte_size).present ();
  }

  bool
  vas_t::is_mapped (void *addr) const
  {

    // Just a wrapper, call actual one
    return is_mapped (to_page (addr));
  }

  vas_t::pte_t *
  vas_t::pte (size_t page)
  {

    // Make sure in range
    if (page * 0x1000 >= 0xf0000000)
      {
        errno = ERANGE;
        return nullptr;
      }

    // PTE size and index
    size_t pte_size = use_pae () ? 8 : 4;
    size_t pte_i = (use_pae ()
                    ? ((page % 0x40000) % 0x200)
                    : (page % 0x400));

    // Get the PTE
    smart_mapping *pt_pg = new smart_mapping (get_pt (root_paging_table,
                                                      page));

    // If failed, deallocate
    if (errno == EBUSY || errno == EAGAIN)
      {
        delete pt_pg;
        return nullptr;
      }

    // Return PTE
    return new pte_t (pt_pg->ptr < char *> ()
                      + pte_i * pte_size, pt_pg);
  }

  vas_t
  vas_t::clone () const
  {

    // The VAS we're creating
    vas_t other;

    // PDE size in bytes
    size_t pde_size = use_pae () ? 8 : 4;

    // Nullifies a page
    auto nullify = [] (uint64_t addr)
    {

      // Map it
      smart_mapping page = addr;

      if (errno == EBUSY)
        {
          return;
        }

      // Nullify
      memset (page.ptr (), 0, 0x1000);
    };

    // Clones a PD
    auto clone_pd = [&] (uint64_t src_data, uint64_t dest_data, size_t offset)
    {

      // Map source
      smart_mapping src = src_data;

      if (errno == EBUSY)
        {
          return;
        }

      // Map destination
      smart_mapping dest = dest_data;

      if (errno == EBUSY)
        {
          return;
        }

      // Clone each PDE
      for (size_t i = 0; i < 0x1000 / pde_size;
           ++i, offset += 0x1000 / pde_size * 0x1000)
        {

          // If we're in the kernel's private area, stop
          if (offset >= 0xf0000000)
            {
              break;
            }

          // Source and destination PDEs
          auto src_pde = paging::pde (src.ptr < char *> () + i * pde_size);
          auto dest_pde = paging::pde (dest.ptr < char *> ()
                                       + i * pde_size);

          // If source PDE is present...
          if (src_pde.present ())
            {

              // ... and writable, make it COW
              if (src_pde.writable ())
                {
                  src_pde.swapped (true);
                  src_pde.writable (false);
                  dest_pde.init (src_pde.address (), false,
                                 src_pde.executable (), false, false, true,
                                 true);
                }

              // Or if it is readonly, writing is not possible, no need to
              // make it COW
              else
                {
                  dest_pde.init (src_pde.address (), false,
                                 src_pde.executable (), false);
                }

              // Increment reference count
              paging::palloc (src_pde.address ());
            }
        }
    };

    // Clones a PDPT
    auto clone_pdpt = [&] (uint64_t src_data, uint64_t dest_data,
                           size_t offset)
    {

      // Map source
      smart_mapping src = src_data;

      if (errno == EBUSY)
        {
          return;
        }

      // Map destination
      smart_mapping dest = dest_data;

      if (errno == EBUSY)
        {
          return;
        }

      // Allocations and their count, used during cleanup in case of errors
      uint64_t allocations[4];
      size_t allocation_count = 0;

      // Releases memory in case of errors
      auto cleanup = [&] ()
      {

        // Release every allocated page
        for (size_t i = 0; i < allocation_count; ++i)
          {
            paging::pfree (allocations[i]);
          }
      };

      // Clone each PDPTE
      for (size_t i = 0; i < 4; ++i, offset += 0x40000000)
        {

          // Don't disturb the kernel
          if (offset >= 0xf0000000)
            {
              break;
            }

          // Source and destination
          auto src_pdpte = paging::pdpte (src.ptr < char *> ()
                                          + i * pde_size);
          auto dest_pdpte = paging::pdpte (dest.ptr < char *> ()
                                           + i * pde_size);

          // If present, clone it
          if (src_pdpte.present ())
            {

              // If destination PDPTE is not already present (may not make
              // sense, but the constructor makes some PDPTEs present),
              // allocating new one
              if (!dest_pdpte.present ())
                {

                  // Allocate a page
                  uint64_t pg = paging::palloc ();

                  if (errno == EAGAIN)
                    {
                      cleanup ();
                      return;
                    }

                  // Nullify
                  nullify (pg);

                  if (errno == EBUSY)
                    {
                      paging::pfree (pg);
                      cleanup ();
                      return;
                    }

                  dest_pdpte.init (pg);
                }

              // Clone PD
              clone_pd (src_pdpte.address (), dest_pdpte.address (),
                        offset);

              // If error, cleanup and exit
              if (errno == EAGAIN || errno == EBUSY)
                {
                  cleanup ();
                  return;
                }
            }
        }
    };

    if (use_pae ())
      {
        clone_pdpt (root_paging_table, other.root_paging_table, 0);
      }
    else
      {
        clone_pd (root_paging_table, other.root_paging_table, 0);
      }

    // Return the VAS
    return other;
  }

  void *
  vas_t::to_ptr (size_t page)
  {
    return reinterpret_cast < void *> (page * 0x1000);
  }

  size_t
  vas_t::to_page (void *ptr)
  {
    return reinterpret_cast < size_t > (ptr) / 0x1000;
  }

  vas_t &
  vas_t::operator = (vas_t &&other)
  {

    // Cleanup...
    _free ();

    // .. then assign
    root_paging_table = other.root_paging_table;
    initialized = other.initialized;
    other.initialized = false;
    return *this;
  }

  void
  vas_t::_free ()
  {

    // If uninitialized, do nothing
    if (!initialized)
      {
        return;
      }

    // PTE size
    size_t pte_size = use_pae () ? 8 : 4;

    // Frees a page table
    auto free_pt = [&] (uint64_t addr, size_t offset)
    {

      // Map PT
      smart_mapping page = addr;

      // Free each page
      for (size_t i = 0; i < 0x1000 / pte_size && offset < 0xf0000000;
           ++i, offset += 0x1000)
        {

          // Release page
          auto pte = paging::pte (page.ptr < char *> () + i * pte_size);

          if (pte.present ())
            {
              paging::pfree (pte.address ());
            }
        }

      // Release PT
      paging::pfree (addr);
    };

    // Frees a page directory
    auto free_pd = [&] (uint64_t addr, size_t offset)
    {

      // Map PD
      smart_mapping page = addr;

      // Coverage of a PDE
      size_t coverage = use_pae () ? 0x200000 : 0x400000;

      // Free each PT
      for (size_t i = 0; i < 0x1000 / pte_size && offset < 0xf0000000;
           ++i, offset += coverage)
        {

          // Free PT
          auto pde = paging::pde (page.ptr < char *> () + i * pte_size);

          if (pde.present ())
            {

              // If unique, free
              if (paging::puser (pde.address ()) == 1)
                {
                  free_pt (pde.address (), offset);
                }

              // Otherwise, decrease reference count
              else
                {
                  paging::pfree (pde.address ());
                }
            }
        }

      // Release PD
      paging::pfree (addr);
    };

    // Frees a page directory pointer table
    auto free_pdpt = [&] (uint64_t addr, size_t offset)
    {

      // Map PDPT
      smart_mapping page = addr;

      // Free each PD
      for (size_t i = 0; i < 4 && offset < 0xf0000000;
           ++i, offset += 0x40000000)
        {

          // Free PD
          auto pdpte = paging::pdpte (page.ptr < char *> () + i * 8);

          if (pdpte.present () && paging::puser (pdpte.address ()) == 1)
            {

              // If unique, free
              if (paging::puser (pdpte.address ()) == 1)
                {
                  free_pd (pdpte.address (), offset);
                }

              // Otherwise, decrease reference count
              else
                {
                  paging::pfree (pdpte.address ());
                }
            }
        }

      // Release PDPT
      paging::pfree (addr);
    };

    // Free VAS
    if (use_pae ())
      {
        free_pdpt (root_paging_table, 0);
      }
    else
      {
        free_pd (root_paging_table, 0);
      }

    initialized = false;
  }

  uint64_t
  vas_t::data () const
  {
    return root_paging_table;
  }

  vas_t::pte_t::pte_t (void *addr, void *pg)
    : paging::pte (addr), page (pg)
  {}

  vas_t::pte_t::~pte_t ()
  {

    // (Destruct and) Delete the object
    delete static_cast < smart_mapping *> (page);
  }

  uint64_t
  init ()
  {

    // Is it already initialized? Then return previous results
    static bool inited;
    if (inited)
      {

        // Load the VAS
        paging::flush (pager_vas_info.root_paging_table);
        return pager_vas_info.root_paging_table;
      }

    // PTE size
    size_t pte_size = use_pae () ? 8 : 4;

    // smart_mapping and friends aren't working yet, use global mapping
    char *page = static_cast < char *> (paging::kpmap (0, false, false));

    // PT count
    size_t pt_count;

    // PD count
    size_t pd_count;

    // PDPT count
    size_t pdpt_count;

    // Set PT, PD and PDPT counts
    if (use_pae ())
      {
        pt_count = 1920;
        pd_count = 4;
        pdpt_count = 1;
      }
    else
      {
        pt_count = 960;
        pd_count = 1;
        pdpt_count = 0;
      }

    // PT addresses
    uint64_t pts[pt_count];

    // PD addresses
    uint64_t pds[pd_count];

    // PDPT addresses
    uint64_t pdpts[pdpt_count];

    // Nullifies a page
    auto nullify = [&] (uint64_t addr)
    {
      paging::kpmap (page, addr, true, false);
      return memset (page, 0, 0x1000);
    };

    // Initializes a PT
    auto init_pt = [&] (uint32_t offset, uint64_t addr)
    {

      // Add to list
      pts[offset / (0x1000 * (0x1000 / pte_size))] = addr;

      // Fill with zeros and unset all present bits as a side effect
      nullify (addr);
    };

    // Initializes a PD
    auto init_pd = [&] (uint32_t offset, uint64_t addr)
    {

      // Map PD
      paging::kpmap (page, addr, true, false);

      // Add to list
      pds[offset / (0x1000 * (0x1000 / pte_size)
                    * (0x1000 / pte_size))] = addr;

      // Queue of PTs to create (without this, this function will be much
      // slow due to much more use of `invlpg' instruction)
      struct
      {
        uint32_t offset;
        uint64_t addr;
      } queue[0x1000 / pte_size];

      // Initialize PDEs
      size_t i = 0;
      for (; i < 0x1000 / pte_size; ++i)
        {
          if (offset >= 0xf0000000)
            {
              break;
            }

          // Add to queue
          queue[i] = { offset, paging::palloc () };

          // Panic if allocation failed (no need for releasing memory, the
          // system is going to halt)
          if (errno == EAGAIN)
            {
              panic ("error: insufficient memory, not enough memory for "
                     "initializing");
            }

          // Fill PDE
          paging::pde (page + i * pte_size).init (queue[i].addr);
          offset += 0x1000 * (0x1000 / pte_size);
        }

      // If we're in supervisor's area, map the kernel
      if (offset >= 0xf0000000)
        {
          memcpy (page + i * pte_size, pd_ + i * pte_size,
                  0x1000 - i * pte_size);
        }

      // Clear queue
      for (size_t j = 0; j < i; ++j)
        {
          init_pt (queue[j].offset, queue[j].addr);
        }
    };

    auto init_pdpt = [&] (uint32_t offset, uint64_t addr)
    {
      paging::kpmap (page, addr, true, false);
      pdpts[0] = addr; // Multiple PDPTs aren't available in i386

      struct
      {
        uint32_t offset;
        uint64_t addr;
      } queue[4];

      // Fill PDPT
      for (size_t i = 0; i < 4; ++i)
        {

          // Break if in kernel's area
          if (offset >= 0xf0000000)
            {
              break;
            }

          // Add to queue
          queue[i] = { offset, paging::palloc () };

          // Panic if allocation failed (no need for releasing memory, the
          // system is going to halt)
          if (errno == EAGAIN)
            {
              panic ("error: insufficient memory, not enough memory for "
                     "initializing");
            }

          // Fill PDPTE
          paging::pdpte (page + i * 8).init (queue[i].addr);
          offset += 0x40000000;
        }

      // Clean queue
      for (size_t i = 0; i < 4; ++i)
        {
          init_pd (queue[i].offset, queue[i].addr);
        }
    };

    // Allocate root paging structure
    pager_vas_info_t vas_info;
    vas_info.root_paging_table = paging::palloc ();

    // Panic if allocation failed (no need for releasing memory, the
    // system is going to halt)
    if (errno == EAGAIN)
      {
        panic ("error: insufficient memory, not enough memory for "
               "initializing");
      }

    // Initialize VAS
    if (use_pae ())
      {
        init_pdpt (0, vas_info.root_paging_table);
      }
    else
      {
        init_pd (0, vas_info.root_paging_table);
      }

    // Where to map next pages?
    size_t page_i = 0;

    // Map PDPTs
    vas_info.pdpt = reinterpret_cast < char *> (page_i * 0x1000);
    for (size_t i = 0; i < pdpt_count; ++i, ++page_i)
      {

        // Map PT if required
        if (page_i % (0x1000 / pte_size) == 0)
          {
            paging::kpmap (page, pts[page_i / (0x1000 / pte_size)],
                           true, false);
          }

        paging::pte (page + (page_i % (0x1000 / pte_size)) * pte_size)
          .init (pdpts[i], true, false);
      }

    // Map PDs
    vas_info.pd = reinterpret_cast < char *> (page_i * 0x1000);
    for (size_t i = 0; i < pd_count; ++i, ++page_i)
      {

        // Map PT if required
        if (page_i % (0x1000 / pte_size) == 0)
          {
            paging::kpmap (page, pts[page_i / (0x1000 / pte_size)],
                           true, false);
          }

        paging::pte (page + (page_i % (0x1000 / pte_size)) * pte_size)
          .init (pds[i], true, false);
      }

    // Map PTs
    vas_info.pt = reinterpret_cast < char *> (page_i * 0x1000);
    for (size_t i = 0; i < pt_count; ++i, ++page_i)
      {

        // Map the PT where we're mapping to kernel area
        if (page_i % (0x1000 / pte_size) == 0)
          {
            paging::kpmap (page, pts[page_i / (0x1000 / pte_size)],
                           true, false);
          }

        paging::pte (page + (page_i % (0x1000 / pte_size)) * pte_size)
          .init (pts[i], true, false);
      }

    // We're done with the page allocated from kernel space, deallocate it
    paging::kpunmap (page);

    // Load the new VAS
    paging::flush (vas_info.root_paging_table);

    // How many free pages do we have in the new VAS?
    size_t pages_avail = 0xf0000000 / 0x1000 - page_i;

    // How many pages should we use for mapping data?
    size_t data_pg_count = (pages_avail * 0x1000
                            / (0x1000 + sizeof (mapping_t)));

    // How many pages should we use for holding data about used and cached
    // mappings?
    size_t mapping_pg_count = pages_avail - data_pg_count;
    auto mappings = reinterpret_cast < mapping_t *> (page_i * 0x1000);

    // We don't have any mapping yet
    vas_info.used_mappings = nullptr;
    vas_info.cached_mappings = nullptr;
    vas_info.data_offset = (reinterpret_cast < void *>
                            ((page_i + mapping_pg_count) * 0x1000));
    vas_info.mappings = mappings;
    vas_info.max_mapping_count = data_pg_count;
    vas_info.mapping_create_index = 0;

    // Allocate and map pages for holding metadata
    for (size_t i = 0; i < mapping_pg_count; ++i, ++page_i)
      {
        paging::pte (vas_info.pt + page_i * pte_size)
          .init (paging::palloc (), true, false);
      }

    // We've modified possibly thousands of PTEs, so reloading `%cr3' is
    // cheaper than executing lots of `invlpg'
    asm volatile ("movl %%cr3, %%eax;"
                  "movl %%eax, %%cr3;"
                  :
                  :
                  : "eax", "memory");

    // Write to global variables and return
    pager_vas_info = vas_info;
    inited = true;
    return vas_info.root_paging_table;
  }
}
