/*
idt.hpp - IDT related functions and structures

This file is a part of Fiku.
Copyright (C) 2021 Akib Azmain Turja

Fiku is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Fiku is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

Authors: Akib Azmain Turja
*/

#pragma once

#include <stdint.h>
#include <stddef.h>

namespace idt
{
  struct gate
  {
  private:

    // The lower 16 bit of offset (aka pointer)
    uint16_t offset_low_:16;

    // Segment selector
    uint16_t selector_:16;

    // Unused 8 bits
    uint8_t zero_:8;

    // Type of the gate
    uint8_t type_:4;

    // Storage segment
    bool storage_segment_:1;

    // Minimum privilege level required to trigger
    uint8_t privilege_level_:2;

    // Whether the interrupt is available
    bool present_:1;

    // Higher 16 bit of offset
    uint16_t offset_high_:16;

  public:

    /**
     * @brief Constructs a null IDT gate
     */
    constexpr
    gate ()
      : gate (0, 0, 0, 0, false)
    {}

    /**
     * @brief Constructs an IDT gate
     *
     * @details @rst
     *
     * This constructor construct a IDT gate with given parameters.
     * For details of parameters, see Intel manual.
     *
     * @endrst
     *
     * @param offset offset to interrupt handler
     * @param selector code
     * :ref:`segment selector <definition-segment_selector>`
     * @param type type of the gate
     * @param privilege_level minimum privilege level required to
     * trigger the interrupt
     * @param present whether the interrupt is present
     * @param storage_segment storage segment
     */
    constexpr
    gate (uint32_t offset, uint16_t selector, uint8_t type,
          uint8_t privilege_level, bool present = true,
          bool storage_segment = false)
      : offset_low_ (0), selector_ (selector), zero_ (0), type_ (type),
        storage_segment_ (storage_segment),
        privilege_level_ (privilege_level), present_ (present),
        offset_high_ (0)
    {
      this->offset (offset);
    }

    /**
     * @brief Returns the offset of the gate
     *
     * @return the offset of the gate
     */
    constexpr uint32_t
    offset () const
    {
      return (static_cast < uint32_t > (offset_high_) << 16
              | static_cast < uint32_t > (offset_low_));
    }

    /**
     * @brief Sets the offset of the gate
     */
    constexpr void
    offset (uint32_t value)
    {
      offset_low_ = value & 0xffff;
      offset_high_ = value >> 16;
    }

    /**
     * @brief Returns the segment selector of the gate
     *
     * @return the segment selector of the gate
     */
    constexpr uint16_t
    selector () const
    {
      return selector_;
    }

    /**
     * @brief Sets the segment selector of the gate
     */
    constexpr void
    selector (uint16_t value)
    {
      selector_ = value;
    }

    /**
     * @brief Returns the type of the gate
     *
     * @return the type of the gate
     */
    constexpr uint8_t
    type () const
    {
      return type_;
    }

    /**
     * @brief Sets the type of the gate
     */
    constexpr void
    type (uint8_t value)
    {
      type_ = value;
    }

    /**
     * @brief Returns the segment storage of the gate
     *
     * @return the segment storage of the gate
     */
    constexpr bool
    segment_storage () const
    {
      return storage_segment_;
    }

    /**
     * @brief Sets the segment storage of the gate
     */
    constexpr void
    segment_storage (bool value)
    {
      storage_segment_ = value;
    }

    /**
     * @brief Returns the privilege level of the gate
     *
     * @return the privilege level of the gate
     */
    constexpr uint8_t
    privilege_level () const
    {
      return privilege_level_;
    }

    /**
     * @brief Sets the privilege level of the gate
     */
    constexpr void
    privilege_level (uint8_t value)
    {
      privilege_level_ = value & 0b11;
    }

    /**
     * @brief Returns whether the interrupt is present
     *
     * @return whether the interrupt is present
     */
    constexpr bool
    present () const
    {
      return present_;
    }

    /**
     * @brief Sets whether the interrupt is present
     */
    constexpr void
    present (bool value)
    {
      present_ = value;
    }
  }
#ifndef DOXYGEN_DOCUMENTATION_BUILD
    __attribute__((packed))
#endif
  ;

  /**
   * @brief Flushes the given IDT
   *
   * @param table pointer to the first gate of the IDT
   * @param count size of the table
   */
  void
  flush (const gate * table, size_t count);

  /**
   * @brief Sets up an IDT
   *
   * @details @rst
   *
   * This function sets up and loads an IDT with 256 entries. If any error
   * occurs, ``nullptr`` is returned.
   *
   * .. warning::
   *
   *      Calling this function will overwrite any IDT returned previously
   *      by this function. Be sure to backup the previous IDT before
   *      calling this function again!
   *
   * @endrst
   *
   * @param flush whether to flush the IDT
   * @param pic whether to initialize the PIC
   * @return pointer to the IDT
   */
  gate *
  init (bool flush = true, bool pic = true);

  /**
   * @brief Structure for passing information to interrupt handler
   */
  struct registers_t
  {

    /**
     * @brief ``cs`` register, holds code segment selector
     */
    uint16_t cs;

    /**
     * @brief ``ds`` register, holds data segment selector
     */
    uint16_t ds;

    /**
     * @brief ``eip`` register, aka extended instruction pointer
     */
    uint32_t eip;

    /**
     * @brief ``eflags`` register, aka extended flags
     */
    uint32_t eflags;

    // General purpose registers
    uint32_t eax;
    uint32_t ebx;
    uint32_t ecx;
    uint32_t edx;
    uint32_t esp;
    uint32_t ebp;
    uint32_t esi;
    uint32_t edi;

    // Is EOI sent to the PIC?
    bool eoi_sent;
  };

  /**
   * @brief Typedef for interrupt handlers
   */
  typedef void (*interrupt_handler) (uint8_t interrupt, uint8_t code,
                                     registers_t &registers);

  /**
   * @brief Sets the handler for the given interrupt
   *
   * @details @rst
   *
   * This function sets the halder of given interrupt. The handler must
   * take three parameter: interrupt number, error code and reference to a
   * structure holding registers. Error code is zero when the interrupt
   * doesn't push an error code.
   *
   * .. warning::
   *
   *      Modifing the registers may crash the system. Be careful!
   *
   * @endrst
   *
   * @param interrupt interrupt number
   * @param handler handler function takeing interrupt number, error code
   * and registers as parameter
   */
  void
  set_handler (uint8_t interrupt, interrupt_handler handler);

  /**
   * @brief Returns the handler of the given interupt
   *
   * @return the handler of the given interupt
   */
  interrupt_handler
  get_handler (uint8_t interrupt);
}
