/*
iterator.impl.hpp - C++ iterator implementation

This file is a part of Fiku.
Copyright (C) 2021 Akib Azmain Turja

Fiku is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Fiku is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

Authors: Akib Azmain Turja
*/

#include <stddef.h>

namespace fiku::internal_
{

    /**
     * @brief Template class for iterators
     *
     * @tparam array_t type of the container, maybe const-qualified
     * @tparam element_t type of element, maybe const-qualified
     */
  template <class array_t, class element_t, class pointer_t = element_t *,
            class reference_t = element_t &>
  class iterator
  {
  public:

    /**
     * @brief Pointer to the container
     *
     */
    array_t *const array;

    /**
     * @brief Index of element pointed by the iterator
     *
     */
    size_t index;

  public:

    /**
     * @brief Defines template parameter element_t as value_type
     *
     */
    typedef element_t value_type;

    /**
     * @brief Defines type of pointer to element as pointer
     *
     */
    typedef pointer_t pointer;

    /**
     * @brief Defines type of constant pointer to element as
     * const_pointer
     *
     */
    typedef const pointer_t const_pointer;

    /**
     * @brief Defines reference to element as reference
     *
     */
    typedef reference_t reference;

    /**
     * @brief Defines constant reference to element as
     * const_reference
     *
     */
    typedef const reference_t const_reference;

    /**
     * @brief Defines ``ptrdiff_t`` as difference_type
     *
     */
    typedef ptrdiff_t difference_type;

    /**
     * @brief Constructs a new iterator
     *
     * @param array pointer to the list
     * @param index initial index of the element
     */
    constexpr
    iterator (array_t * array, size_t index):
      array (array),
      index (index)
    {
    }

    /**
     * @brief Dereferences the iterator
     *
     * @return reference to element
     */
    constexpr reference
    operator * ()
    {
      return (*array)[index];
    }

    /**
     * @brief Returns the pointer to the element
     *
     * @return pointer to the element
     */
    constexpr pointer
    operator -> ()
    {
      return &(*array)[index];
    }

    /**
     * @brief Increments the iterator
     *
     * @return reference to the iterator after incrementing
     */
    constexpr iterator &
    operator ++ ()
    {
      ++index;
      return *this;
    }

    /**
     * @brief Post-increments the iterator
     *
     * @return reference to the iterator before incrementing
     */
    constexpr iterator
    operator ++ (int)
    {
      iterator
        copy = *this;
      index++;
      return copy;
    }

    /**
     * @brief Returns an iterator which is advanced by ``n``
     * positions
     *
     * @param n count of elements to advance
     * @return iterator which is advanced by ``n`` positions
     */
    constexpr iterator
    operator + (difference_type n) const
    {
      return
        iterator (array, index + n);
    }

    /**
     * @brief Advances the iterator ``n`` positions
     *
     * @param n count of elements to advance
     * @return reference to the iterator
     */
    constexpr iterator &
    operator += (difference_type n)
    {
      index += n;
      return *this;
    }

    /**
     * @brief Decrements the iterator
     *
     * @return reference to the iterator after decrementing
     */
    constexpr iterator &
    operator -- ()
    {
      --index;
      return *this;
    }

    /**
     * @brief Post-decrements the iterator
     *
     * @return reference to the iterator before decrementing
     */
    constexpr iterator
    operator -- (int)
    {
      iterator
        copy = *this;
      index--;
      return copy;
    }

    /**
     * @brief Returns an iterator which is ``n`` positions behind
     *
     * @param n count of elements to advance
     * @return iterator which is ``n`` positions behind
     */
    constexpr iterator
    operator - (difference_type n) const
    {
      return
        iterator (array, index - n);
    }

    /**
     * @brief Returns difference between two iterators
     *
     * @param other another iterator
     * @return difference between two iterators
     */
    constexpr difference_type
    operator - (iterator other) const
    {
      return
        index -
        other.
        index;
    }

    /**
     * @brief Retreats the iterator ``n`` positions
     *
     * @param n count of elements to retreat
     * @return reference to the iterator
     */
    constexpr iterator &
    operator -= (difference_type n)
    {
      index -= n;
      return *this;
    }

    /**
     * @brief Checks whether another iterator equals to the iterator
     *
     * @param other another iterator to compare
     * @return whether another iterator equals to the iterator
     */
    constexpr bool
    operator == (const iterator & other)
    {
      return array == other.array && index == other.index;
    }

    /**
     * @brief Checks whether another iterator not equals to the
     * iterator
     *
     * @param other another iterator to compare
     * @return whether another iterator not equals to the iterator
     */
    constexpr bool
    operator != (const iterator & other)
    {
      return !(*this == other);
    }

    /**
     * @brief Checks whether another iterator is more advanced than
     * the iterator
     *
     * @param other another iterator to compare
     * @return whether another iterator is more advanced than the
     * iterator
     */
    constexpr bool
    operator > (const iterator & other)
    {
      return array == other.array && index > other.index;
    }

    /**
     * @brief Checks whether another iterator is more advanced or
     * equals to the iterator
     *
     * @param other another iterator to compare
     * @return whether another iterator is more advanced or equals
     * to the iterator
     */
    constexpr bool
    operator >= (const iterator & other)
    {
      return array == other.array && index >= other.index;
    }

    /**
     * @brief Checks whether the iterator is more advanced than
     * another iterator
     *
     * @param other another iterator to compare
     * @return whether another iterator is more advanced than the
     * iterator
     */
    constexpr bool
    operator < (const iterator & other)
    {
      return array == other.array && index > other.index;
    }

    /**
     * @brief Checks whether the iterator is more advanced or equals
     * to another iterator
     *
     * @param other another iterator to compare
     * @return whether another iterator is more advanced or equals
     * to the iterator
     */
    constexpr bool
    operator <= (const iterator & other)
    {
      return array == other.array && index <= other.index;
    }
  };

  /**
   * @brief Template class for reverse iterators
   *
   * @tparam array_t type of the container, maybe const-qualified
   * @tparam element_t type of elements, maybe const-qualified
   */
  template < class iterator_t,
             class element_t = typename iterator_t::value_type,
             class pointer_t = typename iterator_t::pointer,
             class reference_t = typename iterator_t::reference >
  class reverse_iterator
  {
  private:

    /**
     * @brief Actual iterator
     *
     */
    iterator_t it;

  public:

    /**
     * @brief Defines template parameter element_t as value_type
     *
     */
    typedef element_t value_type;

    /**
     * @brief Defines type of pointer to element as pointer
     *
     */
    typedef pointer_t pointer;

    /**
     * @brief Defines type of constant pointer to element as
     * const_pointer
     *
     */
    typedef const pointer_t const_pointer;

    /**
     * @brief Defines reference to element as reference
     *
     */
    typedef reference_t reference;

    /**
     * @brief Defines constant reference to element as
     * const_reference
     *
     */
    typedef const reference_t const_reference;

    /**
     * @brief Defines ``ptrdiff_t`` as difference_type
     *
     */
    typedef ptrdiff_t difference_type;

    /**
     * @brief Constructs a new iterator
     *
     * @param array pointer to the list
     * @param index initial index of the element
     */
    constexpr
    reverse_iterator (iterator_t it):
      it (--it)
    {
    }

    /**
     * @brief Dereferences the iterator
     *
     * @return reference to element
     */
    constexpr reference
    operator * ()
    {
      return *it;
    }

    /**
     * @brief Returns the pointer to the element
     *
     * @return pointer to the element
     */
    constexpr pointer
    operator -> ()
    {
      return it.operator -> ();
    }

    /**
     * @brief Increments the iterator
     *
     * @return reference to the iterator after incrementing
     */
    constexpr reverse_iterator &
    operator ++ ()
    {
      --it;
      return *this;
    }

    /**
     * @brief Post-increments the iterator
     *
     * @return reference to the iterator before incrementing
     */
    constexpr reverse_iterator
    operator ++ (int)
    {
      iterator
        copy = *this;
      it--;
      return copy;
    }

    /**
     * @brief Returns an iterator which is advanced by ``n``
     * positions
     *
     * @param n count of elements to advance
     * @return iterator which is advanced by ``n`` positions
     */
    constexpr reverse_iterator
    operator + (difference_type n) const
    {
      return
        reverse_iterator (it - n);
    }

    /**
     * @brief Advances the iterator ``n`` positions
     *
     * @param n count of elements to advance
     * @return reference to the iterator
     */
    constexpr reverse_iterator &
    operator += (difference_type n)
    {
      it -= n;
      return *this;
    }

    /**
     * @brief Decrements the iterator
     *
     * @return reference to the iterator after decrementing
     */
    constexpr reverse_iterator &
    operator -- ()
    {
      ++it;
      return *this;
    }

    /**
     * @brief Post-decrements the iterator
     *
     * @return reference to the iterator before decrementing
     */
    constexpr reverse_iterator
    operator -- (int)
    {
      iterator
        copy = *this;
      it++;
      return copy;
    }

    /**
     * @brief Returns an iterator which is ``n`` positions behind
     *
     * @param n count of elements to advance
     * @return iterator which is ``n`` positions behind
     */
    constexpr reverse_iterator
    operator - (difference_type n) const
    {
      return
        reverse_iterator (it + n);
    }

    /**
     * @brief Returns difference between two iterators
     *
     * @param other another iterator
     * @return difference between two iterators
     */
    constexpr difference_type
    operator - (reverse_iterator other) const
    {
      return other.it - it;
    }

    /**
     * @brief Retreats the iterator ``n`` positions
     *
     * @param n count of elements to retreat
     * @return reference to the iterator
     */
    constexpr reverse_iterator &
    operator -= (difference_type n)
    {
      it += n;
      return *this;
    }

    /**
     * @brief Checks whether another iterator equals to the iterator
     *
     * @param other another iterator to compare
     * @return whether another iterator equals to the iterator
     */
    constexpr bool
    operator == (const reverse_iterator &other)
    {
      return it == other.it;
    }

    /**
     * @brief Checks whether another iterator not equals to the
     * iterator
     *
     * @param other another iterator to compare
     * @return whether another iterator not equals to the iterator
     */
    constexpr bool
    operator != (const reverse_iterator &other)
    {
      return it != other.it;
    }

    /**
     * @brief Checks whether another iterator is more advanced than
     * the iterator
     *
     * @param other another iterator to compare
     * @return whether another iterator is more advanced than the
     * iterator
     */
    constexpr bool
    operator > (const reverse_iterator &other)
    {
      return it < other.it;
    }

    /**
     * @brief Checks whether another iterator is more advanced or
     * equals to the iterator
     *
     * @param other another iterator to compare
     * @return whether another iterator is more advanced or equals
     * to the iter
     * ator
     */
    constexpr bool
    operator >= (const reverse_iterator &other)
    {
      return it <= other.it;
    }

    /**
     * @brief Checks whether the iterator is more advanced than
     * another iterator
     *
     * @param other another iterator to compare
     * @return whether another iterator is more advanced than the
     * iterator
     */
    constexpr bool
    operator < (const reverse_iterator &other)
    {
      return it > other.it;
    }

    /**
     * @brief Checks whether the iterator is more advanced or equals
     * to another iterator
     *
     * @param other another iterator to compare
     * @return whether another iterator is more advanced or equals
     * to the iterator
     */
    constexpr bool
    operator <= (const reverse_iterator &other)
    {
      return it >= other.it;
    }
  };

  template < class array_t, class element_t >  iterator < array_t, element_t >
  operator + (ptrdiff_t n, iterator < array_t,  element_t > it)
  {
    return it + n;
  }

  template < class array_t, class element_t > reverse_iterator < array_t,
                                                                 element_t >
  operator + (ptrdiff_t n, reverse_iterator < array_t, element_t > it)
  {
    return it - n;
  }
}
