/*
multiboot2.hpp - Functions and classes for parsing Multiboot2 boot
information

This file is a part of Fiku.
Copyright (C) 2021 Akib Azmain Turja

Fiku is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Fiku is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

Authors: Akib Azmain Turja
*/

#pragma once

#include <stdint.h>
#include <stddef.h>
#include <fiku/iterator.impl.hpp>

namespace multiboot2
{
  class boot_info
  {
  private:

    // Pointer to boot info
    const void *data;

  public:
    struct tag_t
    {

      /**
       * @brief Tag type
       */
      uint32_t type;

      /**
       * @brief Tag size
       */
      uint32_t size;

      template < class tag > static tag &
      cast (void *ptr)
      {
        return *static_cast < tag *>((tag::type == (static_cast < tag_t *>
                                                    (ptr)->type))
                                     ? ptr : nullptr);
      }

      template < class tag > static const tag &
      cast (const void *ptr)
      {
        return *static_cast < const tag *> ((tag::type ==
                                             (static_cast < const tag_t *>
                                              (ptr)->type))
                                            ? ptr : nullptr);
      }

      template < class tag > const tag &
      cast () const
      {
        return cast < tag > (this);
      }
    };

    struct cmd_t
    {

      /**
       * @brief Tag type
       */
      constexpr static const uint32_t type = 1;

      /**
       * @brief Tag type
       */
      uint32_t type__ = type;

      /**
       * @brief Tag size
       */
      uint32_t size;

      /**
       * @brief String holding command line, must be treated as a
       * pointer
       */
      char cmd[0];
    };

    struct bootloader_t
    {

      /**
       * @brief Tag type
       */
      constexpr static const uint32_t type = 2;

      /**
       * @brief Tag type
       */
      uint32_t type__ = type;

      /**
       * @brief Tag size
       */
      uint32_t size;

      /**
       * @brief String holding bootloader name, must be treated as
       * a pointer
       */
      char name[0];
    };

    struct module_t
    {

      /**
       * @brief Tag type
       */
      constexpr static const uint32_t type = 3;

      /**
       * @brief Tag type
       */
      uint32_t type__ = type;

      /**
       * @brief Tag size
       */
      uint32_t size;

      /**
       * @brief Module start
       */
      uint32_t start;

      /**
       * @brief Module end
       */
      uint32_t end;

      char args[0];
    };

    struct meminfo_t
    {

      /**
       * @brief Tag type
       */
      constexpr static const uint32_t type = 4;

      /**
       * @brief Tag type
       */
      uint32_t type__ = type;

      /**
       * @brief Tag size
       */
      uint32_t size;

      /**
       * @brief Lower memory
       */
      uint32_t mem_lower;

      /**
       * @brief Upper memory
       */
      uint32_t mem_upper;
    };

    struct memmap_t
    {

      /**
       * @brief Tag type
       */
      constexpr static const uint32_t type = 6;

      /**
       * @brief Tag type
       */
      uint32_t type__ = type;

      /**
       * @brief Tag size
       */
      uint32_t size;

      /**
       * @brief Size of each entry
       */
      uint32_t entry_size;

      /**
       * @brief Entry version
       */
      uint32_t entry_version;

      struct entry_t
      {
        uint64_t base_addr;
        uint64_t length;
        uint32_t type;
      };

      /**
       * @brief Entries
       */
      entry_t entries[1];

      const entry_t & operator [] (size_t index) const
      {
        return *(reinterpret_cast < const entry_t *>
                 (reinterpret_cast < const char *> (entries)
                  + entry_size * index));
      }

      size_t entry_count () const
      {
        return ((size - offsetof (multiboot2::boot_info::memmap_t,
                                  entries)) / entry_size);
      }
    };

    typedef fiku::internal_::iterator < const boot_info,
                                        const tag_t > iterator;
    typedef iterator const_iterator;
    typedef fiku::internal_::reverse_iterator < iterator
                                                > reverse_iterator;
    typedef reverse_iterator const_reverse_iterator;

    boot_info (const void *structure);

    size_t size () const;

    size_t tag_count () const;

    const_iterator begin () const;
    const_iterator cbegin () const;
    const_reverse_iterator rbegin () const;
    const_reverse_iterator crbegin () const;
    const_iterator end () const;
    const_iterator cend () const;
    const_reverse_iterator rend () const;
    const_reverse_iterator crend () const;

    const tag_t &operator [] (size_t index) const;
  };
}
