/*
paging.hpp - Paging related definitions for i386 architecture

This file is a part of Fiku.
Copyright (C) 2021 Akib Azmain Turja

Fiku is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Fiku is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

Authors: Akib Azmain Turja
*/

#pragma once

#include <stdint.h>
#include <stddef.h>

namespace paging
{
  class pde_pte_base
  {
  protected:
    void *page_struct;

  public:
    pde_pte_base (void *page_struct);

    pde_pte_base (const pde_pte_base &) = delete;

    virtual
    ~pde_pte_base () = default;

    void
    init ();

    void
    init (uint32_t physaddr, bool writable = true, bool executable = true,
          bool privileged = true, bool present = true);

    uint64_t
    address () const;

    void
    address (uint64_t value);

    bool
    writable () const;

    void
    writable (bool value);

    bool
    executable () const;

    void
    executable (bool value);

    bool
    privileged () const;

    void
    privileged (bool value);

    bool
    accessed () const;

    void
    accessed (bool value);

    bool
    written () const;

    void
    written (bool value);

    bool
    global () const;

    void
    global (bool value);

    bool
    present () const;

    void
    present (bool value);

    bool
    cache () const;

    void
    cache (bool value);

    bool
    writet () const;

    void
    writet (bool value);

    pde_pte_base &
    operator = (const pde_pte_base &) = delete;
  };

  class pdpte : public pde_pte_base
  {
  public:
    pdpte (void *page_struct);

    virtual
    ~pdpte () = default;

    void
    init ();

    void
    init (uint32_t physaddr, bool present = true);

    bool
    writable () const = delete;

    void
    writable (bool value) = delete;

    bool
    executable () const = delete;

    void
    executable (bool value) = delete;

    bool
    privileged () const = delete;

    void
    privileged (bool value) = delete;

    bool
    accessed () const = delete;

    void
    accessed (bool value) = delete;

    bool
    written () const = delete;

    void
    written (bool value) = delete;

    bool
    global () const = delete;

    void
    global (bool value) = delete;
  };

  class pde : public pde_pte_base
  {
  public:
    pde (void *page_struct);

    virtual
    ~pde () = default;

    void
    init ();

    void
    init (uint32_t physaddr, bool writable = true, bool executable = true,
          bool privileged = true, bool size = false, bool present = true,
          bool swapped = false);

    uint64_t
    address () const;

    void
    address (uint64_t value);

    bool
    written () const;

    void
    written (bool value);

    bool
    swapped () const;

    void
    swapped (bool value);

    bool
    size () const;

    void
    size (bool value);
  };

  class pte : public pde_pte_base
  {
  public:
    pte (void *page_struct);

    virtual
    ~pte () = default;

    void
    init ();

    void
    init (uint32_t physaddr, bool writable = true, bool executable = true,
          bool privileged = true, bool present = true,
          bool swapped = false);

    bool
    swapped () const;

    void
    swapped (bool value);
  };

  /* Invalidates TLB cache for the page associated with given address
   *
   * Note: it reloads %cr3 if ``invlpg'' is not available, invalidating
   * whole TLB. Set ``no_reload'' to true to avoid this.
   *
   * address - address to virtual memory
   *
   * no_reload - set to true to disable reloading %cr3
   */
  void
  invlpg (void *, bool no_reload = true);

  // Flushes paging structure
  void
  flush (size_t structure);

  // Initializes a paging structure
  void *
  init (const void *boot_info, bool flush = true);

  /* Allocates a 4 KiB page
   *
   * pts - pointer to page tables (must be virtual and contiguous)
   *
   * count - count of pages to allocate
   *
   * alignment - alignment of page
   *
   * size - size of page tables in bytes
   *
   * writable - whether the page is writable
   *
   * executable - whether the page is executable
   *
   * privileged - whether the page can only be access in supervisor mode
   *
   * global - whether the page is global
   *
   * Returns the physical address of the page
   */
  uint64_t
  palloc ();

  // Increments reference count of a page
  void
  palloc (uint64_t addr);

  // Deallocates a 4 KiB page
  void
  pfree (uint64_t addr);

  // Returns whether a page is used
  bool
  pused (uint64_t addr);

  // Returns the user count of a page (i.e VASs count refering to the page)
  size_t
  puser (uint64_t addr);

  // Allocates a 4 KiB page in kernel address space
  void *
  kpalloc (size_t count, size_t alignment, bool writable, bool executable);

  // Maps given page to given physical address
  void
  kpmap (void *page, uint64_t addr, bool writable, bool executable);

  // Maps a free page to given physical address
  void *
  kpmap (uint64_t addr, bool writable, bool executable);

  // Returns the physical address where the page is mapped to
  uint64_t
  kpmapping (void *page);

  // Unmaps (but doesn't free) a page
  void
  kpunmap (void *page);

  // Return the PTE associated to the given page
  pte
  kpte (void *page);

  // Deallocates a 4 KiB page in kernel address space
  void
  kpfree (void *page, size_t count);
}
