/*
proc.hpp - Process management and thread scheduling functions

This file is a part of Fiku.
Copyright (C) 2021 Akib Azmain Turja

Fiku is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Fiku is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

Authors: Akib Azmain Turja
*/

#pragma once

#include <stdint.h>
#include <fiku/idt.hpp>
#include <fiku/timers/base.hpp>

namespace proc
{
  struct thrd_t
  {

    // Is the thread attached to the scheduler? (mustn't be touched)
    bool attached;

    enum
      {
        // The thread is running
        thrd_running,

        // The thread isn't running, but we can run it
        thrd_runnable,

        // The thread is killed and can never be started
        thrd_killed,

        // The thread is waiting for receiving message from a synchronous
        // port
        thrd_waiting_msg_recv,

        // The thread sent a message and is waiting for the delivery to be
        // complete
        thrd_waiting_msg_dlvr
      } thrd_state;

    // Time to live (in milliseconds)
    size_t ttl;

    // CPU time used by the thread
    uintmax_t cpu_time;

    // Virtual Address Space
    uint32_t vas;

    // Register state of the thread
    idt::registers_t *reg_state;

    // Is it a userspace thread?
    bool usr_thrd;

    // Kernel to handler interrupts
    void *kernel_stack;

    // Kernel context after interrupt
    void *kernel_context;

    // Function to call after the thread is detached by the scheduler (set
    // to nullptr to ignore)
    void (*after_detach) (void *);

    // Argument to after_detach function
    void *after_detach_arg;

    // Next thread to execute
    thrd_t *next_thrd;
  };

  // Prepares a thread before adding for the first time
  void
  thrd_prep (thrd_t *thrd);

  // Adds a thread
  void
  thrd_add (thrd_t *thrd);

  // Removes a thread
  void
  thrd_rm (thrd_t *thrd);

  // Returns current thread
  proc::thrd_t *
  thrd_current ();

  // Switches to another thread
  void
  thrd_yield (size_t time_lived, idt::registers_t *regs);

  // Starts the first thread
  __attribute__ ((noreturn)) void
  thrd_start ();

  // Sets the timer to use
  void
  set_timer (timer::basic_timer &timer);

  // Returns the timer being used
  timer::basic_timer &
  get_timer ();

  struct port_t
  {

    // Whether the port has protection
    bool protection;

    // Receiver threads
    thrd_t *receivers;

    // Reference count
    size_t ref_count;
  };

  enum port_priv
    {

      // Send right
      send = 1 << 0,

      // Receive right
      receive = 1 << 1,

      // Right to give the right to send to threads
      give_send = 1 << 2,

      // Right to give the right to receive to threads
      give_recv = 1 << 3,

      // Right to give this right to threads
      give_allow = 1 << 4,
    };

  // Allocates a port
  port_t *
  port_alloc (bool protection);

  // Deallocates a port
  void
  port_free (port_t *port);

  // Gives a thread the right to do certain things with a port (set rights
  // to 0 to revoke any previous rights)
  void
  port_priv (port_t *port, thrd_t *thrd, size_t rights);

  // Send data to a port
  void
  port_send (port_t *port, size_t time_lived, idt::registers_t *regs);

  // Receives data
  idt::registers_t *
  port_recv (port_t *port, size_t time_lived, idt::registers_t *regs);

  // TSS or Task State Segment
  struct tss_t
  {
    uint16_t prev_tss;
    uint16_t reserved0_;
    uint32_t esp0;
    uint16_t ss0;
    uint16_t reserved1_;
    uint32_t esp1;
    uint16_t ss1;
    uint16_t reserved2_;
    uint32_t esp2;
    uint16_t ss2;
    uint16_t reserved3_;
    uint32_t cr3;
    uint32_t eip;
    uint32_t eflags;
    uint32_t eax;
    uint32_t ecx;
    uint32_t edx;
    uint32_t ebx;
    uint32_t esp;
    uint32_t ebp;
    uint32_t esi;
    uint32_t edi;
    uint16_t es;
    uint16_t reserved4_;
    uint16_t cs;
    uint16_t reserved5_;
    uint16_t ss;
    uint16_t reserved6_;
    uint16_t ds;
    uint16_t reserved7_;
    uint16_t fs;
    uint16_t reserved8_;
    uint16_t gs;
    uint16_t reserved9_;
    uint16_t ldt_selector;
    uint16_t reserved10_;
    bool debug_trap:1;
    uint16_t reserved11_:15;
    uint16_t iomap_base;
    uint32_t ssp;
  } __attribute__ ((packed));

  // Return TSS
  tss_t *
  get_tss ();
}
