/*
vas.hpp - Virtual Address Space management functions for i386 architecture

This file is a part of Fiku.
Copyright (C) 2021 Akib Azmain Turja

Fiku is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Fiku is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

Authors: Akib Azmain Turja
*/

/*
 * For the reader, some term used here:
 *
 * VAS - Virtual Address Space
 */

#pragma once

#include <stddef.h>
#include <stdint.h>
#include <fiku/paging.hpp>

namespace vas
{

  // Class for managing non-supervisor VAS
  class vas_t
  {
  private:
    uint64_t root_paging_table;
    bool initialized;

    // Frees the memory used
    void
    _free ();

  public:

    // PTE type returned by `pte'
    class pte_t : public paging::pte
    {
    private:
      void *page;
      pte_t (void *addr, void *pg);

    public:
      pte_t (const pte_t &) = delete;

      pte_t (pte_t &&) = delete;

      ~pte_t ();

      pte_t &
      operator = (const pte_t &) = delete;

      pte_t &
      operator = (pte_t &&) = delete;

      friend class vas_t;
    };

    // Constructs a vas_t object
    vas_t ();

    vas_t (const vas_t &) = delete;

    // Move constructs a vas_t object
    vas_t (vas_t &&other);

    // Destructs vas_t
    ~vas_t ();

    // Maps given page
    void
    map (size_t page, bool writable, bool executable);

    // Maps given page to given physical address
    void
    map (size_t page, uint64_t physaddr, bool writable, bool executable);

    // Returns the mapping of given page
    uint64_t
    mapping (size_t page) const;

    // Returns whether the given page is writable
    bool
    writable (size_t page) const;

    // Returns whether the given page is executable
    bool
    executable (size_t page) const;

    // Touches the given page (i.e swaps the page in if swapped)
    void
    touch (size_t page);

    // Unmaps the given page
    void
    unmap (size_t page, bool free = true);

    // Checks if the page is mapped
    bool
    is_mapped (size_t page) const;

    // Checks if the page associated the given address is mapped
    bool
    is_mapped (void *addr) const;

    // Returns PTE for given page (note: must be freed with delete when
    // done with it, should not be used to change the present bit)
    pte_t *
    pte (size_t page);

    // Clones the VAS
    vas_t
    clone () const;

    // Returns the physical address of root paging table
    uint64_t
    data () const;

    // Converts page to pointer
    static void *
    to_ptr (size_t page);

    // Converts pointer to page
    static size_t
    to_page (void *ptr);

    vas_t &
    operator = (const vas_t &other) = delete;

    // Move assigns a vas_t object to this
    vas_t &
    operator = (vas_t &&other);
  };

  // Loads a VAS, pass nullptr to disable VAS
  void
  load (const vas_t *vas);

  // Initializes VAS class and functions
  uint64_t
  init ();
}
