/*
elf.cpp - ELF loader

This file is a part of Fiku.
Copyright (C) 2021 Akib Azmain Turja

Fiku is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Fiku is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

Authors: Akib Azmain Turja
*/

#include <fiku/loader.hpp>
#include <stdint.h>
#include <stddef.h>
#include <fiku/panic.h>
#include <fiku/vas.hpp>
#include <string.h>
#include <errno.h>
#include <fiku/paging.hpp>
#include <stdio.h>

namespace
{

  // eint = Extended Integer
  // Used for parsing integers with endianness unknown at compile time
  template < size_t size > struct eint
  {
    uint8_t data[size];

    constexpr uintmax_t
    decode (uint8_t endianness) const
    {
      uintmax_t ret = 0;

      for (size_t i = 0; i < size; ++i)
        {
          if (endianness == 1) // Little-endian
            {
              ret |= data[i] << (i * 8);
            }
          else // Big-endian
            {
              ret |= data[i] << ((size - 1 - i) * 8);
            }
        }

      return ret;
    }
  };

#ifdef __x86_64__
#define BITS 64
#else
#define BITS 32
#endif

  // File header
  struct fhdr
  {
    uint8_t magic[4];         // Must be "\x7fELF" without trailing null
    uint8_t bits;             // 1 for 32-bit, 2 for 64-bit
    uint8_t endianess;        // 1 for little-endian, 2 for big
    uint8_t i_version;        // Version in ei_ident, 1 is the latest
    uint8_t osabi;            // Target OS
    uint8_t abi_version;      // ABI version
    uint8_t padding[7];       // Padding (unused)
    eint < 2 > type;          // Object file type
    eint < 2 > machine;       // ISA / Machine type
    eint < 4 > version;       // 1 for original version
    eint < BITS / 8 > entry;  // Entry point
    eint < BITS / 8 > ph_off; // Program header offset
    eint < BITS / 8 > sh_off; // Section header offset
    eint < 4 > flags;         // Unused
    eint < 2 > size;          // sizeof (this)
    eint < 2 > ph_entry_size; // Size of program header entry
    eint < 2 > ph_count;      // Number of entries in program header
    eint < 2 > sh_entry_size; // Size of section header entry
    eint < 2 > sh_count;      // Number of entries in section header
    eint < 2 > shstr_i;       // Index of section containing section names
  };

  // Program Header
  struct phdr
  {
    eint < 4 > type;              // Segment type
#if BITS == 64
    eint < 4 > flags;             // Flags
#endif
    eint < BITS / 8 > file_off;   // Offset of contents
    eint < BITS / 8 > virt_off;   // Address of contents in memory
    eint < BITS / 8 > phys_off;   // Physical offset of contents in memory
    eint < BITS / 8 > file_size;  // Size of segment in file
    eint < BITS / 8 > mem_size;   // Size of segment in memory
#if BITS == 32
    eint < 4 > flags;             // Flags
#endif
    eint < BITS / 8 > alignment;  // Alignment
  };

  bool
  valid (const void *file)
  {
    const fhdr *hdr = static_cast < const fhdr *> (file);

    if (hdr->magic[0] != 0x7f
        || hdr->magic[1] != 'E'
        || hdr->magic[2] != 'L'
        || hdr->magic[3] != 'F'
        || hdr->i_version != 1
        || hdr->version.decode (hdr->endianess) != 1
#if BITS == 32
        || hdr->bits != 1
        || hdr->machine.decode (hdr->endianess) != 0x3
#else
        || hdr->bits != 2
        || hdr->machine.decode (hdr->endianess) != 0x3e
#endif
        || hdr->osabi != 0
        || hdr->abi_version != 0
        || hdr->type.decode (hdr->endianess) != 0x2)
      {
        return false;
      }

    return true;
  }
}

namespace loader
{
  prog_img
  load (const void *file, const void *)
  {
    if (!valid (file))
      {
        panic ("error: invalid executable file\n");
      }

    // The base address
    const char *base = static_cast < const char *> (file);

    // File header
    const fhdr *hdr = static_cast < const fhdr *> (file);

    // Program header table
    const phdr *phdrt = (reinterpret_cast < const phdr *>
                           (base + hdr->ph_off.decode (hdr->endianess)));

    // Count of entries in program header table
    size_t phdrc = hdr->ph_count.decode (hdr->endianess);

    // Switch to the VAS of VAS manager
    vas::init ();

    // VAS holding the program image
    vas::vas_t *img = new vas::vas_t;

    // Make sure no errors occured
    if (!img)
      {
        return { nullptr, 0 };
      }
    else if (errno == EAGAIN)
      {
        delete img;
        return { nullptr, 0 };
      }

    // Iterate over all PHTE
    for (size_t i = 0; i < phdrc; ++i)
      {

        // Check if the segment has to be loaded
        if (phdrt[i].type.decode (hdr->endianess) == 0x1)
          {
            if (phdrt[i].alignment.decode (hdr->endianess) < 0x1000)
              {
                delete img;
                return { nullptr, 0 };
              }

            void *addr = (reinterpret_cast < void *>
                          (phdrt[i].virt_off.decode (hdr->endianess)));
            uintmax_t flags = phdrt[i].flags.decode (hdr->endianess);
            bool writable = flags >> 1 & 1;
            bool executable = flags >> 2 & 1;

            // Allocate memory
            for (size_t j = 0;
                 j * 0x1000 < phdrt[i].mem_size.decode (hdr->endianess);
                 ++j)
              {
                img->map (img->to_page (addr) + j, true, executable);
              }

            paging::flush (img->data ());

            // Copy data
            memset ((reinterpret_cast < void *>
                     (phdrt[i].virt_off.decode (hdr->endianess))),
                    0, phdrt[i].mem_size.decode (hdr->endianess));
            memcpy ((reinterpret_cast < void *>
                     (phdrt[i].virt_off.decode (hdr->endianess))),
                    base + phdrt[i].file_off.decode (hdr->endianess),
                    phdrt[i].file_size.decode (hdr->endianess));

            // Set writing permission
            if (!writable)
              {
                vas::init ();

                for (size_t j = 0;
                     j * 0x1000 < phdrt[i].mem_size.decode (hdr->endianess);
                     ++j)
                  {
                    auto pte = img->pte (img->to_page (addr) + j);
                    pte->writable (writable);
                    delete pte;
                  }
              }
          }
      }

    return { img, static_cast < uintptr_t > (hdr->entry.decode
                                             (hdr->endianess)) };
  }
}
