/*
multiboot2.cpp - Functions for parsing Multiboot2 boot information

This file is a part of Fiku.
Copyright (C) 2021 Akib Azmain Turja

Fiku is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Fiku is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

Authors: Akib Azmain Turja
*/

#include <fiku/multiboot2.hpp>

namespace
{
  struct boot_info_s
  {
    uint32_t size;
    uint32_t zero;
    char tags[0];
  }
  __attribute__((packed));

  const void *
  loop_until (uint32_t tag_type, const void *tags,
                          long max = -1, size_t *count = nullptr)
  {
    const char *offset = static_cast < const char *> (tags);
    size_t loop_count = 0;
    const multiboot2::boot_info::tag_t * tag;

    while ((max >= 0 ? max > static_cast < long > (loop_count) : true)
           && ((tag
                = (reinterpret_cast < const multiboot2::boot_info::tag_t *>
                   (offset)))->type
               != tag_type))
      {
        offset += (tag->size + 7) & ~0b111;
        loop_count++;
      }

    if (count)
      *count = loop_count;

    return offset;
  }
}

namespace multiboot2
{
  boot_info::boot_info (const void *ptr)
    : data (ptr)
  {}

  size_t
  boot_info::size () const
  {
    return static_cast < const boot_info_s *> (data)->size;
  }

  size_t
  boot_info::tag_count () const
  {
    size_t count = 0;

    // Loop until the last tag is reached, then return
    loop_until (0, static_cast < const boot_info_s * > (this->data)->tags,
                -1, &count);
    return count;
  }

  const boot_info::tag_t &
  boot_info::operator [] (size_t index) const
  {
    return *static_cast < const boot_info::tag_t * >
      (loop_until (0, static_cast < const boot_info_s * > (this->data)->tags,
                   index));
  }

  boot_info::const_iterator
  boot_info::begin () const
  {
    return const_iterator (this, 0);
  }

  boot_info::const_iterator
  boot_info::cbegin () const
  {
    return const_iterator (this, 0);
  }

  boot_info::const_iterator
  boot_info::end () const
  {
    return const_iterator (this, tag_count ());
  }

  boot_info::const_iterator
  boot_info::cend () const
  {
    return const_iterator (this, tag_count ());
  }

  boot_info::const_reverse_iterator
  boot_info::rbegin () const
  {
    return const_reverse_iterator (begin ());
  }

  boot_info::const_reverse_iterator
  boot_info::crbegin () const
  {
    return const_reverse_iterator (cbegin ());
  }

  boot_info::const_reverse_iterator
  boot_info::rend () const
  {
    return const_reverse_iterator (end ());
  }

  boot_info::const_reverse_iterator
  boot_info::crend () const
  {
    return const_reverse_iterator (cend ());
  }
}
