/*
new.cpp - Implementation of operator new

This file is a part of Fiku.
Copyright (C) 2021 Akib Azmain Turja

Fiku is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Fiku is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

Authors: Akib Azmain Turja
*/

#include <new.hpp>
#include <errno.h>
#include <stdlib.h>

namespace
{
  gstd::new_handler handler;

  void
  handle_error ()
  {
    if (handler)
      {
        handler ();
      }
    else
      {
        errno = EAGAIN;
      }
  }
}

namespace gstd
{
  new_handler
  get_new_handler () noexcept
  {
    return handler;
  }

  new_handler
  set_new_handler (new_handler handler_func) noexcept
  {
    new_handler prev = handler;
    handler = handler_func;
    return prev;
  }
}


void *
operator new (size_t size)
{
  void *ptr = malloc (size);

  if (!ptr)
    {
      handle_error ();
    }

  return ptr;
}

void *
operator new (size_t size, gstd::align_val_t alignment)
{
  void *ptr = aligned_alloc (static_cast < size_t > (alignment), size);

  if (!ptr)
    {
      handle_error ();
    }

  return ptr;
}

void *
operator new (size_t size, const gstd::nothrow_t &) noexcept
{
  return malloc (size);
}

void *
operator new (size_t size, gstd::align_val_t alignment,
              const gstd::nothrow_t &) noexcept
{
  return aligned_alloc (size, static_cast < size_t > (alignment));
}

void *
operator new (size_t, void *ptr) noexcept
{
  return ptr;
}

void *
operator new[] (size_t size)
{
  void *ptr = malloc (size);

  if (!ptr)
    {
      handle_error ();
    }

  return ptr;
}

void *
operator new[] (size_t size, gstd::align_val_t alignment)
{
  void *ptr = aligned_alloc (static_cast < size_t > (alignment), size);

  if (!ptr)
    {
      handle_error ();
    }

  return ptr;
}

void *
operator new[] (size_t size, const gstd::nothrow_t &) noexcept
{
  return malloc (size);
}

void *
operator new[] (size_t size, gstd::align_val_t alignment,
                const gstd::nothrow_t &) noexcept
{
  return aligned_alloc (size, static_cast < size_t > (alignment));
}

void *
operator new[] (size_t, void *ptr) noexcept
{
  return ptr;
}

void
operator delete (void *ptr) noexcept
{
  free (ptr);
}

void
operator delete (void *ptr, size_t) noexcept
{
  free (ptr);
}

void
operator delete (void *ptr, gstd::align_val_t) noexcept
{
  free (ptr);
}

void
operator delete (void *ptr, size_t, gstd::align_val_t) noexcept
{
  free (ptr);
}

void
operator delete (void *ptr, gstd::nothrow_t &) noexcept
{
  free (ptr);
}

void
operator delete (void *ptr, gstd::align_val_t, gstd::nothrow_t &) noexcept
{
  free (ptr);
}

void
operator delete (void *, void *) noexcept
{}

void
operator delete[] (void *ptr) noexcept
{
  free (ptr);
}

void
operator delete[] (void *ptr, size_t) noexcept
{
  free (ptr);
}

void
operator delete[] (void *ptr, gstd::align_val_t) noexcept
{
  free (ptr);
}

void
operator delete[] (void *ptr, size_t, gstd::align_val_t) noexcept
{
  free (ptr);
}

void
operator delete[] (void *ptr, gstd::nothrow_t &) noexcept
{
  free (ptr);
}

void
operator delete[] (void *ptr, gstd::align_val_t,
                   gstd::nothrow_t &) noexcept
{
  free (ptr);
}

void
operator delete[] (void *, void *) noexcept
{}
