/*
alloc.cpp - Dynamic memory allocation function implementations

This file is a part of Fiku.
Copyright (C) 2021 Akib Azmain Turja

Fiku is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Fiku is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

Authors: Akib Azmain Turja
*/

/*
 * For the readers, some terms used here:
 *
 * superblock - constant amount (depends on platform) of data, on x86 it is
 * 0x1000 or 4096 bytes
 *
 * page - 0x1000 or 4096 bytes of data
 *
 * metapage - page holding metadata
 *
 * AVLX - AVL (Adelson-Velsky and Landis) eXtended tree. This is an
 * extended version of AVL tree. AVL tree doesn't support nodes with same
 * key, AVLX adds suppert for that by adding a bottom child node which
 * holds node with idenitical key.
 *
 *             C
 *           / | \
 *          B  C  D
 *        /    |  |
 *       A     C  D
 *
 *  figure: an AVLX tree
 */

#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <stdint.h>
#include <fiku/paging.hpp>
#include <errno.h>
#include <fiku/cpu.h>
#include <fiku/spinlock.h>

namespace
{

  // Structure for holding the data about a superblock
  struct superblock_t
  {

    // Address to superblock
    void *key;

    // Size of superblock
    size_t data;

    // Hold whether the superblock is allocated with paging::kpalloc
    bool page_allocated;

    // Height of AVLX tree
    ptrdiff_t height;

    // Left node in AVLX tree
    superblock_t *left;

    // Right node in AVLX tree
    superblock_t *right;

    // Bottom node in AVLX tree
    superblock_t *bottom;
  };

  // Structure for holding information about a address and its block node
  struct addr_t;

  // Structure for holding information about a block
  struct block_t
  {

    // Size of block, key of AVL tree
    size_t key;

    // Address to block
    void *data;

    // Pointer to previous block
    addr_t *prev;

    // Pointer to next block
    addr_t *next;

    // Height of AVLX tree
    ptrdiff_t height;

    // Left node in AVLX tree
    block_t *left;

    // Right node in AVLX tree
    block_t *right;

    // Bottom node in AVLX tree
    block_t *bottom;
  };

  // Structure for holding information about a address and its block node
  struct addr_t
  {

    // Size of block, key of AVL tree
    void *key;

    // Address to block
    block_t *data;

    // Height of AVLX tree
    size_t height;

    // Whether the block is used
    bool used;

    // Left node in AVLX tree
    addr_t *left;

    // Right node in AVLX tree
    addr_t *right;

    // Bottom node in AVLX tree
    addr_t *bottom;
  };

  template < class arg_t > constexpr arg_t
  max (arg_t a, arg_t b)
  {
    return a < b ? b : a;
  }

  template < class arg_t > constexpr arg_t
  min (arg_t a, arg_t b)
  {
    return a < b ? a : b;
  }

  template < class node_t > node_t *
  avlx_rotate_left (node_t *root)
  {

    // Rotate
    node_t *pivot = root->right;
    root->right = pivot->left;
    pivot->left = root;

    // Update heights
    auto update_height = [] (node_t *node) -> void
    {
      node->height = 1 + max ((node->left ? node->left->height : 0),
                              (node->right ? node->right->height : 0));
    };

    update_height (root);
    update_height (pivot);

    // Return new root
    return pivot;
  }

  template < class node_t > node_t *
  avlx_rotate_right (node_t *root)
  {

    // Rotate
    node_t *pivot = root->left;
    root->left = pivot->right;
    pivot->right = root;

    // Update heights
    auto update_height = [] (node_t *node) -> void
    {
      node->height = 1 + max ((node->left ? node->left->height : 0),
                              (node->right ? node->right->height : 0));
    };

    update_height (root);
    update_height (pivot);

    // Return new root
    return pivot;
  }

  template < class node_t > node_t *
  avlx_insert (node_t *root, node_t *node)
  {
    if (!root)
      {
        node->height = 1;
        node->left = nullptr;
        node->right = nullptr;
        node->bottom = nullptr;
        return node;
      }

    // Insert
    if (node->key == root->key)
      {
        node->bottom = root->bottom;
        root->bottom = node;
        return root;
      }
    if (node->key < root->key)
      {
        root->left = avlx_insert (root->left, node);
      }
    else
      {
        root->right = avlx_insert (root->right, node);
      }

    // Update height
    root->height = 1 + max ((root->left ? root->left->height : 0),
                            (root->right ? root->right->height : 0));

    // Balance
    auto bf = [] (node_t *node) -> ptrdiff_t
    {
      return ((node->left ? node->left->height : 0)
              - (node->right ? node->right->height : 0));
    };

    if (bf (root) > 1)
      {
        if (node->key < root->left->key)
          {
            root = avlx_rotate_right (root);
          }
        else
          {
            root->left = avlx_rotate_left (root->left);
            root = avlx_rotate_right (root);
          }
      }
    else if (bf (root) < -1)
      {
        if (node->key < root->right->key)
          {
            root->right = avlx_rotate_right (root->right);
            root = avlx_rotate_left (root);
          }
        else
          {
            root = avlx_rotate_left (root);
          }
      }

    return root;
  }

  template < class node_t > node_t *
  avlx_delete (node_t *root, node_t *node)
  {

    // If root is null, return null
    if (!root)
      {
        return nullptr;
      }

    // If the current node is the node to delete, delete it
    else if (root == node)
      {

        // If the node has bottom nodes, replace with the bottom node
        if (node->bottom)
          {
            node->bottom->left = root->left;
            node->bottom->right = root->right;
            root = node->bottom;
            return root;
          }

        // Otherwise, if the node has no child, simple delete it
        else if (!node->left && !node->right)
          {
            root = nullptr;
          }

        // Otherwise, if the node has one child, replace it with that
        else if (!node->left)
          {
            root = node->right;
          }
        else if (!node->right)
          {
            root = node->left;
          }

        // Otherwise, the node has two children, replace the node with the
        // node with the maximum key value
        else
          {

            // Get the node with minimum key in right sub-tree
            node_t *right_min = node->right;
            while (right_min->left)
              {
                right_min = right_min->left;
              }

            right_min->right = avlx_delete (root->right, right_min);
            right_min->left = root->left;
            root = right_min;
          }

        // Remove external references from node to prevent infinite loops
        node->left = nullptr;
        node->right = nullptr;
        node->bottom = nullptr;
      }

    // Otherwise, find and delete the node in sub-trees
    else if (node->key == root->key)
      {
        root->bottom = avlx_delete (root->bottom, node);
      }
    else if (node->key < root->key)
      {
        root->left = avlx_delete (root->left, node);
      }
    else
      {
        root->right = avlx_delete (root->right, node);
      }

    // If root in nullptr, return
    if (!root)
      {
        return nullptr;
      }

    // Update height
    root->height = 1 + max ((root->left ? root->left->height : 0),
                            ((root->right) ? root->right->height : 0));

    // Balance
    auto bf = [] (node_t *node) -> ptrdiff_t
    {
      return ((node->left ? node->left->height : 0)
              - (node->right ? node->right->height : 0));
    };

    if (bf (root) > 1)
      {
        if (bf (root->left) >= 0)
          {
            root = avlx_rotate_right (root);
          }
        else
          {
            root->left = avlx_rotate_left (root->left);
            root = avlx_rotate_right (root);
          }
      }
    else if (bf (root) < -1)
      {
        if (bf (root->right) <= 0)
          {
            root = avlx_rotate_left (root);
          }
        else
          {
            root->right = avlx_rotate_right (root->right);
            root = avlx_rotate_left (root);
          }
      }

    // Return new node
    return root;
  }

  block_t *
  avlx_search_size (block_t *root, size_t size, size_t alignment,
                   size_t loss = 0)
  {
    block_t *best = nullptr;
    bool loss_inited;

    if (loss == 0)
      {
        loss_inited = false;
      }

    if (!root)
      {
        return nullptr;
      }
    else
      {

        // Calculate loss due to alignment
        size_t align_loss = (((reinterpret_cast < size_t > (root->data)
                               + (alignment - 1)) & ~(alignment - 1))
                             - reinterpret_cast < size_t > (root->data));

        if (root->key >= align_loss + size)
          {
            size_t tmp_loss = root->key - size;

            // Return if loss is zero
            if (tmp_loss == 0)
              {
                return root;
              }

            // If the loss is less than before or loss is uninitialized,
            // this is the best node among the visited nodes
            if (!loss_inited || tmp_loss < loss)
              {
                loss = tmp_loss;
                loss_inited = true;
                best = root;
              }

            // Check nodes with same size
            if (root->bottom)
              {
                auto node = avlx_search_size (root->left, size, alignment,
                                              loss);

                // If it returned a node, it is better than before
                if (node)
                  {
                    loss = node->key - size;
                    loss_inited = true;
                    best = node;
                  }
              }
          }

        // Check nodes with less sizes if possible
        if (size < root->key)
          {
            auto node = avlx_search_size (root->left, size, alignment,
                                          loss);

            // If it returned a node, it is better than before
            if (node)
              {
                loss = node->key - size;
                loss_inited = true;
                best = node;
              }

            // Return if loss is zero
            if (loss_inited && loss == 0)
              {
                return best;
              }
          }

        // Check nodes with more size
        auto node = avlx_search_size (root->right, size, alignment, loss);

        // If it returned a node, it is better than before
        if (node)
          {
            loss = node->key - size;
            loss_inited = true;
            best = node;
          }
      }

    // Return the base found
    return best;
  }

  template < class node_t > node_t *
  avlx_search (node_t *root, decltype (node_t::key) key)
  {

    // If root is a valid node, test it and it's children
    if (root)
      {
        if (key == root->key)
          {
            return root;
          }
        else if (key < root->key)
          {
            return avlx_search (root->left, key);
          }
        else
          {
            return avlx_search (root->right, key);
          }
      }

    // Otherwise, not found, and return nullptr
    else
      {
        return nullptr;
      }
  }

  // Structure for holding metadata
  struct metadata_t
  {

    // (Random) Magic number for a allocated metadata block
    static constexpr uint64_t magic_alloc = 0xd9a2ced9969024d9;

    // Magic number of the metadata block
    uint64_t magic;

    // Union for holding the metadata
    union data_t
    {

      // Don't rename; used to determining the position of allocated memory
      // of the other members
      char addr[0];

      // Possible data structures to be stored
      superblock_t superblock_s;
      block_t block_s;
      addr_t addr_s;
    };

    // Memory for the data
    data_t data;
  };

  // Structure the header of a page allocated for metadata
  struct meta_header_t
  {

    // The number of metadata blocks in the page
    static const size_t block_count;

    // The number of free metadata blocks in the page
    size_t free_block_count;

    // Data of previous superblock
    meta_header_t *prev;

    // Data of next superblock
    meta_header_t *next;

    // Metadata blocks
    metadata_t blocks[0];
  };

  constexpr const size_t meta_header_t::block_count =
    ((0x1000 - offsetof (meta_header_t, blocks)) / sizeof (metadata_t));

  // Pointer to first metapage
  meta_header_t *first_metapage;

  // Pointer to last metapage
  meta_header_t *last_metapage;

  meta_header_t *
  alloc_meta_page ()
  {

    // Allocate new page
    auto header = (static_cast < meta_header_t *>
                   (paging::kpalloc (1, 0, true, false)));

    if (errno == EAGAIN)
      {
        return nullptr;
      }

    // Nullify contents
    memset (header, 0, 0x1000);

    // Set header fields
    header->next = nullptr;
    header->prev = last_metapage;
    header->free_block_count = header->block_count;

    // Update pointers to first and last metapage
    if (!first_metapage)
      {
        first_metapage = header;
      }
    else
      {
        last_metapage->next = header;
      }

    last_metapage = header;

    // Return
    return header;
  }

  void
  free_meta_page (meta_header_t *header)
  {

    // Update previous metapage header
    if (header->prev)
      {
        header->prev->next = header->next;
      }
    else
      {
        first_metapage = header->next;
      }

    // Update next metapage header
    if (header->next)
      {
        header->next->prev = header->prev;
      }
    else
      {
        last_metapage = header->prev;
      }

    // Deallocate
    paging::kpfree (header, 1);
  }

  void *
  alloc_metadata ()
  {

    // Find a free node
    for (auto pg = first_metapage; pg != nullptr; pg = pg->next)
      {

        // Check if current page has any free block
        if (pg->free_block_count > 0)
          {

            // Search for a free block
            for (size_t i = 0; i < pg->block_count; ++i)
              {
                if (pg->blocks[i].magic != pg->blocks[i].magic_alloc)
                  {
                    --pg->free_block_count;
                    pg->blocks[i].magic = pg->blocks[i].magic_alloc;
                    return pg->blocks[i].data.addr;
                  }
              }
          }
      }

    // None found, allocate a new page
    auto pg = alloc_meta_page ();
    --pg->free_block_count;

    // Return the first node
    pg->blocks[0].magic = pg->blocks[0].magic_alloc;
    return pg->blocks[0].data.addr;
  }

  void
  free_metadata (void *node)
  {

    // Nullify node
    memset ((static_cast < char *> (node)
             - (offsetof (metadata_t, data)
                + offsetof (metadata_t::data_t, addr))),
            0, sizeof (metadata_t));

    // Increase free block count and deallocate page is totally free
    auto page = (reinterpret_cast < meta_header_t *>
                 (reinterpret_cast < size_t > (node) & ~0xfff));

    // Increase free block count
    ++page->free_block_count;

    // Deallocate the page if unused
    if (page->free_block_count == page->block_count)
      {
        free_meta_page (page);
      }
  }

  // AVL tree for superblocks
  superblock_t *superblock_tree;

  superblock_t *
  alloc_superblock (size_t size, size_t alignment)
  {

    // If size is not divisible by 0x1000, increment it to a such value
    size = (size + 0xfff) / 0x1000 * 0x1000;

    // Allocate new page
    void *mem = paging::kpalloc (size / 0x1000, alignment, true, false);

    // Check if the page is allocated
    if (errno == EAGAIN)
      {
        return nullptr;
      }

    auto data = static_cast < superblock_t *> (alloc_metadata ());

    data->key = mem;
    data->data = size;
    data->page_allocated = true; // Is it allocated with paging::kpalloc?

    // Update pointers to first and last superblock
    superblock_tree = avlx_insert (superblock_tree, data);

    return data;
  }

  void
  free_superblock (superblock_t *data)
  {

    // Deallocate only if allocated with paging::kpalloc
    if (data->page_allocated)
      {

        // Remove from AVL tree
        superblock_tree = avlx_delete (superblock_tree, data);

        // Deallocate page(s)
        paging::kpfree (data->key, data->data / 0x1000);

        // Deallocate metadata
        free_metadata (data);
      }
  }

  // AVL tree with block size as key
  block_t *size_tree;

  // AVL tree with block address as key
  addr_t *addr_tree;

  spnlk_t lock;

  int
  accuire ()
  {
    cpu_cli ();
    return spnlk_lock (&lock);
  }

  int
  release ()
  {
    int ret = spnlk_unlock (&lock);
    cpu_sti ();
    return ret;
  }

  void
  init ()
  {
    static bool initialized;
    if (!initialized)
      {

        // Initialize variables
        first_metapage = nullptr;
        last_metapage = nullptr;
        superblock_tree = nullptr;
        size_tree = nullptr;
        addr_tree = nullptr;
        initialized = true;
        spnlk_init (&lock);
      }
  }
}

#ifdef LIBC_MALLOC
void *
malloc (size_t size)
{
  return aligned_alloc (alignof (max_align_t), size);
}
#endif

#ifdef LIBC_CALLOC
void *
calloc (size_t num, size_t size)
{
  return memset (malloc (num * size), 0, num * size);
}
#endif

#ifdef LIBC_REALLOC
void *
realloc (void *ptr, size_t size)
{
  return aligned_realloc (alignof (max_align_t), ptr, size);
}
#endif

#ifdef LIBC_FREE
void
free (void *ptr)
{

  // Do nothing if nullptr
  if (ptr == nullptr)
    {
      return;
    }

  // Accuire lock
  accuire ();

  addr_t *addr = avlx_search (addr_tree, ptr);

  // Delete address node before modifing
  addr_tree = avlx_delete (addr_tree, addr);

  // Set block to unused
  addr->used = false;

  // Merge next free blocks with this one
  while (true)
    {
      if (!addr->data->next || addr->data->next->used)
        {
          break;
        }

      addr_tree = avlx_delete (addr_tree, addr->data->next);
      size_tree = avlx_delete (size_tree, addr->data->next->data);

      addr->data->key += addr->data->next->data->key;
      auto next = addr->data->next->data->next;

      free_metadata (addr->data->next->data);
      free_metadata (addr->data->next);

      addr->data->next = next;
    }

  // Merge previous free blocks with this one
  while (true)
    {
      if (!addr->data->prev || addr->data->prev->used)
        {
          break;
        }

      addr_tree = avlx_delete (addr_tree, addr->data->prev);
      size_tree = avlx_delete (size_tree, addr->data->prev->data);

      addr->data->key += addr->data->prev->data->key;
      addr->data->data = addr->data->prev->data->data;
      addr->key = addr->data->data;
      auto prev = addr->data->prev->data->prev;

      free_metadata (addr->data->prev->data);
      free_metadata (addr->data->prev);

      addr->data->prev = prev;
    }

  // If a dynamic superblock became totally free, free the superblock
  auto superblock = avlx_search (superblock_tree, addr->key);

  if (superblock && superblock->page_allocated
      && superblock->data == addr->data->key)
    {
      free_superblock (superblock);
      free_metadata (addr->data);
      free_metadata (addr);
    }
  else
    {
      addr_tree = avlx_insert (addr_tree, addr);
      size_tree = avlx_insert (size_tree, addr->data);
    }

  // Release the lock
  release ();
}
#endif

#ifdef LIBC_ALIGNED_ALLOC
void *
aligned_alloc (size_t alignment, size_t size)
{
  init ();

  // What to do when no memory is wanted? Return an invalid pointer
  if (size == 0)
    {
      return nullptr;
    }

  // Accuire lock
  accuire ();

  block_t *node = avlx_search_size (size_tree, size, alignment);
  addr_t *addr;

  // Check if a suitable block is found
  if (node)
    {
      size_tree = avlx_delete (size_tree, node);
      addr = avlx_search (addr_tree, node->data);
      addr_tree = avlx_delete (addr_tree, addr);
    }

  // Otherwise, allocate new superblock
  else
    {
      superblock_t *superblock = alloc_superblock (size, alignment);

      if (errno == EAGAIN)
        {
          errno = 0;
          release ();
          return nullptr;
        }

      node = static_cast < block_t *> (alloc_metadata ());

      if (!node)
        {

          // Make sure there is no memory leakage
          free_superblock (superblock);
          release ();
          return nullptr;
        }

      addr = static_cast < addr_t *> (alloc_metadata ());

      if (!addr)
        {

          // Make sure there is no memory leakage
          free_superblock (superblock);
          free_metadata (superblock);
          free_metadata (node);
          release ();
          return nullptr;
        }

      // Set fields of the new nodes
      node->key = superblock->data;
      node->data = superblock->key;
      node->prev = nullptr;
      node->next = nullptr;
      addr->key = node->data;
      addr->data = node;
    }

  // If the block perfectly fits, just allocate it
  if (node->key == size && (reinterpret_cast < uint32_t > (node->data)
                            & (alignment - 1)) == 0)
    {
      addr->used = true;
      addr_tree = avlx_insert (addr_tree, addr);
      release ();
      return node->data;
    }
  else
    {
      size_t aligned_offset = (((reinterpret_cast < size_t > (node->data)
                                 + (alignment - 1)) & ~(alignment - 1))
                               - reinterpret_cast < size_t > (node->data));

      bool prev_block = false;
      block_t *block_prev = nullptr;
      addr_t *addr_prev = nullptr;
      bool next_block = false;
      block_t *block_next = nullptr;
      addr_t *addr_next = nullptr;

      // If the block doesn't fulfill alignment requirements, break it
      if (aligned_offset)
        {
          prev_block = true;
          block_prev = static_cast < block_t *> (alloc_metadata ());
          addr_prev = static_cast < addr_t *> (alloc_metadata ());

          if (block_prev && addr_prev)
            {
              block_prev->key = aligned_offset;
              block_prev->data = node->data;
              block_prev->prev = node->prev;
              block_prev->next = addr;
              node->prev = addr_prev;
              addr_prev->key = block_prev->data;
              addr_prev->data = block_prev;
            }
        }

      // If block is larger than required, break it (again)
      if (node->key - (size + aligned_offset) != 0)
        {
          next_block = true;
          block_next = static_cast < block_t *> (alloc_metadata ());
          addr_next = static_cast < addr_t *> (alloc_metadata ());

          if (block_next && addr_next)
            {
              block_next->key = node->key - (size + aligned_offset);
              block_next->data = (static_cast < char *> (node->data)
                                  + aligned_offset + size);

              block_next->next = node->next;
              block_next->prev = addr;
              node->next = addr_next;
              addr_next->key = block_next->data;
              addr_next->data = block_next;
            }
        }

      // If any allocated failed, make sure there is no memory leakage and
      // return nullptr
      if ((prev_block && (!block_prev || !addr_prev))
          || (next_block && (!block_next || !addr_next)))
        {
          if (block_prev)
            {
              free_metadata (block_prev);
            }

          if (addr_prev)
            {
              free_metadata (addr_prev);
            }

          if (block_next)
            {
              free_metadata (block_next);
            }

          if (addr_next)
            {
              free_metadata (addr_next);
            }
        }

      node->key = size;
      node->data = static_cast < char *> (node->data) + aligned_offset;

      addr->key = node->data;
      addr->used = true;

      addr_tree = avlx_insert (addr_tree, addr);

      if (prev_block)
        {
          size_tree = avlx_insert (size_tree, block_prev);
          addr_tree = avlx_insert (addr_tree, addr_prev);
        }

      if (next_block)
        {
          size_tree = avlx_insert (size_tree, block_next);
          addr_tree = avlx_insert (addr_tree, addr_next);
        }

      // Release lock and return
      release ();
      return node->data;
    }
}
#endif

#ifdef LIBC_ALIGNED_REALLOC
void *
aligned_realloc (size_t alignment, void *ptr, size_t size)
{

  // Accuire lock
  accuire ();

  // Get the size of block
  size_t prev_size = avlx_search (addr_tree, ptr)->data->key;

  // Release lock
  release ();

  // If the current block meets the requirements, do nothing
  if (prev_size == size &&
      (reinterpret_cast < size_t > (ptr) & (alignment - 1)) == 0)
    {
      return ptr;
    }

  // Allocate new memory
  void *new_ptr = aligned_alloc (alignment, size);

  // Copy data
  memcpy (new_ptr, ptr, min (prev_size, size));

  // Free previous memory
  free (ptr);

  // Return pointer to new memory
  return new_ptr;
}
#endif
