/*
stdio.cpp - Standard I/O C function implementations

This file is a part of Fiku.
Copyright (C) 2021 Akib Azmain Turja

Fiku is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Fiku is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

Authors: Akib Azmain Turja
*/

#include <stdio.h>
#include <stdarg.h>
#include <stddef.h>
#include <stdint.h>
#include <fiku/vga.h>
#include <stdlib.h>
#include <string.h>

namespace
{
  // Calls given function for each character in computed formatted string
  // ('c' stands for caller)
  template < class func_t > int
  vcprintf (func_t function, bool limit, size_t max, const char *format,
            va_list ap)
  {

    // How many bytes written? Zero.
    int written = 0;

    // Wrapper function to prevent stack overflow
    auto func = [&] (char ch) -> void
    {
      if (!limit || static_cast < size_t > (written) < max)
        {
          function (ch);
        }
    };

    // Iterate through the array
    for (size_t i = 0; format[i] != '\0'; i++)
      {

        // Is it the special character?
        if (format[i] == '%')
          {

            // Is the character repeated?
            if (format[++i] == '%')
              {
                func ('%');
                written++;
                continue;
              }

            // All options that can be specified...

            // Justify on which side?
            bool left_justify = false;

            // Force printing the sign? No.
            bool force_signed = false;

            // Use space when positive? No.
            bool space_if_no_sign = false;

            // Do alternative convertion? No.
            bool alternative_convertion = false;

            // Pad with zeros?
            bool zero_padding = false;

            // Minimum width
            unsigned int min_width = 0;

            // Have you specified the precision, caller?
            bool precision_specified = false;

            // Precision
            unsigned int precision = 0;

            // Argument length specifiers
            bool len_hh = false;
            bool len_h = false;
            bool len_ll = false;
            bool len_l = false;
            bool len_j = false;
            bool len_z = false;
            bool len_t = false;
            bool len_L = false;

            // Parse flags
            for (;; i++)
              {

                // Justify on left?
                if (format[i] == '-')
                  {
                    left_justify = true;
                  }

                // Force to print the sign?
                else if (format[i] == '+')
                  {
                    force_signed = true;
                  }

                // Print space instead of plus sign?
                else if (format[i] == ' ')
                  {
                    space_if_no_sign = true;
                  }

                // Alternative convertion?
                else if (format[i] == '#')
                  {
                    alternative_convertion = true;
                  }

                // Pad with zeros?
                else if (format[i] == '0')
                  {
                    zero_padding = true;
                  }

                // No rule to find more flags.  Stop.
                else
                  {
                    break;
                  }
              }

            // Get minimum width if specified
            if (format[i] == '*' || (format[i] >= '0' && format[i] <= '9'))
              {
                int min_width_signed;

                // Where's the minimun width specified?
                if (format[i] == '*')
                  {
                    min_width_signed = va_arg (ap, int);
                    i++;
                  }
                else
                  {
                    min_width_signed = atoi (format + i);
                    for (; (format[i] >= '0' && format[i] <= '9'); i++);
                  }

                // Negative width?
                if (min_width_signed < 0)
                  {

                    // Assuming the caller meant to justify left
                    min_width = -min_width_signed;
                    left_justify = true;
                  }
                else
                  {
                    min_width = min_width_signed;
                  }
              }

            // Did you specify the precision, caller?
            if (format[i] == '.')
              {
                i++;
                int precision_signed;

                // A dedicated argument for precision?
                if (format[i] == '*')
                  {
                    precision_signed = va_arg (ap, int);
                    i++;
                  }

                // Embed in format?
                else if (format[i] >= '0' && format[i] <= '9')
                  {
                    precision_signed = atoi (format + i);
                    for (; format[i] >= '0' && format[i] <= '9'; i++);
                  }

                // Neither specified, assuming 0
                else
                  {
                    precision_signed = 0;
                  }

                // Is precision positive?
                if (precision_signed >= 0)
                  {
                    precision = precision_signed;
                    precision_specified = true;
                  }
              }

            // Parse length modifiers
            for (;; i++)
              {

                // Modifier "h"?
                if (format[i] == 'h')
                  {

                    // Is it "hh" or just "h"?
                    if (format[i + 1] == 'h')
                      {
                        len_hh = true;
                        i++;
                      }
                    else
                      {
                        len_h = true;
                      }
                  }

                // Modifier "l"?
                else if (format[i] == 'l')
                  {

                    // Is it "ll" or just "l"?
                    if (format[i + 1] == 'l')
                      {
                        len_ll = true;
                        i++;
                      }
                    else
                      {
                        len_l = true;
                      }
                  }

                // Modifier "j"?
                else if (format[i] == 'j')
                  {
                    len_j = true;
                  }

                // Modifier "z"?
                else if (format[i] == 'z')
                  {
                    len_z = true;
                  }

                // Modifier "t"?
                else if (format[i] == 't')
                  {
                    len_t = true;
                  }

                // Modifier "L"?
                else if (format[i] == 'L')
                  {
                    len_L = true;
                  }

                // No more modifiers, break
                else
                  {
                    break;
                  }
              }

            // What should be converted to string? character?
            if (format[i] == 'c')
              {
                int ch = va_arg (ap, int);
                unsigned int padding = min_width > 1 ? min_width - 1 : 0;

                // Print padding if required
                if (!left_justify)
                  {
                    for (size_t i = 0; i < padding; i++)
                      {
                        func (' ');
                        written++;
                      }
                  }

                // Print the character
                func (static_cast < char > (ch));
                written++;

                // Print padding if required
                if (!left_justify)
                  {
                    for (size_t i = 0; i < padding; i++)
                      {
                        func (' ');
                        written++;
                      }
                  }
              }

            // Write a string?
            else if (format[i] == 's')
              {
                const char *str = static_cast < const char *> (va_arg (ap, void *));
                size_t len = strlen (str);

                // Is the precision, maximum bytes to write (excluding
                // padding), specified?
                if (precision_specified && len > precision)
                  {
                    len = precision;
                  }

                unsigned int padding = min_width > len ? min_width - len : 0;

                // Print the padding if required
                if (!left_justify)
                  {
                    for (size_t i = 0; i < padding; i++)
                      {
                        func (' ');
                        written++;
                      }
                  }

                // Write the string
                for (size_t i = 0; i < len; i++)
                  {
                    func (str[i]);
                    written++;
                  }

                // Print the padding if required
                if (left_justify)
                  {
                    for (size_t i = 0; i < padding; i++)
                      {
                        func (' ');
                        written++;
                      }
                  }
              }
            else if (format[i] == 'd' || format[i] == 'i' || format[i] == 'u')
              {
                uintmax_t num;
                bool signed_convertion = false;
                bool negative = false;

                // Is it signed or unsigned?
                if (format[i] == 'u')
                  {
                    if (len_ll)
                      {
                        num = va_arg (ap, unsigned long long);
                      }
                    else
                      {
                        num = va_arg (ap, unsigned int);
                      }
                  }
                else
                  {

                    // We are doing signed convertion, so convert signed number
                    // to unsigned
                    signed_convertion = true;
                    intmax_t snum;
                    if (len_ll)
                      {
                        snum = va_arg (ap, long long);
                      }
                    else
                      {
                        snum = va_arg (ap, int);
                      }

                    // Is it negative?
                    if (snum < 0)
                      {
                        negative = true;
                        num = -snum;
                      }
                    else
                      {
                        num = snum;
                      }
                  }

                // Precision specified?
                if (!precision_specified)
                  {
                    precision = 1;
                  }

                // A number, that meets the following condition:
                // num / divisor = x where 0 <= x < 10
                size_t divisor = 1;
                size_t len = (signed_convertion
                              && ((negative | force_signed | space_if_no_sign)
                                  ? 1 : 0));

                for (uintmax_t i = num; i != 0; i /= 10)
                  {
                    if (i >= 10)
                      {
                        divisor *= 10;
                      }

                    len++;
                  }

                // The number of digits required to print the number
                size_t digit_count = precision > len ? precision : len;

                // How many zeros required to satisfy the precision?
                size_t extra_zeros = precision > len ? precision - len : 0;

                // How bytes for padding?
                size_t padding = (min_width > digit_count
                                  ? min_width - digit_count : 0);

                // How bytes left for printing?
                size_t left = len;

                // Print padding if required
                if (!left_justify)
                  {
                    for (size_t i = 0; i < padding; i++)
                      {
                        func (zero_padding && !precision_specified
                              ? '0' : ' ');
                        ++written;
                      }
                  }

                // Print the extra leading zeros
                for (uint_fast16_t i = 0; i < extra_zeros; i++)
                  {
                    func ('0');
                    ++written;
                  }

                // Print the sign if required
                if (signed_convertion)
                  {

                    // If it's negative, printing sign is mandatory
                    if (negative)
                      {
                        func ('-');
                        ++written;
                        --left;
                      }
                    else
                      {

                        // Does the caller explicitly specfied to print the
                        // sign?
                        if (force_signed)
                          {
                            func ('+');
                            ++written;
                            --left;
                          }

                        // Space instead of the plus sign?
                        else if (space_if_no_sign)
                          {
                            func (' ');
                            ++written;
                            --left;
                          }
                      }
                  }

                // Print the number
                for (; left != 0; divisor /= 10)
                  {
                    func ('0' + (num / divisor));
                    ++written;
                    --left;
                    num %= divisor;
                  }

                // Print padding if required
                if (left_justify)
                  {
                    for (size_t i = 0; i < padding; i++)
                      {
                        func (' ');
                        written++;
                      }
                  }
              }

            // Convert to octal or hexadecimal?
            else if (format[i] == 'o' || format[i] == 'x' || format[i] == 'X')
              {
                uintmax_t unum;
                size_t target_type_size;

                // What's the size of the integer type being converted?
                if (len_hh)
                  {
                    target_type_size = sizeof (unsigned char);
                  }
                else if (len_h)
                  {
                    target_type_size = sizeof (unsigned short);
                  }
                else if (len_l)
                  {
                    target_type_size = sizeof (unsigned long);
                  }
                else if (len_ll)
                  {
                    target_type_size = sizeof (unsigned long long);
                  }
                else if (len_j)
                  {
                    target_type_size = sizeof (uintmax_t);
                  }
                else if (len_z)
                  {
                    target_type_size = sizeof (size_t);
                  }
                else if (len_L)
                  {
                    target_type_size = sizeof (ptrdiff_t);
                  }
                else
                  {
                    target_type_size = sizeof (unsigned int);
                  }

                if (len_ll)
                  {
                    unum = va_arg (ap, unsigned long long);
                  }
                else
                  {
                    unum = va_arg (ap, unsigned int);
                  }

                // How many bits are used per digit in target base? 3 for
                // octal, 4 for hexadecimal
                size_t bits_per_digit = format[i] == 'o' ? 3 : 4;

                // What's the maximum number that can be expressed in one
                // digit?
                size_t max_one_digit_num = format[i] == 'o' ? 0x7 : 0xf;

                // What's the maximum number of digits we can print?
                size_t max_digit_count = (target_type_size * 8
                                          / bits_per_digit);

                // Precision specified?
                if (!precision_specified)
                  {
                    precision = 1;
                  }

                size_t len = 0;

                // Count the number of digits required to print the number
                for (uint_fast16_t i = 0; i < max_digit_count; i++)
                  {
                    if (len > 0 || (unum >> ((max_digit_count - 1 - i)
                                             * bits_per_digit)
                                    & max_one_digit_num) != 0)
                      {
                        len++;
                      }
                  }

                // Alternative convertion to octal?
                if (alternative_convertion && format[i] == 'o' &&
                    precision < len + 1)
                  {
                    precision = len + 1;
                  }

                // The number of digits required to print the number
                size_t digit_count = precision > len ? precision : len;

                // Alternative conversion to hexadecimal?
                if (alternative_convertion && format[i] != 'o' && unum != 0)
                  {
                    digit_count += 2;
                  }

                // How many zeros required to satisfy the precision?
                size_t extra_zeros = precision > len ? precision - len : 0;

                // How bytes for padding?
                size_t padding = (min_width > digit_count
                                  ? min_width - digit_count : 0);

                // Print padding if required
                if (!left_justify)
                  {
                    for (size_t i = 0; i < padding; i++)
                      {
                        func (zero_padding && !precision_specified
                              ? '0' : ' ');
                        written++;
                      }
                  }

                // Alternative conversion to hexadecimal?
                if (alternative_convertion && format[i] != 'o'
                    && unum != 0)
                  {
                    func ('0');
                    func (format[i]);
                    written += 2;
                  }

                // Print the extra leading zeros
                for (uint_fast16_t i = 0; i < extra_zeros; i++)
                  {
                    func ('0');
                    written++;
                  }

                // Digits to use, first 8 for octal, all (except the
                // trailing-null of course) for hexadecimal
                const char *digits = format[i] == 'x' ?
                  "0123456789abcdef" : "0123456789ABCDEF";

                // Write the num
                for (uint_fast16_t i = 0; i < len; i++)
                  {
                    func (digits[(unum >> ((len - 1 - i) * bits_per_digit))
                                 & max_one_digit_num]);
                    written++;
                  }

                // Print padding if required
                if (left_justify)
                  {
                    for (size_t i = 0; i < padding; i++)
                      {
                        func (' ');
                        written++;
                      }
                  }
              }

            // Caller: I want to know the number of bytes written till now
            else if (format[i] == 'n')
              {
                void *ptr = va_arg (ap, void *);
                if (len_hh)
                  {
                    *static_cast < signed char *> (ptr) = written;
                  }
                else if (len_h)
                  {
                    *static_cast < short *> (ptr) = written;
                  }
                else if (len_l)
                  {
                    *static_cast < long *> (ptr) = written;
                  }
                else if (len_ll)
                  {
                    *static_cast < long long *> (ptr) = written;
                  }
                else if (len_j)
                  {
                    *static_cast < intmax_t * > (ptr) = written;
                  }
                else if (len_z)
                  {
                    *static_cast < size_t *> (ptr) = written;
                  }
                else if (len_t)
                  {
                    *static_cast < ptrdiff_t *> (ptr) = written;
                  }
              }

            // Write a pointer?
            else if (format[i] == 'p')
              {
                constexpr const char str[] = "(nullptr)";
                void *ptr = va_arg (ap, void *);
                size_t len = 0;

                if (ptr)
                  {

                    // Count the number of digits required to print the
                    // pointer
                    for (uint_fast16_t i = sizeof (ptr) * 2; i > 0; --i)
                      {
                        if (len > 0 || (reinterpret_cast < size_t > (ptr)
                                        >> ((i - 1) * 4) & 0xf) != 0)
                          {
                            len++;
                          }
                      }

                    len += 2;
                  }
                else
                  {
                    len = sizeof (str) - 1;
                  }

                size_t padding = min_width > len ? min_width - len : 0;

                // Justify if required
                if (!left_justify)
                  {
                    for (size_t i = 0; i < padding; i++)
                      {
                        func (' ');
                        written++;
                      }
                  }

                if (ptr)
                  {
                    // Print leading "0x"
                    func ('0');
                    func ('x');
                    written += 2;
                    constexpr const char digits[] = "0123456789abcdef";

                    // Print the pointer in hexadecimal notation
                    for (uint_fast16_t i = 0; i < len - 2; i++)
                      {
                        func (digits[reinterpret_cast < size_t > (ptr)
                                      >> ((len - 3 - i) * 4) & 0xf]);
                        written++;
                      }
                  }
                else
                  {
                    for (size_t i = 0; i < sizeof (str) - 1; ++i)
                      {
                        func (str[i]);
                      }
                  }

                // Justify if required
                if (left_justify)
                  {
                    for (size_t i = 0; i < padding; i++)
                      {
                        func (' ');
                        written++;
                      }
                  }
              }
          }

        // Nothing special, write it and continue
        else
          {
            func (format[i]);
            written++;
          }
      }

    return written;
  }

  void
  putc (char ch)
  {
    vga_putc (ch);
  }
}

#ifdef LIBC_VPRINTF
int
vprintf (const char *format, va_list ap)
{
  return vcprintf (putc, false, 0, format, ap);
}
#endif

#ifdef LIBC_PRINTF
int
printf (const char *format, ...)
{
  va_list ap;
  va_start (ap, format);
  int written = vcprintf (putc, false, 0, format, ap);
  va_end (ap);
  return written;
}
#endif
