/*
main.cpp - Main function

This file is a part of Fiku.
Copyright (C) 2021 Akib Azmain Turja

Fiku is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Fiku is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

Authors: Akib Azmain Turja
*/

#include <fiku/gdt.hpp>
#include <fiku/idt.hpp>
#include <fiku/multiboot2.hpp>
#include <fiku/paging.hpp>
#include <fiku/panic.h>
#include <fiku/cpu.h>
#include <fiku/vga.h>
#include <stdio.h>
#include <string.h>
#include <fiku/vas.hpp> // FIXME
#include <stdlib.h>
#include <errno.h>
#include <fiku/cpuid.h>
#include <fiku/proc.hpp>
#include <fiku/timers/pit.hpp>
#include <threads.h>
#include <fiku/syscall.hpp>
#include <fiku/cpu.h>
#include <fiku/io.h>
#include <fiku/loader.hpp>

void *multiboot_boot_info_ptr_;

namespace
{
  gdt::entry *gdt_;
  idt::gate *idt_;
  void *pgs_;

  // Initializes things like GDT, IDT and VGA
  __attribute__ ((constructor (0))) void
  init ()
  {
    cpu_init ();
    vga_init ();
    gdt_ = gdt::init ();
    idt_ = idt::init ();
    uint32_t esp[2];
    asm ("movl %%esp, %0;"
         "int $0x27;"
         "movl %%esp, %1;"
         : "=r" (esp[0]), "=r" (esp[1]));
    pgs_ = paging::init (multiboot_boot_info_ptr_);
    syscall::init ();
  }

  // Parses command line into argument vector (argv)
  size_t
  parse_cmd (const char *str, char ***argv)
  {
    struct arg
    {

      // String
      char *data;

      // Size
      size_t size;

      // Capacity
      size_t cap;

      // Pushes a character
      void
      push (char ch)
      {

        // Increase capacity if needed
        if (size == cap)
          {
            cap += 64;

            if (data)
              {
                data = static_cast < char *> (aligned_realloc
                                              (alignof (char), data, cap));
              }
            else
              {
                data = static_cast < char *> (aligned_alloc
                                              (alignof (char), cap));
              }
          }

        // Write character
        data[size++] = ch;
      }

      // Releases unused memory
      void
      shrink ()
      {
        cap = size;
        data = static_cast < char *> (aligned_realloc (alignof (char),
                                                       data, cap));
      }
    };

    struct args
    {

      // Strings
      char **data;

      // String count
      size_t size;

      // Pushes a string
      void
      push (char *arg)
      {

        // Increase capacity
        if (data)
          {
            data = static_cast < char **> (aligned_realloc
                                           (alignof (char *), data,
                                            (size + 1) * sizeof (char *)));
          }
        else
          {
            data = static_cast < char **> (aligned_alloc
                                           (alignof (char *),
                                            sizeof (char *)));
          }

        // Push the string
        data[size++] = arg;
      }
    };

    // List of arguments
    args arg_list = { nullptr, 0 };

    // Parse command line
    while (true)
      {

        // Break if string is parsed
        if (!*str)
          {
            break;
          }

        // String
        arg arg_str = { nullptr, 0, 0 };

        // Whether to escape next character?
        bool escape = false;

        // Are we inside quotes?
        bool quote = false;

        // Parse a argument
        while (true)
          {

            // Break if string is parsed
            if (!*str)
              {
                break;
              }

            // If the character is a space, handle it accordingly
            else if (*str == ' ')
              {
                if (escape)
                  {
                    escape = false;
                    arg_str.push (' ');
                  }
                else if (quote)
                  {
                    arg_str.push (' ');
                  }
                else
                  {
                    ++str;
                    break;
                  }
              }

            // If it is a escape character, handle it accordingly
            else if (*str == '\\')
              {
                if (escape)
                  {
                    escape = false;
                    arg_str.push ('\\');
                  }
                else
                  {
                    escape = true;
                  }
              }

            // If it is a quote, handle it accordingly
            else if (*str == '"')
              {
                if (escape)
                  {
                    escape = false;
                    arg_str.push ('"');
                  }
                else
                  {
                    quote = !quote;
                  }
              }

            // Otherwise, just add character to the argument
            else
              {
                escape = false;
                arg_str.push (*str);
              }

            ++str;
          }

        // Add trailing null, release unused memory and push it to argv
        arg_str.push ('\0');
        arg_str.shrink ();
        arg_list.push (arg_str.data);
      }

    // Return argc and argc
    *argv = arg_list.data;
    return arg_list.size;
  }

  int
  port_test (void *p)
  {
    if (!p)
      {
        uintptr_t port;
        printf ("creating port...\n");
        asm volatile ("int $0xff"
                      : "=b" (port)
                      : "a" (0x1), "c" (0b10));
        thrd_yield ();
        printf ("creating thrd...\n");
        thrd_t new_thrd;
        thrd_create (&new_thrd, port_test, &port);
        size_t data;
        printf ("waiting...\n");
        asm volatile ("pushl %%edx;"
                      "pushl %%esi;"
                      "pushl %%edi;"
                      "pushl %%ebp;"
                      "int $0xff;"
                      "popl %%ebp;"
                      "popl %%edi;"
                      "popl %%esi;"
                      "popl %%edx;"
                      : "=c" (data)
                      : "a" (0x4), "b" (port));
        panic ("%lx\n", data);
      }
    else
      {
        printf ("sending...\n");
        asm volatile ("int $0xff"
                      :
                      : "a" (0x3), "b" (*static_cast < uintptr_t *> (p)),
                        "c" (0xfeeddead));
        return 0;
      }
  }
}

size_t min_a, max_a, min_b, max_b;

extern "C" void
main ()
{
  multiboot2::boot_info boot_info (multiboot_boot_info_ptr_);
  const char *cmd_line = nullptr;

  for (auto &tag : boot_info)
    {
      switch (tag.type)
        {
        case 1:
          {
            cmd_line = tag.cast < multiboot2::boot_info::cmd_t > ().cmd;
            break;
          }
        }
    }

  // Parse arguments
  char **argv;
  size_t argc = parse_cmd (cmd_line, &argv);

  // Print arguments
  printf ("argc = %lu\n", argc);

  for (size_t i = 0; i < argc; ++i)
    {
      printf ("  %lu: \"%s\"\n", i, argv[i]);
    }

  vga_flush ();

  auto task_a = [] (void *arg) -> int
  {
    auto task_b = [] (void *arg) -> int
    {
      for (size_t i = 0; i < 1000; ++i)
        {
          asm volatile ("cli");
          printf ("b: arg: %p, equal: %s (%lu)\n", arg, thrd_equal (*reinterpret_cast < thrd_t *> (arg), thrd_current ()) ? "true" : "false", i);
          vga_flush ();
          asm volatile ("sti");
        }
      return 4;
    };

    asm volatile ("cli");
    printf ("a: creating thread\n");
    vga_flush ();
    asm volatile ("sti");
    thrd_t thrd;
    thrd_create (&thrd, task_b, &thrd);
    asm volatile ("cli");
    printf ("a: yielding\n");
    vga_flush ();
    thrd_yield ();
    printf ("a: waiting\n");
    vga_flush ();
    asm volatile ("sti");
    int retval;
    thrd_join (thrd, &retval);
    asm volatile ("cli");
    printf ("a: b returned %i\n", retval);
    printf ("a: arg: %p, equal: %s\n", arg, thrd_equal (*reinterpret_cast < thrd_t *> (arg), thrd_current ()) ? "true" : "false");
    vga_flush ();
    asm volatile ("sti");
    return 0;
  };

  thrd_t thread_a;
  thrd_create (&thread_a, task_a, &thread_a);

  struct arg_t
  {
    mtx_t *mutex;
    const char *str;
  };

  auto thrd = [] (void *argp)
  {
    arg_t *arg = reinterpret_cast < arg_t *> (argp);
    mtx_lock (arg->mutex);
    mtx_lock (arg->mutex);
    for (size_t i = 0; i < 10; ++i)
      {
        printf ("%s: %lu\n", arg->str, i);
        vga_flush ();
        thrd_yield ();
      }
    mtx_unlock (arg->mutex);
    mtx_unlock (arg->mutex);
    return 0;
  };

  mtx_t mutex;
  mtx_init (&mutex, mtx_plain | mtx_recursive);
  arg_t arg_a = { &mutex, "a" };
  arg_t arg_b = { &mutex, "b" };
  thrd_t thrd_a_, thrd_b_;
  thrd_create (&thrd_a_, thrd, &arg_a);
  thrd_create (&thrd_b_, thrd, &arg_b);

  void (*thrd_a) () = [] ()
  {
    size_t min = 50, max = 50;
    for (size_t i = 0; i < 100000; ++i)
      {
        size_t val = timer::pit.read ();
        cpu_cli ();
        printf ("executing a (for %2lu us) (%lu)...\n", val, i);
        vga_flush ();
        cpu_sti ();
        if (val > max)
          {
            max = val;
          }
        if (val < min)
          {
            min = val;
          }
        asm volatile ("sti");
      }
    min_a = min;
    max_a = max;
    while (true) asm ("int $0x30");
  };

  char stack_a[0x1000];

  void (*thrd_b) () = [] ()
  {
    size_t min = 50, max = 50;
    for (size_t i = 0; i < 100000; ++i)
      {
        uint64_t addr = paging::palloc ();
        size_t val = timer::pit.read ();
        cpu_cli ();
        printf ("executing b (for %2lu us) (%lu)...\n", val, i);
        vga_flush ();
        cpu_sti ();
        paging::pfree (addr);
        if (val > max)
          {
            max = val;
          }
        if (val < min)
          {
            min = val;
          }
      }
    min_b = min;
    max_b = max;
    while (true) asm ("int $0x30");
  };

  char stack_b[0x1000];

  vas::init ();
  proc::thrd_t thread;
//  thread.ttl = 2000;
  thread.thrd_state = thread.thrd_runnable;
  asm ("movl %%cr3, %0" : "=r" (thread.vas));
  thread.reg_state = new idt::registers_t;
  thread.reg_state->eip = reinterpret_cast < uint32_t > (thrd_a);
  thread.reg_state->eflags = 0x202;
  thread.reg_state->cs = 0x8;
  thread.reg_state->ds = 0x10;
  thread.reg_state->esp = reinterpret_cast < uint32_t > (stack_a + 0x1000);
  proc::thrd_prep (&thread);
  delete thread.reg_state;
  thread.reg_state = nullptr;

  proc::thrd_t thread_;
  thread_.thrd_state = thread_.thrd_runnable;
//  thread_.ttl = 10;
  asm ("movl %%cr3, %0" : "=r" (thread_.vas));
  thread_.reg_state = new idt::registers_t;
  thread_.reg_state->eip = reinterpret_cast < uint32_t > (thrd_b);
  thread_.reg_state->eflags = 0x202;
  thread_.reg_state->cs = 0x8;
  thread_.reg_state->ds = 0x10;
  thread_.reg_state->esp = reinterpret_cast < uint32_t > (stack_b + 0x1000);
  proc::thrd_prep (&thread_);
  delete thread_.reg_state;
  thread_.reg_state = nullptr;

  char stack_c[0x1000];
  vas::vas_t mem;
  mem.map (0x100, true, true);
  proc::thrd_t thread_u;
  thread_u.thrd_state = thread_u.thrd_runnable;
//  thread_u.ttl = 1000;
  thread_u.kernel_stack = stack_c + 0x1000;
  thread_u.vas = mem.data ();
  thread_u.reg_state = new idt::registers_t;
  thread_u.reg_state->eip = 0x100000;
  thread_u.reg_state->eflags = 0x202;
  thread_u.reg_state->cs = 0x1b;
  thread_u.reg_state->ds = 0x23;
  proc::thrd_prep (&thread_u);
  delete thread_u.reg_state;
  thread_u.reg_state = nullptr;
  paging::flush (mem.data ());
  *reinterpret_cast < uint16_t *> (0x100000) = 0xfeeb; // jmp ,

  proc::thrd_add (&thread_u);
  proc::thrd_add (&thread);
  proc::thrd_add (&thread_);

  auto printer = [] (uint8_t, size_t t, idt::registers_t &regs)
  {

    // Send EOI
    io_outb (0x20, 0x20);
    regs.eoi_sent = true;

    proc::thrd_yield (t, &regs);
  };

  auto switcher = [] (uint8_t, uint8_t, idt::registers_t &regs)
  {
    proc::thrd_yield (timer::pit.read (), &regs);
  };

  loader::prog_img img = { nullptr, 0 };

  for (auto &tag : boot_info)
    {
      if (tag.type == 3)
        {
          auto start_phys =
            tag.cast < multiboot2::boot_info::module_t > ().start;
          auto end_phys =
            tag.cast < multiboot2::boot_info::module_t > ().end;
          start_phys += 0xf0000000;
          end_phys += 0xf0000000;
          img = loader::load (reinterpret_cast < void *> (start_phys),
                              reinterpret_cast < void *> (end_phys));
          break;
        }
    }

  if (!img.addr_space)
    {
      panic ("error: unable to load program");
    }

  char stack_ul[0x1000];
  proc::thrd_t thread_ul;
  thread_ul.thrd_state = thread_ul.thrd_runnable;
//  thread_u.ttl = 1000;
  thread_ul.kernel_stack = stack_ul + 0x1000;
  thread_ul.vas = img.addr_space->data ();
  thread_ul.reg_state = new idt::registers_t;
  thread_ul.reg_state->eip = img.entry;
  thread_ul.reg_state->eflags = 0x202;
  thread_ul.reg_state->cs = 0x1b;
  thread_ul.reg_state->ds = 0x23;
  proc::thrd_prep (&thread_ul);
  delete thread_ul.reg_state;
  thread_ul.reg_state = nullptr;
  proc::thrd_add (&thread_ul);

//  thrd_t port_test_thrd;
//  thrd_create (&port_test_thrd, port_test, nullptr);

  timer::pit.set (1000, printer);
  idt::set_handler (0x30, switcher);
  proc::set_timer (timer::pit);
  cpu_current ()->cli_count = 0; // Don't sti, IF will set after the thread
                                 // has started
  proc::thrd_start ();
  vga_flush ();
  cpu_freeze ();
}
